/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_coord.c: Coordinates structures/routines
 *
 * Coordinates consist of three integer values corresponding to x, y and z coordinates.
 * Lists of coordinates are implemented using the coordinate_list struct.
 * This is used for various purposes throughout the program.
 * 
 * A coordinate list consists of a simple array containing the coordinates,
 * but also has a hash table which can be used to speed up finding whether or not
 * coordinates are contained in a list (taking O(1) time versus O(n) otherwise). See fw_coord.h for more information on this structure.
 *
 * We also use the concept of multi-threaded coordinate lists, which are sets of coordinate lists.
 * This is done so that each thread can write to a separate coordinate list when rasterizing polygons.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <mpi.h>

#include "fw_main.h"
#include "fw_coord.h"
#include "fw_voxel.h"

/* coordinate_hash: Simple coordinate hashing function
 * Returns hash value of @c belonging to @cl
 */

static inline int coordinate_hash(coordinate_list *cl, coordinate *c)
{
	return abs((c->x * c->y * c->z)) % cl->size;
}


/* coordinate_list_create: Create a new coordinate list of size @size.
 * Returns a pointer to the new coordinate list.
 */

coordinate_list	*coordinate_list_create(int size)
{
	coordinate_list *result;

	assert(size >= 0);

	result = fw_malloc(sizeof(struct coordinate_list));	
	result->size = size;
	result->used = 0;
	result->coords = fw_malloc(size * sizeof(struct coordinate));
	result->coordhash = fw_malloc(size * sizeof(coordinate_hash_entry));
	memset(result->coordhash, '\0', size * sizeof(coordinate_hash_entry));

	return result;
}


/* coordinate_list_free: Free the coordinate list @cl
 */

void			 coordinate_list_free(coordinate_list *cl)
{
	int i;
	coordinate_hash_entry *che, *chenext;

	// Free the hash table

	for(i = 0; i < cl->size; i++) {
		for(che = cl->coordhash[i].next; che != NULL; che = chenext) {
			chenext = che->next;
			fw_free(che);
		}
	}

	fw_free(cl->coordhash);
	fw_free(cl->coords);
	fw_free(cl);
}


/* coordinate_list_lookup: Check if coordinate 'c' is already in 'cl'.
 * Returns 1 if found, 0 if not. (See fw_coord.h for information on the hash structure)
 */

int			 coordinate_list_lookup(coordinate_list *cl, coordinate *c)
{
	assert(cl != NULL);
	assert(c != NULL);

	int hashindex = coordinate_hash(cl, c);
	coordinate_hash_entry *che;

	for(che = &cl->coordhash[hashindex]; che != NULL; che = che->next) {
		if(che->used == CSLOT_USED && (cl->coords[che->index].x == c->x && cl->coords[che->index].y == c->y && cl->coords[che->index].z == c->z)) {
			return 1;
		}
	}

	return 0;
}


/* coordinate_list_add: Add @c to coordinate list @cl. If @c is already in @cl, no action
 * is taken. (See fw_coord.h for information on the hash structure)
 */

void			 coordinate_list_add(coordinate_list *cl, coordinate *c)
{
	assert(cl != NULL);
	assert(c != NULL);
	assert(cl->used < cl->size);

	int hashindex = coordinate_hash(cl, c);
	coordinate_hash_entry *che, *cheprev;


	// Already in table?

	for(che = cheprev = &cl->coordhash[hashindex]; che != NULL; che = che->next) {
		if(che->used == CSLOT_USED && (cl->coords[che->index].x == c->x && cl->coords[che->index].y == c->y && cl->coords[che->index].z == c->z)) {
			return;
		}
		cheprev = che;
	}


	// Find unused slot. first slot is unused => use that, otherwise allocate a new slot

	if(cl->coordhash[hashindex].used == CSLOT_UNUSED)
		che = &cl->coordhash[hashindex];
	else
		che = cheprev->next = fw_malloc(sizeof(struct coordinate_hash_entry));

	cl->coords[cl->used].x = c->x;
	cl->coords[cl->used].y = c->y;
	cl->coords[cl->used].z = c->z;
	che->index = cl->used;
	che->used = CSLOT_USED;
	che->next = NULL;
	++cl->used;
}

/* coordinate_list_merge: Merge @mergelist into @cl
 */ 

void			 coordinate_list_merge(coordinate_list *cl, coordinate_list *mergelist)
{
	int i;

	assert(cl != NULL);
	assert(mergelist != NULL);

	for(i = 0; i < mergelist->used; i++)
		coordinate_list_add(cl, &mergelist->coords[i]);
}


/* coordinate_list_print: Print @cl to stdout, for debugging purposes
 */

void 			coordinate_list_print(coordinate_list *cl)
{
	int i;

	assert(cl != NULL);

	printf("[coordinate_list] Coordinate list %p size(%d) used(%d)\n", cl, cl->size, cl->used);

	for(i = 0; i < cl->used; i++) {
		printf("(%d,%d,%d), ", cl->coords[i].x, cl->coords[i].y, cl->coords[i].z);
	}

	printf("\n");	
}


/* coordinate_mtlist_create: Multi-threaded coordinate lists 
 * A multi-threaded coordinate list is a set of @nthreads coordinate lists,
 * one for each process, each of size @size. The lists may subsequently be merge into one large coordinate list.
 * Returns a pointer to the new mt-coordinate list
 */

mt_coordinate_list              *coordinate_mtlist_create(int size, int nthreads)
{
	int i;
	struct mt_coordinate_list *res = fw_malloc(sizeof(mt_coordinate_list));

	res->threads = nthreads;
	res->clists = fw_malloc(sizeof(struct coordinate_list *) * nthreads);
	

	for(i = 0; i < nthreads; i++) {
		res->clists[i] = coordinate_list_create(size);
	}

	return res;
}

/* coordinate_mtlist_add: Add coordinate @c to multi-threaded coordinate list @mtcl, where the calling
 * thread is @threadnum.
 */

void                            coordinate_mtlist_add(mt_coordinate_list *mtcl, coordinate *c, int threadnum)
{
	assert(mtcl != NULL);
	assert(c != NULL);
	assert(threadnum < mtcl->threads);

	coordinate_list_add(mtcl->clists[threadnum], c);
}

/* coordinate_mtlist_free: Free multi-threaded coordinate list @mtcl
 */

void                            coordinate_mtlist_free(mt_coordinate_list *mtcl)
{
	int i;

	assert(mtcl != NULL);

	for(i = 0; i < mtcl->threads; i++) {
		coordinate_list_free(mtcl->clists[i]);
	}

	fw_free(mtcl->clists);
	fw_free(mtcl);
}

/* coordinate_list_merge_mt: Merge the multi-threaded coordinate list @mergelist into a single
 * regular coordinate list @cl
 */

void                            coordinate_list_merge_mt(coordinate_list *cl, mt_coordinate_list *mergelist)
{
	int i;

	assert(mergelist != NULL); // cl != null checked in coordinate_list_merge

	for(i = 0; i < mergelist->threads; i++)
		coordinate_list_merge(cl, mergelist->clists[i]);

}


/* coordinate_to_rank: Find the rank which has c in its coordinate space
 */

int				coordinate_to_rank(coordinate c)
{
	int res;
	int cartnodecoords[2];

	extern int rank, singlemode;
	extern MPI_Comm cpfwcomm;


	/* Map face -> node number based on triangle's centroid.
	 */

	if(singlemode) {
		res = rank;
	} else {
		coordinate_to_cartrank(c, &cartnodecoords[0]);
		MPI_Cart_rank(cpfwcomm, &cartnodecoords[0], &res);
	}

	return res;
}

/* coordinate_to_cartrank: Find the Cartesian coordinates @cartranks of the rank which contains
 * @c
 */

void				coordinate_to_cartrank(coordinate c, int *cartranks)
{
	int res;
	int cartnodecoords[2];

	extern int n, m, k, singlemode, rank;
	extern int griddims[2];


	/* Map face -> node number based on triangle's centroid.
	 */

	if(singlemode) {
		cartranks[0] = cartranks[1] = 0;
	} else {
		cartranks[0] = c.x/(n/griddims[0]);
		cartranks[1] = c.y/(m/griddims[1]);

	}
}

/* coordinate_build_mpitype: Build the coordiante MPI type
 */

void				coordinate_build_mpitype(MPI_Datatype *dt)
{
	coordinate c;
	MPI_Datatype types[3] = { MPI_INT, MPI_INT, MPI_INT };
	int blocklengths[3] = { 1, 1, 1 };
	MPI_Aint displacements[3];

	MPI_Aint startaddr, tmpaddr;

	displacements[0] = 0;
	MPI_Address(&c.x, &startaddr);
	MPI_Address(&c.y, &tmpaddr);
	displacements[1] = tmpaddr - startaddr;
	MPI_Address(&c.z, &tmpaddr);
	displacements[2] = tmpaddr - startaddr;
	

	MPI_Type_struct(3, blocklengths, displacements, types, dt);

	MPI_Type_commit(dt);

}

/* coordinate_cart_get{left,right,upper,lower}neighbour: 
 * These functions retrieve the left/right/upper/lower neighbours of the node with rank 'rank' in 2D space. 
 *
 * Consider the following figure:
 *
 *        ______________
 *       |____|__2_|____|
 *       |__1_|__X_|__3_|
 *       |____|__4_|____|
 * 
 * Then calling coordinate_cart_get{left,upper,right,lower}neighbour(X) returns, respectively, 1, 2, 3, 4.
 * If no neighbors exist a negative number is returned.
 *
 */      

int				coordinate_cart_getleftneighbor(int noderank)
{

	coordinate c;

	extern int n, m, k;
	extern int singlemode;
	extern int griddims[2];
	extern native_voxel_space localvxspace;
	extern MPI_Comm cpfwcomm;

	int ncoords[2];


	assert(!singlemode);

	MPI_Cart_coords(cpfwcomm, noderank, 2, &ncoords[0]);

	if(ncoords[0] == 0)
		return -1; // No neighbour


	c.x = (ncoords[0] - 1) * (n/griddims[0]);
	c.y = (ncoords[1]) * (m/griddims[1]);;
	c.z = 0;	

	return coordinate_to_rank(c);
	
}

// See comment above

int				coordinate_cart_getrightneighbor(int noderank)
{

	coordinate c;

	extern int n, m, k;
	extern int singlemode;
	extern int griddims[2];
	extern native_voxel_space localvxspace;
	extern MPI_Comm cpfwcomm;

	int ncoords[2];

	assert(!singlemode);

	MPI_Cart_coords(cpfwcomm, noderank, 2, &ncoords[0]);

	if(ncoords[0] == griddims[0] - 1)
		return -1; // No neighbour


	c.x = (ncoords[0] + 1) * (n/griddims[0]);
	c.y = (ncoords[1]) * (m/griddims[1]);;
	c.z = 0;

	return coordinate_to_rank(c);
}

// See comment above

int				coordinate_cart_getupperneighbor(int noderank)
{

	coordinate c;

	extern int n, m, k;
	extern int singlemode;
	extern int griddims[2];
	extern native_voxel_space localvxspace;
	extern MPI_Comm cpfwcomm;

	int ncoords[2];

	assert(!singlemode);

	MPI_Cart_coords(cpfwcomm, noderank, 2, &ncoords[0]);

	if(ncoords[1] == griddims[1] - 1)
		return -1; // No neighbour


	c.x = (ncoords[0]) * (n/griddims[0]);
	c.y = (ncoords[1] + 1) * (m/griddims[1]);;
	c.z = 0;	

	return coordinate_to_rank(c);
}


// See comment above

int				coordinate_cart_getlowerneighbor(int noderank)
{

	coordinate c;

	extern int n, m, k;
	extern int singlemode;
	extern int griddims[2];
	extern native_voxel_space localvxspace;
	extern MPI_Comm cpfwcomm;

	int ncoords[2];

	assert(!singlemode);

	MPI_Cart_coords(cpfwcomm, noderank, 2, &ncoords[0]);

	if(ncoords[1] == 0)
		return -1; // No neighbour


	c.x = (ncoords[0]) * (n/griddims[0]);
	c.y = (ncoords[1] - 1) * (m/griddims[1]);;
	c.z = 0;	

	return coordinate_to_rank(c);
}
