/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_coord.h: Coordinates structures. See fw_coord.c for extensive comment.
 */

#ifndef FW_COORD
#define FW_COORD

#include <math.h>
#include <mpi.h>

#define CSLOT_UNUSED	0
#define CSLOT_USED	1

typedef struct coordinate {
	int x, y, z;
} coordinate;

/* Coordinate hash entry. Each coordinate list consists of a set of coordinates and
 * a set of coordinate_hash_entries:
 * 
 * coordinate_list:   coords      [ coordinate0 coordinate1 ... cN ]
 *                                  |---^-----------^       |----^
 *                    coordhash   [ 0   1   2 --------------|      ]
 *
 * The purpose of the coordhash is to quickly detect whether or not a coordinate is contained
 * inside the coordinate list (O(1) vs. O(n)).
 */

typedef struct coordinate_hash_entry {
	char used;
	int index;
	struct coordinate_hash_entry *next;
} coordinate_hash_entry;

typedef struct coordinate_list {
	coordinate *coords;
	coordinate_hash_entry *coordhash;
	int size;
	int used;
} coordinate_list;

typedef struct mt_coordinate_list {
	int threads;
	coordinate_list **clists;
} mt_coordinate_list;



// Find min/max x/y/z components of vertex coordinates a,b,c

#define Vmin(a,b,c,f)	( (a).f < (b).f ? ((c).f < (a).f ? (c).f : (a).f) : ((c).f < (b).f ? (c).f : (b).f) )
#define Vmax(a,b,c,f)	( (a).f > (b).f ? ((c).f > (a).f ? (c).f : (a).f) : ((c).f > (b).f ? (c).f : (b).f) )



coordinate_list			*coordinate_list_create(int size);
void			 	coordinate_list_free(coordinate_list *cl);
int  		                coordinate_list_lookup(coordinate_list *cl, coordinate *c);
void			 	coordinate_list_add(coordinate_list *cl, coordinate *c);
void                     	coordinate_list_print(coordinate_list *cl);
void		 		coordinate_list_merge(coordinate_list *cl, coordinate_list *mergelist);
void		 		coordinate_list_merge_mt(coordinate_list *cl, mt_coordinate_list *mergelist);
int				coordinate_to_rank(coordinate c);
void				coordinate_to_cartrank(coordinate c, int *cartranks);
int				coordinate_cart_getleftneighbor(int rank);
int				coordinate_cart_getrightneighbor(int rank);
int				coordinate_cart_getupperneighbor(int rank);
int				coordinate_cart_getlowerneighbor(int rank);


mt_coordinate_list		*coordinate_mtlist_create(int size, int nthreads);
void				coordinate_mtlist_add(mt_coordinate_list *mtcl, coordinate *c, int threadnum);
void				coordinate_mtlist_free(mt_coordinate_list *mtcl);

void				coordinate_build_mpitype(MPI_Datatype *dt);

static inline coordinate	coordinate_create(int x, int y, int z)
{
	struct coordinate res = {x, y, z};
	return res;
}


static inline int		coordinate_length(coordinate c)
{
	return (int) ceil(sqrt(c.x * c.x + c.y * c.y + c.z * c.z));
}


static inline int		coordinate_equal(coordinate c1, coordinate c2)
{
	return c1.x == c2.x && c1.y == c2.y && c1.z == c2.z;
}


static inline coordinate	coordinate_add(coordinate c1, coordinate c2)
{
	coordinate res;
	
	res.x = c1.x + c2.x;
	res.y = c1.y + c2.y;
	res.z = c1.z + c2.z;

	return res;
}

static inline coordinate	coordinate_sub(coordinate c1, coordinate c2)
{
	coordinate res;
	
	res.x = c1.x - c2.x;
	res.y = c1.y - c2.y;
	res.z = c1.z - c2.z;

	return res;
}


static inline coordinate	coordinate_abs(coordinate c)
{
	coordinate res;
	
	res.x = c.x > 0 ? c.x : -c.x;
	res.y = c.y > 0 ? c.y : -c.y;
	res.z = c.z > 0 ? c.z : -c.z;

	return res;
}



static inline coordinate	coordinate_scale(coordinate c, int k)
{
	coordinate res;
	
	res.x = c.x * k;
	res.y = c.y * k;
	res.z = c.z * k;

	return res;
}



static inline coordinate	coordinate_scale_float(coordinate c, double k)
{
	coordinate res;
	
	res.x = (int) round(c.x * k);
	res.y = (int) round(c.y * k);
	res.z = (int) round(c.z * k);

	return res;
}

#endif
