/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 */

/* fw_extract_cpu.c: Provides voxelization routines. Although this file is called '_cpu', the entry
 * point for all voxelization calls is still the extract_points() function, which uses the auxiliary
 * extract_cpu_3d_triangle (for CPU voxelization) and routines in fw_extract_gpu.c for GPU voxelization.
 */

#include <sys/mman.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#ifdef VOXELTEST
#include <sys/time.h>
#include <sys/resource.h>
#endif

#include "fw_coord.h"
#include "fw_vector.h"
#include "fw_wingededge.h"
#include "fw_extract_cpu.h"
#include "fw_extract_gpu.h"
#include "fw_main.h"


// Prototypes
static coordinate_list  *extract_cpu_3d_triangle(vector v1, vector Tc_1, vector Tc_2, vector Tc_3, quaternion *q1, coordinate_list *res);
static coordinate_list *extract_bresenham_2d(coordinate start, coordinate end, coordinate_list *res);
static coordinate_list *extract_bresenham_3d(coordinate start, coordinate end, coordinate_list *res);

// References # of threads/gpumode variables in fw_main.c
extern int threadcount, gpumode;

/* extract_cpu_init: Initialize OpenMP threads. Threads are created in this routine
 * so that (expensive) thread-creation does not happen when doing the actual voxelization.
 */

void extract_cpu_init(void)
{
	int tmp = 0;

        omp_set_dynamic(0);
        omp_set_num_threads(threadcount);
#pragma omp parallel shared(tmp)
        {
#pragma omp atomic
                ++tmp;
        }

	if(tmp != threadcount) {
		printf("[extract_init] only created %d of %d threads, aborting.\n", tmp, threadcount);
		abort();
	}
}

/* extract_points: Voxelize the triangle spanned by v1, v2 and v3, returning a coordinate list
 * of points intersecting the triangle.
 */

struct coordinate_list *extract_points(struct vertex *v1, struct vertex *v2, struct vertex *v3)
{
	// Vector quantities from v1->v2, v1->v3, etc.

	vector v1_v = vector_create(&v1->coords), v2_v = vector_create(&v2->coords), v3_v = vector_create(&v3->coords);
	vector v1_v2 = vector_sub(v2_v, v1_v);
	vector v1_v3 = vector_sub(v3_v, v1_v);
	vector v2_v3 = vector_sub(v3_v, v2_v);

	// Lengths of above vectors

	double v1_v2_length = vector_length(v1_v2), 
		v1_v3_length = vector_length(v1_v3), 
		v2_v3_length = vector_length(v2_v3);

	// Check if any coordinates are equal. If so, the triangle is degenerate.
	int v1v2_eq = vector_coordinate_equal(v1_v, v2_v), v1v3_eq = vector_coordinate_equal(v1_v, v3_v),
		v2v3_eq = vector_coordinate_equal(v2_v, v3_v);

	// Temporary vectors used below.
	vector tracev1, tracev2, tracev3;

	vector Tc_1, Tc_2, Tc_3; // Transformed vertices
	quaternion q1;		 // Inverse rotation quaternion


	coordinate_list *res;	// Result
	double area;		// Area of triangle


	/* First, we check if the triangle is degenerate: 
	 *   1) is v1 = v2 = v3? if so, the triangle is a single voxel
	 *   2) are two vertices equal so that the entire triangle is a single line? if so, we can use Bresenham's 3D line drawing algorithm for the
	 *      entire voxelization.
	 */

	if(v1v2_eq || v1v3_eq || v2v3_eq) {
		if(v1v2_eq && v1v3_eq && v2v3_eq) {
			res = coordinate_list_create(1);
			coordinate_list_add(res, &v1->coords);
			return res;
		} else if(v1v2_eq) {
			res = coordinate_list_create(ceil(v1_v3_length) + 2);
			extract_bresenham_3d(v1->coords, v3->coords, res);
			return res;
		} else if(v1v3_eq) {
			res = coordinate_list_create(ceil(v2_v3_length) + 2);
			extract_bresenham_3d(v2->coords, v3->coords, res);
			return res;
		} else { // v2v3_eq must be true
			res = coordinate_list_create(ceil(v1_v2_length) + 2);
			extract_bresenham_3d(v1->coords, v2->coords, res);
			return res;
		}
	}

	/* 3) Are the vertices colinear? If so, we can also apply Bresenham's 3D line drawing algorithm directly
	 */


// The following two macros take the min/max of a,b,c and returns, respectively, ar,br or cr

#define VVmin(ar,br,cr,a,b,c)	( (a) < (b) ? ((c) < (a) ? (cr) : (ar)) : ((c) < (b) ? (cr) : (br)) )
#define VVmax(ar,br,cr,a,b,c)	( (a) > (b) ? ((c) > (a) ? (cr) : (ar)) : ((c) > (b) ? (cr) : (br)) )

	if(vector_parallel(v1_v2, v2_v3) && vector_parallel(v1_v3, v2_v3) && vector_parallel(v1_v2, v1_v3)) {

		// Select starting/ending point of triangle by choosing the vertex with the lowest/highest distance from the origin.

		vector minYv = VVmin(v1_v, v2_v, v3_v, vector_length(v1_v), vector_length(v2_v), vector_length(v3_v));
		vector maxYv = VVmax(v1_v, v2_v, v3_v, vector_length(v1_v), vector_length(v2_v), vector_length(v3_v));

		res = coordinate_list_create( ceil(vector_length(vector_sub(maxYv, minYv))) + 2 );
		extract_bresenham_3d(vector_to_coordinate(vector_round(minYv)), vector_to_coordinate(vector_round(maxYv)), res);

		return res;
	}

	/* First compute area of polygon to estimate the size of the coordinate list.
	 * We slightly over-allocate since we are using discrete pixels instead of continuous mathematical lines.
	 */
	
	area = ceil(we_triangle_area(v1, v2, v3)) + ceil(v1_v2_length) + ceil(v1_v3_length) + ceil(v2_v3_length);

	/* Create resulting coordinate list 
	 */
	
	res = coordinate_list_create( (int) ceil(area) );

	/* The following if-statements execute such that (tracev1,tracev3) is the longest edge 
	 * of the triangle.
	 */

	if(v1_v2_length <= v1_v3_length && v1_v2_length <= v2_v3_length) {

		/* v1_v2 is shortest edge; trace1/2/3 are determined by the longest
		 * edge of v1_v3 and v2_v3
		 */

		if(v1_v3_length < v2_v3_length) {
			tracev1 = v2_v;
			tracev2 = v1_v;
			tracev3 = v3_v;
		} else {
			tracev1 = v1_v;
			tracev2 = v2_v;
			tracev3 = v3_v;
		}
	} else if(v1_v3_length <= v1_v2_length && v1_v3_length <= v2_v3_length) {

		/* v1_v3 is shortest edge; trace1/2/3 are determined by the longest
		 * edge of v1_v2 and v2_v3
		 */

		if(v1_v2_length < v2_v3_length) {
			tracev1 = v3_v;
			tracev2 = v1_v;
			tracev3 = v2_v;
		} else {
			tracev1 = v1_v;
			tracev2 = v3_v;
			tracev3 = v2_v;
		}
	} else {
		/* v2_v3 is shortest edge; trace1/2/3 are determined by the longest
		 * edge of v1_v2 and v1_v3
		 */

		if(v1_v2_length < v1_v3_length) {
			tracev1 = v3_v;
			tracev2 = v2_v;
			tracev3 = v1_v;
		} else {
			tracev1 = v2_v;
			tracev2 = v3_v;
			tracev3 = v1_v;
		}
	}

	/* Rotate triangle to z=0 plane, returning transformed vertices in Tc_1, Tc_2 and Tc_3.
	 */

	extract_rotate_triangle(tracev1, tracev2, tracev3, &Tc_1, &Tc_2, &Tc_3, &q1);

	// If not benchmarking, try GPU first and then CPU

#ifndef VOXELTEST
	
	/* extract_gpu_3d_triangle returns NULL if not using GPU or if triangle is too large
	 */

# ifndef GPU_DISABLE
	if(!extract_gpu_3d_triangle(tracev1, Tc_1, Tc_2, Tc_3, &q1, res))
# endif
		extract_cpu_3d_triangle(tracev1, Tc_1, Tc_2, Tc_3, &q1, res);
#else
	// If benchmarking, extract_gpu_3d_triangle will abort if a triangle > max extents
	// is passed so that benchmark results only reflect GPU performance.

# ifndef GPU_DISABLE
	if(gpumode)
		extract_gpu_3d_triangle(tracev1, Tc_1, Tc_2, Tc_3, &q1, res);
	else
		extract_cpu_3d_triangle(tracev1, Tc_1, Tc_2, Tc_3, &q1, res);
# else
		extract_cpu_3d_triangle(tracev1, Tc_1, Tc_2, Tc_3, &q1, res);
# endif
		

#endif	

	return res;
}

/* extract_bresenham_2d: Bresenham's 2D line drawing algorithm.
 * This function rasterizes a line starting at @start and ending at @end. The coordinate list,
 * which has to be allocated in advance, will contain the line coordinates
 * and be returned after execution. See also thesis Appendix B. 
 *
 * This routine is used in extract_cpu_3d_triangle
 *
 * This implementation was inspired by the implementations at http://www.cit.gu.edu.au/~anthony/info/graphics/bresenham.procs
 * (original author unknown) and D. Heran and M. P. Baker: "Computer Graphics with OpenGL", 3rd ed.
 */

static coordinate_list *extract_bresenham_2d(coordinate start, coordinate end, coordinate_list *res)
{
	int i, e1;
	coordinate lmn, d, d2, inc, coord1;
	
	d = coordinate_sub(end, start);		// deltas in x,y directions
	lmn = coordinate_abs(d); 		// absolute deltas in x,y directions
	d2 = coordinate_scale(lmn, 2);		// 2x dx/dy, for convenience
	
	inc.x = (d.x < 0) ? -1 : 1; 		// d.x < 0 => increment = -1 in x-direction, if not we increment in steps of 1 
	inc.y = (d.y < 0) ? -1 : 1;		// similar to inc.x for y-direction


	// If lmn.x > lmn.y, we must step in x-direction.
	// If not, step in the y-direction since the y-direction contains most pixels.

	// See thesis Appendix B for more information on the code below.

	if(lmn.x > lmn.y) {
		// Initial error rho_0
		e1 = d2.y - lmn.x;
		for(i = 0; i < lmn.x + 1; i++) {
			coordinate_list_add(res, &start);

			// If e1>0 (rho_k) increase y-coordinate

			if(e1 > 0) {
				start.y += inc.y;
				e1 -= d2.x;
			}


			// Incrementally compute rho_k+1
			e1 += d2.y;
			start.x += inc.x;
		}
	} else {
		// Initial error rho_0
		e1 = d2.x - lmn.y;
		for(i = 0; i < lmn.y + 1; i++) {
			coordinate_list_add(res, &start);

			// If e1>0 (rho_k) increase x-coordinate

			if(e1 > 0) {
				start.x += inc.x;
				e1 -= d2.y;
			}

			// Incrementally compute rho_k+1
			e1 += d2.x;
			start.y += inc.y;			
		}
	}

	return res;
}

/* extract_bresenham_2d: Bresenham's 3D line drawing algorithm.
 * This function voxelizes a line starting at @start and ending at @end. The coordinate list,
 * which has to be allocated in advance, will contain the line coordinates
 * and be returned after execution. See also thesis Appendix B. 
 *
 * This routine is used in extract_points in the case that the triangle degenerates to a line.
 *
 * This implementation was inspired by the implementations at http://www.cit.gu.edu.au/~anthony/info/graphics/bresenham.procs
 * (original author unknown) and D. Heran and M. P. Baker: "Computer Graphics with OpenGL", 3rd ed.
 */


static coordinate_list *extract_bresenham_3d(coordinate start, coordinate end, coordinate_list *res)
{
	int i, e1, e2;
	coordinate lmn, d, d2, inc, coord1;
	
	d = coordinate_sub(end, start);		// deltas in x,y,z directions
	lmn = coordinate_abs(d); 		// absolute deltas in x,y directions
	d2 = coordinate_scale(lmn, 2);		// 2x deltas in x,y,z directions
	
	inc.x = (d.x < 0) ? -1 : 1; 		// d.x < 0 => increment = -1 in x-direction, if not we increment in steps of 1 
	inc.y = (d.y < 0) ? -1 : 1;		// similar to inc.x for y-direction
	inc.z = (d.z < 0) ? -1 : 1;		// similar to inc.x for z-direction


	// The following if-statements ensure that we increment
	// along the axis where the line has the longest extent
	// See also thesis Appendix B

	if(lmn.x > lmn.y && lmn.x > lmn.z) {
		// e1,e2 are initial erros in y,z directions
		e1 = d2.y - lmn.x;
		e2 = d2.z - lmn.x;
		for(i = 0; i < lmn.x + 1; i++) {
			coordinate_list_add(res, &start);
			// e1, e2 correspond to rho_{k,y} and rho_{k,z}
			if(e1 > 0) {
				start.y += inc.y;
				e1 -= d2.x;
			}
			if(e2 > 0) {
				start.z += inc.z;
				e2 -= d2.x;
			}

			e1 += d2.y;
			e2 += d2.z;

			start.x += inc.x;			
		}
	} else if(lmn.y >= lmn.x && lmn.y >= lmn.z) {
		// e1,e2 are initial erros in x,z directions

		e1 = d2.x - lmn.y;
		e2 = d2.z - lmn.y;
		for(i = 0; i < lmn.y + 1; i++) {
			coordinate_list_add(res, &start);
			// e1, e2 correspond to rho_{k,x} and rho_{k,z}

			if(e1 > 0) {
				start.x += inc.x;
				e1 -= d2.y;
			}
			if(e2 > 0) {
				start.z += inc.z;
				e2 -= d2.y;
			}

			e1 += d2.x;
			e2 += d2.z;

			start.y += inc.y;			
		}
	} else {
		// e1,e2 are initial erros in x,y directions

		e1 = d2.y - lmn.z;
		e2 = d2.x - lmn.z;
		for(i = 0; i < lmn.z + 1; i++) {
			coordinate_list_add(res, &start);
			// e1, e2 correspond to rho_{k,y} and rho_{k,x}

			if(e1 > 0) {
				start.y += inc.y;
				e1 -= d2.z;
			}
			if(e2 > 0) {
				start.x += inc.x;
				e2 -= d2.z;
			}

			e1 += d2.y;
			e2 += d2.x;

			start.z += inc.z;
		}
	}
	
	return res;
}

/* extract_rotate_triangle: Transform a 3D triangle to the z=0 plane. Approach (see also thesis Chapter 4):
 * Given a triangle, (v1, v3) is the longest vertex, (v1, v2) 2nd longest and (v2, v3) shortest.
 *   1. Translate v1 to the origin.
 *   2. Rotate (v1, v3) to z=0 plane. (alpha)
 *   3. Rotate so that (v1, v3) is aligned with the x-axis of the z = 0 plane. (beta)
 *   4. Rotate the final vertex, v2, into the z = 0 plane. (gamma)
 *   5. Now the triangle is in a single plane and the longest vertex is located along the x-axis.
 *      Find voxels intersected by the edges by using Bresenham's line-drawing algorithm.
 *   6. For x = 0 to the v3 vertex's x coordinate:
 *         Let y = 0. While !voxel belongs to some line of the triangle:
 *                       Add (x,y,0) to the result set.
 *                       Increase/decrease y based on v2's y being > 0 or < 0.
 *   7. Apply inverse 4,3,2,1 to each voxel found.
 *
 * Steps 1-4 is done in this routine. Steps 5-7 are done in extract_cpu_3d_triangle (for the CPU)
 * or extract_gpu_3d_triangle in fw_extract_gpu.c (for the GPU).
 *
 * Rotation happens using quaternions. See fw_vector.h for details. In step 7, inverse quaternions
 * are combined to form a single quaternion -- which immediately cuts computational complexity by 66%,
 * at the price of a small one-time cost.
 *
 * This routine takes the vertices @v1, @v2, @v3 of the triangle as input, and returns an inverse quaternion
 * in @ivquat in addition to transformed vertices in @tv1, @tv2, @tv3.
 *
 * It is assumed that (v1,v3) is the longest vertex of the triangle. This is ensured in extract_points.
 */

void		extract_rotate_triangle(vector v1, vector v2, vector v3, vector *tv1, vector *tv2, vector *tv3, quaternion *ivquat)
{
	/* temporary vectors/angle values used when rotating/translating */

	vector tmpvec1, tmpvec2;
	double tmpang; 
	quaternion q1, q2, q3;

	// -- Step 1 - translate coords
	vector Tc_1 = vector_sub(v1, v1), Tc_2 = vector_sub(v2, v1),
			Tc_3 = vector_sub(v3, v1);


	// -- Step 2 - project (v1, v3) to xy-plane, find angle and rotate about perpendicular vector.	
	// After this step, (v1, v3) is in the plane z=0.

	// If Tc_3.x and Tc_3.y are zero, we have a special case with rotation about the positive y-axis Pi/2 radians.

	if(Tc_3.x != 0 || Tc_3.y != 0) {
		tmpvec1 = vector_create_components(Tc_3.x, Tc_3.y, 0.0);
		tmpvec2 = vector_create_components(-Tc_3.y, Tc_3.x, 0.0); // tmpvec2 perp tmpvec1
	} else {
		tmpvec1 = vector_create_components(0.0, 0.0, Tc_3.z);
		tmpvec2 = VECTOR_UNIT_Y;
	}

	tmpang = vector_angle(Tc_3, tmpvec1);
	q1 = quaternion_create(tmpvec2, Tc_3.z > 0 ? tmpang : -tmpang);
	

	// Rotate Tc_2 and Tc_3. Since Tc_1 is already at the origin, rotation is unnecessary.

	Tc_2 = vector_rotate(Tc_2, &q1);
	Tc_3 = vector_rotate(Tc_3, &q1);


	// -- Step 3 - rotate (v1,v3) to align with the x-axis.

	tmpang = vector_angle(Tc_3, VECTOR_UNIT_X);
	q2 = quaternion_create(VECTOR_UNIT_Z, Tc_3.y < 0 ? tmpang : -tmpang);


	// Rotate Tc_2 and Tc_3. Rotating Tc_1 is, again, unnecessary.

	Tc_2 = vector_rotate(Tc_2, &q2);
	Tc_3 = vector_rotate(Tc_3, &q2);	

	// -- Step 4 - rotate Tc_2 to lie in z = 0 plane.

	tmpvec1 = vector_create_components(0.0, Tc_2.y, 0.0);
	tmpvec2 = vector_create_components(0.0, Tc_2.y, Tc_2.z);
	tmpang = vector_angle(tmpvec1, tmpvec2);

	if((Tc_2.z >= 0 && Tc_2.y >= 0) || (Tc_2.z < 0 && Tc_2.y < 0)) {
		tmpang = -tmpang;
	}

	q3 = quaternion_create(VECTOR_UNIT_X, tmpang);

	// Rotate Tc_2 - Tc_1 and Tc_3 are already in z=0 plane

	Tc_2 = vector_rotate(Tc_2, &q3);

	*tv1 = Tc_1;
	*tv2 = Tc_2;
	*tv3 = Tc_3;

	quaternion_invert(&q1);
	quaternion_invert(&q2);
	quaternion_invert(&q3);
	q2 = quaternion_combine(&q2, &q3);
	q1 = quaternion_combine(&q1, &q2);

	*ivquat = q1;
}

/* extract_cpu_3d_triangle: See comment above extract_rotate_triangle. This routine performs steps 5-7. We take in
 * three vertices @Tc_1, @Tc_2, @Tc_3 in z=0, and rasterize as described in thesis Chapter 4. @q1 is the inverse quaternion,
 * @v1 is the initial translation and @res is the output list of coordinates.
 */

static coordinate_list  *extract_cpu_3d_triangle(vector v1, vector Tc_1, vector Tc_2, vector Tc_3, quaternion *q1, coordinate_list *res)
{
	
	// -- Step 5 - extract voxels of the triangle's lines


	vector r_1 = vector_round(Tc_1), r_2 = vector_round(Tc_2), r_3 = vector_round(Tc_3);
	coordinate_list *l1, *l2, *lm;
	int *lc;
	int lidx, i, j, lcy, lcx;

	l1 = coordinate_list_create(vector_length(vector_sub(r_2, r_1)) + 2);
	l2 = coordinate_list_create(vector_length(vector_sub(r_3, r_2)) + 2);
	lc = fw_malloc(sizeof(int) * ((int) r_3.x + 1));

	// These will be zero anyway, but we make sure that some obscure rounding error doesn't ruin everything
	r_1.z = r_2.z = r_3.z = 0;

	extract_bresenham_2d(vector_to_coordinate(r_1), vector_to_coordinate(r_2), l1);
	extract_bresenham_2d(vector_to_coordinate(r_2), vector_to_coordinate(r_3), l2);
	lm = coordinate_list_create(l1->used + l2->used);
	coordinate_list_merge(lm, l1);
	coordinate_list_merge(lm, l2);


	/* Make lc array (see thesis Chap. 4). We do this by passing over a merged list of the
	 * coordinates in l1 and l2 and always picking the coordinate with the minimum y (if v2's y is positive)
	 * or the maximum y (if negative).
	 */

	for(i = 0, lidx = 0; i <= (int) r_3.x; i++) {
		lcy = lm->coords[lidx].y;
		lcx = lm->coords[lidx++].x;

		if(r_2.y > 0) {
			while(lidx < l1->used + l2->used && lm->coords[lidx].x == lcx) {
				if(lm->coords[lidx].y < lcy)
					lcy = lm->coords[lidx].y;
				++lidx;
			} 
		} else {
			while(lidx < l1->used + l2->used && lm->coords[lidx].x == lcx) {
				if(lm->coords[lidx].y > lcy)
					lcy = lm->coords[lidx].y;
				++lidx;
			} 
		}

		lc[i] = lcy;
	}

	coordinate_list_free(lm);
	coordinate_list_free(l1);
	coordinate_list_free(l2);


	// -- Steps 6/7 - extract actual voxels & inverse rotation/translation

	const int minmaxY = (int) r_2.y;
	coordinate tmpc, tmpc2;
	int x, y;

	/* When multi-threading the algorithm, each thread will write to one coordinate list.
	 * The list will subsequently be merged on the master thread.
	 * This is done to avoid synchronization in the inner loops.
	 */

	mt_coordinate_list *mtcl = coordinate_mtlist_create( res->size, threadcount );

	mlockall(MCL_CURRENT|MCL_FUTURE);

	if(minmaxY > 0) {
#pragma omp parallel for private(tmpc, tmpc2)
		for(x = 0; x <= (int) r_3.x; x++) {
			tmpc.z = 0;
			for(tmpc.y = 0; tmpc.y <= lc[x]; tmpc.y++) {

			/* If running the benchmarking program,
			 * we do not hash coordinate values since this makes
			 * the time reflect the hashing time and not extraction time.
			 * We also use 32-bit in test mode since the GPU also uses 32 bits.
			 * Also note that the vector_* functions are inlined.
			 */
				tmpc.x = x;
#ifndef VOXELTEST
				tmpc2 = vector_to_coordinate(vector_round(vector_add(vector_rotate(vector_create(&tmpc), q1), v1)));
				coordinate_mtlist_add(mtcl, &tmpc2, omp_get_thread_num());
#else
				tmpc2 = vector_to_coordinate(vector_round(vector_addf(vector_rotatef(vector_create(&tmpc), q1), v1)));
				mtcl->clists[omp_get_thread_num()]->coords[mtcl->clists[omp_get_thread_num()]->used++] = tmpc2;
#endif
			}
		}
	} else {
#pragma omp parallel for private(tmpc, tmpc2)
		for(x = 0; x <= (int) r_3.x; x++) {
			tmpc.z = 0;
			for(tmpc.y = 0; tmpc.y >= lc[x]; tmpc.y--) {
				tmpc.x = x;
#ifndef VOXELTEST
				tmpc2 = vector_to_coordinate(vector_round(vector_add(vector_rotate(vector_create(&tmpc), q1), v1)));
				coordinate_mtlist_add(mtcl, &tmpc2, omp_get_thread_num());
#else
				tmpc2 = vector_to_coordinate(vector_round(vector_addf(vector_rotatef(vector_create(&tmpc), q1), v1)));
				mtcl->clists[omp_get_thread_num()]->coords[mtcl->clists[omp_get_thread_num()]->used++] = tmpc2;
#endif

			}
		}
	}


/* It's not fair to include the merge step in the test program since it includes hashing
 */

#ifndef VOXELTEST
	coordinate_list_merge_mt(res, mtcl);
#else

	// Use quicker merging for the test program.
	// The disadvantage of using this method is that lookups in coordinate lists are slow.

	for(i = 0; i < mtcl->threads; i++) {
		for(j = 0; j < mtcl->clists[i]->used; j++) {
			res->coords[res->used++] = mtcl->clists[i]->coords[j];
		}
	}

#endif
	coordinate_mtlist_free(mtcl);


	fw_free(lc);

	munlockall();


	return res;

}

/* The following are functions used only in the benchmarking program for voxelization
 */

#if defined(VOXELTEST) && !defined(GENSCRIPT)


/* usage: Print usage information, where the initial argv[0] is in @av0.
 */

void usage(const char *av0)
{
	printf("\nUsage: %s <gpu-extract|cpu-extract> <threads> <v1x,v1y,v1z> <v2x,v2y,v2z> <v3x,v3y,v3z> [onetriangle]\n", av0);
	printf("The program will create 2^9 = 512 triangles from the three vertices specified by trying all sign combinations\n");
	printf("of the 9 specified values. Vertices are specified as integers. 	If onetriangle is specified, 512 tests with a single triangle will be run.\n\n");
}

/* Entry point to voxelization benchmarking program.
 *  @argc argument count
 *  @argv argument vector
 */

int main(int argc, char **argv)
{
	int onetriangle = 0;
	unsigned int i;
	coordinate v1_orig, v2_orig, v3_orig;
	
	double walltime = 0.0, Ucputime = 0.0, Scputime = 0.0, Tcputime = 0.0;
	double avgarea = 0.0, maxarea = -1.0, minarea = 0xffffffff;

	printf("\nFramework for Polygonal Structures on Clusters\n");
	printf("Leif Christian Larsen <leifchl@idi.ntnu.no>\n\n");
	printf("VOXELIZATION BENCHMARKING PROGRAM.\n\n");

	if(argc != 6 && argc != 7) {
		usage(argv[0]);
		return -1;
	}

	threadcount = atoi(argv[2]);

	if(threadcount < 0) {
		printf("[voxeltest] Invalid thread count.\n");
		usage(argv[0]);
		return -1;
	}

	if(!strcmp(argv[1], "gpu-extract")) {
		extract_cpu_init();
#ifndef GPU_DISABLE
		extract_gpu_init(&argc, argv);
		gpumode = 1;
#else
		printf("[cpfw] ABORT: Program was compiled with GPU_DISABLE flag. GPU support disabled. Aborting\n");
		return -1;
#endif
	} else if(!strcmp(argv[1], "cpu-extract")) {
		extract_cpu_init();
		gpumode = 0;
	} else {
		printf("[voxeltest] Unknown extraction mode `%s'.\n", argv[1]);
		usage(argv[0]);
		return -1;
	}

	if(sscanf(argv[3], "%d,%d,%d", &v1_orig.x, &v1_orig.y, &v1_orig.z) != 3) {
		printf("[voxeltest] Invalid vertex: %s\n", argv[3]);
		usage(argv[0]);
		return -1;
	}

	if(sscanf(argv[4], "%d,%d,%d", &v2_orig.x, &v2_orig.y, &v2_orig.z) != 3) {
		printf("[voxeltest] Invalid vertex: %s\n", argv[4]);
		usage(argv[0]);
		return -1;
	}

	if(sscanf(argv[5], "%d,%d,%d", &v3_orig.x, &v3_orig.y, &v3_orig.z) != 3) {
		printf("[voxeltest] Invalid vertex: %s\n", argv[5]);
		usage(argv[0]);
		return -1;
	}

	if(argc == 7 && !strcmp(argv[6], "onetriangle")) {
		printf("[voxeltest] One-triangle mode. Running 512 tests with single triangle\n");
		onetriangle = 1;
	} else if(argc == 7) {
		printf("[voxeltest] Argument 6 must be onetriangle.\n");
		usage(argv[0]);
		return -1;
	}

	printf("Voxelizing triangles (. signifies start of triangle voxelization, : signifies finished): ");
	fflush(stdout);

	for(i = 0; i < 512; i++) {
		struct vertex v1 = {.coords = {v1_orig.x, v1_orig.y, v1_orig.z}}, v2 = {.coords = {v2_orig.x, v2_orig.y, v2_orig.z}}, v3 = {.coords = {v3_orig.x, v3_orig.y, v3_orig.z}};
		struct timeval start, stop;
		struct rusage rstart, rstop;

		coordinate_list *cl;

		if(!onetriangle) {
			if(i&(1<<0))
				v1.coords.x = -v1.coords.x;
			if(i&(1<<1))
				v1.coords.y = -v1.coords.y;
			if(i&(1<<2))
				v1.coords.z = -v1.coords.z;

			if(i&(1<<3))
				v2.coords.x = -v2.coords.x;
			if(i&(1<<4))
				v2.coords.y = -v2.coords.y;
			if(i&(1<<5))
				v2.coords.z = -v2.coords.z;

			if(i&(1<<6))
				v3.coords.x = -v3.coords.x;
			if(i&(1<<7))
				v3.coords.y = -v3.coords.y;
			if(i&(1<<8))
				v3.coords.z = -v3.coords.z;
		}

		double area = we_triangle_area(&v1, &v2, &v3);

		avgarea += area/512;
		if(area < minarea)
			minarea = area;
		if(area > maxarea)
			maxarea = area;

		printf(".");
		fflush(stdout);

		gettimeofday(&start, NULL);
		getrusage(RUSAGE_SELF, &rstart);
		cl = extract_points(&v1, &v2, &v3);
		gettimeofday(&stop, NULL);
		getrusage(RUSAGE_SELF, &rstop);

		printf(":");
		fflush(stdout);


		walltime += (stop.tv_sec - start.tv_sec) + (stop.tv_usec - start.tv_usec)/1000000.0;
		Ucputime += (rstop.ru_utime.tv_sec - rstart.ru_utime.tv_sec) + (rstop.ru_utime.tv_usec - rstart.ru_utime.tv_usec)/1000000.0;
		Scputime += (rstop.ru_stime.tv_sec - rstart.ru_stime.tv_sec) + (rstop.ru_stime.tv_usec - rstart.ru_stime.tv_usec)/1000000.0;
	
		if(cl)
			coordinate_list_free(cl);

	}

	Tcputime = Ucputime + Scputime;


	printf("\n\nStatistics\n");
	printf("==========\n");
	printf("Used %d threads, algorithm = %s\n", threadcount, argv[1]);
	printf("Total/per triangle wallclock time.........: %.5f/%.5f\n", walltime, walltime/512);
	printf("Total/per triangle CPU user time..........: %.5f/%.5f\n", Ucputime, Ucputime/512);
	printf("Total/per triangle CPU system time........: %.5f/%.5f\n", Scputime, Scputime/512);
	printf("Total/per triangle CPU user+sys time......: %.5f/%.5f\n", Tcputime, Tcputime/512);
	printf("Average/max/min triangle area ............: %.5f/%.5f/%.5f\n", avgarea, maxarea, minarea);


	return 0;	
}
#endif
