/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_extract_gpu.c: Provides voxelization on the GPU (however, this file contains only the
 * CPU code necessary for the GPU code to run.)
 *
 * Note that only triangles with extents < max GPU texture size may be extracted on the GPU.
 * Hence all triangles of greater size will still use the CPU version.
 */


#ifndef GPU_DISABLE

#include <sys/mman.h>
#include <stdio.h>
#ifdef USE_GLEW
#include <GL/glew.h>
#endif
#include <GL/glut.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glext.h>
#include <Cg/cgGL.h>



#include "fw_main.h"
#include "fw_coord.h"
#include "fw_extract_cpu.h"
#include "fw_extract_gpu.h"

// Prototypes for internal/debugging functions

static void extract_gpu_cgerror(void);
static void print_texture(float *data, int w, int h);


/* Cg variables */
static CGcontext cgc;							// Cg context
static CGprofile cgpr_f, cgpr_v;					// Fragment/vertex processor profiles
static CGprogram cgprog_V, cgprog_F, cgprog_F_clear;			// Passthrough vertex program, fragment voxelization program and fragment clear program
static CGparameter cgprog_V_param_mvp;					// Model-View-Projection matrix parameter to passthrough vertex program
static CGparameter cgprog_F_param_rotmtx, cgprog_F_param_minXY;		// Inverse rotation matrix and minXY translation parameters to fragment voxelization program


static int maxsize = 0;			// maximum texture size
static GLuint fb;			// framebuffer ID
static GLuint texture_id;		// texture ID

static int gpuinit = 0;			// has the GPU been initialized?

/* extract_gpu_init: Called once, when program starts. Initializes glut and OpenGL.
 * Original program calling arguments are passed in @argc and @argv, since they are needed
 * by glutInit.
 */

void extract_gpu_init(int *argc, char **argv)
{
	gpuinit++;

	/* ---- Initialize GLUT */

        glutInit(argc, argv);
        glutCreateWindow("cpfw"); /* This doesnt actually create a window, just initializes glut */
#ifdef USE_GLEW
	glewInit(); /* initialize GLEW (opengl extension library) */
#endif	



	/* ---- Get max texture size */

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxsize);
	printf("[extract-gpu] Maximum texture size is: %dx%d.\n", maxsize, maxsize);
	maxsize--; /* Avoid "last texture address not available" bug on some NVIDIA GeForce 7 series GPUs */

	/* ---- OpenGL initialization */

	/* Map pixels and texels 1:1
	 */

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, maxsize, 0.0, maxsize, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0.0, 0.0, maxsize, maxsize);

	/* Set up off-screen rendering
	 */

	glGenFramebuffersEXT( 1, &fb );



	/* ---- Initialize Cg */


	/* On error, dct_gpu_cgError will be called. Thus we do not need
	 * to check return values from cg functions.
	 */

	cgSetErrorCallback(extract_gpu_cgerror);

	/* Create context, profiles and load programs
	 */

	cgc = cgCreateContext();
	cgpr_f = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	cgpr_v = cgGLGetLatestProfile(CG_GL_VERTEX);



	if(cgpr_f == CG_PROFILE_UNKNOWN) {
		printf("[extract_gpu] Error: Unknown fragment profile. Do you really have a GPU installed?\n");
		abort();
	}

	if(cgpr_v == CG_PROFILE_UNKNOWN) {
		printf("[extract_gpu] Error: Unknown vertex profile. Do you really have a GPU installed?\n");
		abort();
	}


	cgGLSetOptimalOptions(cgpr_f);
	cgGLSetOptimalOptions(cgpr_v);

	cgprog_V = cgCreateProgramFromFile(cgc, CG_SOURCE, "fw_extract_gpu.cg", cgpr_v, "extract_vp", NULL);
	cgprog_F = cgCreateProgramFromFile(cgc, CG_SOURCE, "fw_extract_gpu.cg", cgpr_f, "extract_fp", NULL);
	cgprog_F_clear = cgCreateProgramFromFile(cgc, CG_SOURCE, "fw_extract_gpu.cg", cgpr_f, "extract_fp_clear", NULL);


	cgGLLoadProgram(cgprog_V);
	cgGLLoadProgram(cgprog_F);
	cgGLLoadProgram(cgprog_F_clear);


	/* MVP matrix parameter - constant */
	cgprog_V_param_mvp = cgGetNamedParameter(cgprog_V, "mvp");
	cgprog_F_param_rotmtx = cgGetNamedParameter(cgprog_F, "rotmtx");
	cgprog_F_param_minXY = cgGetNamedParameter(cgprog_F, "minXY");



	/* Ready to go? */

	if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT ) {
		printf("[extract_gpu] Framebuffer: Error: Not complete. Check that card supports FBOs. Exit\n");
		abort();
	}

	cgGLEnableProfile(cgpr_f);
	cgGLEnableProfile(cgpr_v);

	cgGLSetManageTextureParameters(cgc, CG_TRUE); /* automatically calls cgGLEnableTextureParameter when program is bound */

	cgGLSetStateMatrixParameter(cgprog_V_param_mvp, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fb );
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);


	glFinish(); /* synchronize OpenGL */
}

/* extract_gpu_3d_triangle: Voxelize 3D triangle on the GPU. @Tc_1, @Tc_2 and @Tc_3 are vertices transformed to z=0 plane.
 * @q1 is inverse rotation quaternion, and @v1 is initial translation. Coordinates are inserted into the @res coordinate list,
 * which is also returned by the function.
 */

coordinate_list  *extract_gpu_3d_triangle(vector v1, vector Tc_1, vector Tc_2, vector Tc_3, quaternion *q1, coordinate_list *res)
{
	int transX, transY, maxX, maxY;
	float *gpureadbuf;
	int gpureadbufsize;
	int i;
	coordinate tmpc;

	/* Immediately discard: No GPU or triangle area > max texture area 
	 * CPU algorithm will be used.
	 */

	if(!gpuinit || res->size > maxsize * maxsize)
#ifndef VOXELTEST
		return NULL;
#else
		{
			printf("[cpfw] ABORTING GPU BENCHMARK: Triangle area too large.\n");
			abort(); // should not happen when benchmarking!
		}
#endif

	Tc_1 = vector_round(Tc_1);
	Tc_2 = vector_round(Tc_2);
	Tc_3 = vector_round(Tc_3);

	/* Unlike in the CPU version, all coordinates must be positive. This involves another translation (which
	 * is also inversely done on the GPU)
	 */

	transX = (int) Vmin(Tc_1, Tc_2, Tc_3, x);
	transY = (int) Vmin(Tc_1, Tc_2, Tc_3, y);

	Tc_1.x -= transX;
	Tc_2.x -= transX;
	Tc_3.x -= transX;
	Tc_1.y -= transY;
	Tc_2.y -= transY;
	Tc_3.y -= transY;

	maxX = (int) Vmax(Tc_1, Tc_2, Tc_3, x);
	maxY = (int) Vmax(Tc_1, Tc_2, Tc_3, y);

	/* Discard if extents > maxsize. Note that this is NOT the same as the initial discarding test!
	 */

	if(maxX > maxsize || maxY > maxsize)
#ifndef VOXELTEST
		return NULL;
#else
		{
			printf("[cpfw] ABORTING GPU BENCHMARK: Triangle extents (maxX = %d vs %d, maxY = %d vs. %d) too large.\n", maxX, maxsize, maxY, maxsize);
			abort(); // should not happen when benchmarking!
		}
#endif

	/* We use a texture which is equal to the size of the bounding box of the triangle.
	 */

	glGenTextures( 1, &texture_id );

	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fb );

	/* Initialize output texture
	 * Disable all filtering and disable looping
	 */

	glBindTexture( GL_TEXTURE_RECTANGLE_ARB, texture_id );	

	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

	/* Each texture element has 4 components. The following call initializes the texture.
	 */

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GL_FLOAT_RGBA32_NV, maxX, maxY, 0, GL_RGBA, GL_FLOAT, NULL );

	/* Bind to framebuffer so we can render to the texture.
	 */

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, texture_id, 0);

	gpureadbufsize = maxX * maxY * 4;
	gpureadbuf = fw_malloc(sizeof(float) * gpureadbufsize);
	mlock(gpureadbuf, gpureadbufsize);


	/* Set translation/quaternion matrix as parameters to fragment shader program
	 */

	quaternion_update_matrix(q1);

	cgGLSetParameter2d(cgprog_F_param_minXY, transX, transY);

	/* Add inverse translation to the matrix. Also, element (4,4) contains -1 so that the alpha field in the texture
	 * marks the coordinate as rasterized
	 */

	q1->mtx[3]  = v1.x;
	q1->mtx[7]  = v1.y;
	q1->mtx[11] = v1.z;
	q1->mtx[15] = -1.0;

	cgGLSetMatrixParameterdr(cgprog_F_param_rotmtx, q1->mtx);

	/* Step 1. Run clearing program, which sets all elements of output texture = 0
	 */

	cgGLBindProgram(cgprog_V);
	cgGLBindProgram(cgprog_F_clear);

	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);		glVertex2f(0.0, 0.0);
		glTexCoord2f(maxX, 0.0);	glVertex2f(maxX, 0.0);
		glTexCoord2f(maxX, maxY);	glVertex2f(maxX, maxY);
		glTexCoord2f(0.0, maxY);	glVertex2f(0.0, maxY);
	glEnd();

	/* Step 2. Run actual voxelization program.
	 */

	cgGLBindProgram(cgprog_F);

	glBegin(GL_TRIANGLES);
		glTexCoord2d(Tc_1.x, Tc_1.y); glVertex2d(Tc_1.x, Tc_1.y);
		glTexCoord2d(Tc_2.x, Tc_2.y); glVertex2d(Tc_2.x, Tc_2.y);
		glTexCoord2d(Tc_3.x, Tc_3.y); glVertex2d(Tc_3.x, Tc_3.y);
	glEnd();


	glFinish();

	/* Step 3. Download to CPU, find voxels.
	 */

	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glReadPixels( 0, 0, maxX, maxY, GL_RGBA, GL_FLOAT, gpureadbuf );
	



	tmpc.z = 0;


	mlockall(MCL_CURRENT|MCL_FUTURE);

	/* Iterate over texture and detect elements where alpha value != 0, indicating an actual value
	 */

	for(i = 0; i < maxX * maxY * 4; i += 4) {
		if(gpureadbuf[i + 3] < 0) {
			tmpc.x = gpureadbuf[i];
			tmpc.y = gpureadbuf[i + 1];
			tmpc.z = gpureadbuf[i + 2];

			/* If running the benchmarking program,
			 * we do not hash coordinate values since this makes
			 * the time reflect the hashing time and not extraction time.
			 * The hashing time is precisely equal on GPU and CPU versions and therefore
			 * does nothing but add unnecessary noise to the benchmarks.
			 * (Naturally, to make benchmarks fair, we also do the same for the CPU version in extract_cpu.c!)
			 */
#ifndef VOXELTEST
			coordinate_list_add(res, &tmpc);
#else
			res->coords[res->used++] = tmpc;
#endif
		}
	}
	munlockall();

	glDeleteTextures(1, &texture_id);

	munlock(gpureadbuf, gpureadbufsize);
	fw_free(gpureadbuf);

	return res; 
}


/* extract_gpu_cgrerror: Cg error callback called when Cg reports an error. 
 */


static void extract_gpu_cgerror(void) 
{
	if(cgGetError()) {
		printf("%s \n", cgGetErrorString(cgGetError()));
		_exit(-1);
	}
} 


/* print_texture: Debugging routine, print @data of size @w, @h
 */


static void print_texture(float *data, int w, int h)
{
	int i,j;
	for(j = 0; j < h; j++) {
		printf("[ ");
		for(i = 0; i < w; i++) {
			printf("%.5f ", data[i + j * w]);
		}
		printf("]\n");
	}
}

#endif
