/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_loadbalancer.c: Implements the load-balancers
 *
 * The load-balancers are described in detail in Chapter 5 of the thesis.
 *
 * The three load-balancers are:
 *   - global (loadbalance_globalextract)
 *   - local (loadbalance_localextract)
 *   - manhattan (loadbalance_manhattanextract)
 *
 * Additionally, this file contains tables which keep track of
 * 	1) how much area is stored on each node,
 *	2) how many faces are stored on each node, and
 *	3) the sum of all areas stored on all nodes 
 *
 * These tables are updated in
 *   - loadbalance_hook_addpolygon when adding faces,
 *   - loadbalance_hook_removevertex when removing a vertex,
 *   - loadbalance_hook_movevertex_pre/post when moving a vertex.
 *
 * See also the comments above the individual functions.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <search.h>
#include <assert.h>

#include "fw_main.h"
#include "fw_wingededge.h"
#include "fw_loadbalancer.h"


static double *faceareas = NULL; 		// faceareas[i] denotes the current area load on node i
static size_t *nodefaces = NULL;		// nodefaces[i] denotes the number of faces on node i

static double totalarea = 0.0;			// sum of all faceareas[i]

extern double delta;				// delta value, from fw_main


/* loadbalance_hook_addpolygon: This function is called every time a new
 * polygon face @f is added to some mesh.
 *
 * The function updates the faceareas, totalarea and nodefaces variables
 */

void		loadbalance_hook_addpolygon(face *f)
{
	extern int singlemode;
	extern int lbmode;

	if(singlemode)
		return;

	assert( faceareas != NULL );
	assert( nodefaces != NULL );
	

	faceareas[f->rnode] += f->area;
	totalarea += f->area;
	nodefaces[f->rnode]++;
}


/* loadbalance_hook_removevertex: This function is called every time a vertex
 * @v is removed from a mesh, but before it is actually removed.
 *
 * The function updates the faceareas, totalarea and nodefaces variables
 */


void		loadbalance_hook_removevertex(vertex *v)
{
	int i;

	extern int rank;

	extern int singlemode;
	extern int lbmode;

	if(singlemode)
		return;

	assert( faceareas != NULL );
	assert( nodefaces != NULL );
	
	
	/* Sometimes, if a vertex remove operation happens, other vertices which do not belong to any face
	 * might be removed. Since this can not possibly affect area computation, we can safely return in this case.
	 */

	if(v == NULL)
		return;

	for(i = 0; i < v->nfaces; i++) {
		faceareas[v->faces[i]->rnode] -= v->faces[i]->area;
		totalarea -= v->faces[i]->area;
		nodefaces[v->faces[i]->rnode]--;
	}
}

/* loadbalance_hook_movevertex_pre: This function is called prior to
 * a vertex being moved. 
 *
 * Since a move operation can be logically
 * viewed as a remove operation followed by an add operation,
 * this function essentially does the 'remove' part by subtracting the
 * areas of the faces incident to the vertex from the entries in faceareas/nodefaces.
 * Also, the totalarea is subtracted. The loadbalance_hook_movevertex_post function,
 * shown below, then adds the areas of the faces after the vertex has been moved.
 *
 * The function updates the faceareas, totalarea and nodefaces variables
 */


void		loadbalance_hook_movevertex_pre(vertex *v)
{
	int i;

	extern int singlemode;
	extern int lbmode;

	if(singlemode)
		return;

	assert( faceareas != NULL );
	assert( nodefaces != NULL );
	

	for(i = 0; i < v->nfaces; i++) {
		faceareas[v->faces[i]->rnode] -= v->faces[i]->area;
		totalarea -= v->faces[i]->area;
		nodefaces[v->faces[i]->rnode]--;
	}
}




/* loadbalance_hook_movevertex_pre: This function is called subsequent to
 * a vertex being moved. 
 * 
 * After a vertex has been moved, this function updates the faceareas, totalarea and
 * nodefaces variables to reflect the (possible) changes in the areas of the
 * faces incident to the vertex which has been moved.
 *
 * See also comment on loadbalance_hook_movevertex_pre.
 *
 * The function updates the faceareas, totalarea and nodefaces variables
 */

void		loadbalance_hook_movevertex_post(vertex *v)
{
	int i;

	extern int singlemode;
	extern int lbmode;

	if(singlemode)
		return;

	assert( faceareas != NULL );
	assert( nodefaces != NULL );

	
	for(i = 0; i < v->nfaces; i++) {
		faceareas[v->faces[i]->rnode] += v->faces[i]->area;
		totalarea += v->faces[i]->area;
		nodefaces[v->faces[i]->rnode]++;
	}
}

/* global_qsort_comparefcn: This is an auxiliary function used when sorting
 * in loadbalance_globalextract. Input are pointers to two ints, which are two nodes
 * and index the faceareas array. 
 *  - If faceareas[*a] < faceareas[*b], return -1 since *a has less area than *b.
 *  -                  =                        0              equal
 *  -                  >                        1              greater
 */

static int global_qsort_comparefcn(const void *a, const void *b)
{
	int aidx = *(int *) a;
	int bidx = *(int *) b;

	if(faceareas[aidx] < faceareas[bidx])
		return -1;
	else if(faceareas[aidx] == faceareas[bidx])
		return 0;
	else
		return 1;
}

/* loadbalance_globalextract: Global load-balancing algorithm, as described in Chapter 5 of the thesis.
 */


void 	loadbalance_globalextract(void)
{
	int i, j, l;			// temporary variables
	int partnernode;

	mesh *msh;
	face *f;

	extern int n, m, k;		// grid size
	extern int size, singlemode;	// # of nodes, whether or not running without MPI
	extern mesh *meshlist;		// global mesh list

	face **faces[size];		// temporary array of face pointers, one for each node (explained further below)

	int nodeidx[size];		// temporary index of the faces array
	int sortednodes[size];


	// any polygons at all or single mode?

	if(meshlist == NULL || singlemode)
		return;

	// allocate faces[i] array, which is an array of pointers to faces.

	for(i = 0; i < size; i++) {
		if(nodefaces[i] > 0)
			faces[i] = fw_malloc( sizeof(face *) * nodefaces[i] );
		else
			faces[i] = NULL;
		sortednodes[i] = i;
	}

	memset(nodeidx, 0, sizeof(int) * size);

	// faces[i][j] points to the jth face of node i.
	// this is done so that we can efficiently access all faces belonging to a node.
	// faces[i] thus corresponds to F^i in the thesis.

	for(msh = meshlist; msh != NULL; msh = msh->next) {
		for(f = msh->faces; f != NULL; f = f->next) {
			faces[f->rnode][nodeidx[f->rnode]++] = f;
		}
	}	


	// This is the main loop of the global load-balancing algorithm
	// presented in the thesis.

	for(i = 0; i < size; i++) {

		if(faceareas[i] <= totalarea/size + delta)
			continue; // we are underloaded/balanced, nothing to do

		// Sort nodes according to increasing faceareas[i] size

		qsort(&sortednodes[0], size, sizeof(int), global_qsort_comparefcn);

		// iterate over nodes in order of increasing load

		for(j = 0; j < size; j++) {
			int partnernode = sortednodes[j];	// corresponds to k in the thesis

			// as described in the thesis, target is the minimum of the available capacity on the partner node
			// and the surplus area stored on node i (which will transfer faces to the partner node)

#define min(a,b) ((a) < (b) ? (a) : (b))
			double target = min( totalarea/size - faceareas[partnernode], faceareas[i] - totalarea/size ) ;
#undef min

			if(partnernode == i)
				continue;
			
			// Check if (1) partner node is overloaded --> in which case we must be done since all succeeding nodes in sortednodes array are also overloaded
			//          (2) this node is not overloaded --> in which case we can continue processing the next node

			if(faceareas[partnernode] >= totalarea/size || faceareas[i] <= totalarea/size)
				break;


			// Select faces and assign them to the partner node until the target <= 0 or no further faces are found.

			do {


				face *targetface = NULL;
				double targetdiff = n * m * k; // very large value 

				// assign the face which matches the target closely. we thus move one large face
				// rather than several small faces to prevent too much communication.
		
				for(l = 0; l < nodefaces[i]; l++) {
					if(target - faces[i][l]->area < targetdiff && target - faces[i][l]->area >= 0 && faces[i][l]->rnode == i) {
						targetdiff = target - faces[i][l]->area;
						targetface = faces[i][l];
					}
				}

				if(!targetface)	// no faces found
					break;


				// Update areas

				targetface->rnode = partnernode;
				faceareas[i] -= targetface->area;
				faceareas[partnernode] += targetface->area;
				target -= targetface->area;
				nodefaces[i]--;
				nodefaces[partnernode]++;

			} while(target > 0); 
		}
	}


	// Free temporary array

	for(i = 0; i < size; i++) {
		if(faces[i])
			fw_free(faces[i]);
	}
}


/* The following two routines implement the local load-balancing strategy.
 * This code is explained in detail in Chapter 5 of the thesis.
 */


/* loadbalance_local: Do local load-balancing on a face @f which has an
 * overloaded centroid node. This function is used as an auxiliary function
 * to loadbalance_localextract. See thesis Chapter 5 for further information
 * on the ideas behind this strategy.
 *
 * @faceareas_tmp is an array holding the current area stored on all nodes.
 * This function finds nodes whose coordinate space intersects the bounding
 * volume of the polygon, detects which of those nodes having the minimum load,
 * and makes that node responsible for the face.
 */

static void		loadbalance_local(face *f, double *faceareas_tmp)
{
	int i, j, r;
	int cartranks_min[2], cartranks_max[2], cartranks_tmp[2];

	int minnode;
	double minarea;

	extern int rank, size, singlemode;
	extern MPI_Comm cpfwcomm;



	if(singlemode)
		return;

	assert( faceareas_tmp != NULL );

	minarea = faceareas_tmp[0];

	/* Convert the minimum/maximum coordinate of f to the 2D Cartesian coordinates
	 * in the MPI cpfwcomm communicator.
	 */

	coordinate_to_cartrank(we_coordinate_min(f), cartranks_min);
	coordinate_to_cartrank(we_coordinate_max(f), cartranks_max);

	minarea = faceareas_tmp[f->rnode];
	minnode = f->rnode;


	/* Iterate through the 2D box starting at the node responsible for the
	 * minimum coordinate of the face, going up (in 2D) to the node responsible
	 * for the maximum coordinate of the face.
	 */

	for(i = cartranks_min[0]; i <= cartranks_max[0]; i++) {
		for(j = cartranks_min[1]; j <= cartranks_max[1]; j++) {
			cartranks_tmp[0] = i; 
			cartranks_tmp[1] = j;
			MPI_Cart_rank(cpfwcomm, cartranks_tmp, &r);

			if(faceareas_tmp[r] < minarea) {
				minarea = faceareas_tmp[r];
				minnode = r;
			}

		}
	}

	f->rnode = minnode;
}



/* loadbalance_localextract: The local load-balancing strategy. For each face,
 * we check if the centroid node is overloaded. If so, loadbalance_local is called for the face.
 * Also a temporary array holding the current face areas (to check if nodes are overloaded or not)
 * is maintained.
 *
 * See thesis Chapter 5 for further details.
 *
 */

void		loadbalance_localextract(void)
{
	extern int size;
	extern mesh *meshlist;

	double faceareas_tmp[size];
	double totalarea_tmp = 0.0;

	mesh *msh;
	face *f;


	memset(faceareas_tmp, 0, sizeof(double) * size);


	for(msh = meshlist; msh != NULL; msh = msh->next) {
		for(f = msh->faces; f != NULL; f = f->next) {
			if(faceareas_tmp[f->rnode] >= totalarea_tmp/size + delta)
				loadbalance_local(f, faceareas_tmp);
			faceareas_tmp[f->rnode] += f->area;
			totalarea_tmp += f->area;		
		}
	}	
}

/* The following functions implement the Manhattan load-balancing strategy.
 * The function loadbalance_manhattan_init() is called on program startup
 * to initiallize the manhattanmaps array.
 *
 * manhattanmaps[i] is an array of size 'size' (# of nodes) which contains 
 * manhattan_map_entry structs. Each struct gives the node and the Manhattan
 * distance to the node. manhattanmaps[i] is sorted in order of increasing Manhattan
 * distances from node i.
 */

struct manhattan_map_entry {
	int node;
	int distance;
};

static struct manhattan_map_entry **manhattanmaps = NULL;

/* loadbalance_manhattan_node_exists: Check if @node is already inserted into the manhattan map
 * of @startnode. @tmp_mhsize is the number of nodes already inserted into the manhattan map.
 * This function returns 1 if the node is already inserted and 0 if not.
 * 
 * This is used as a convenience in loadbalance_manhattan_init.
 */

static inline int 	loadbalance_manhattan_node_exists(int startnode, int node, int tmp_mhsize)
{
	int i;

	for(i = 0; i < tmp_mhsize; i++) {
		if(manhattanmaps[startnode][i].node == node)
			return 1;
	}

	return 0;
}


/* loadbalance_manhattan_init: Initialize the manhattanmaps array.
 *
 * For all nodes, we add the node to a queue. Next, the node is taken from the queue, and
 * the left, right, upper and lower neighbours of the node are expanded similarly until
 * all nodes have been visited.
 * 
 * All nodes are added successively in the manhattanmaps array. Since the Manhattan distance
 * from a node to its left, right, upper, and lower neighbous is 1, the array will contain
 * the nodes in the order of increasing Manhattan distance.
 *
 */

void		loadbalance_manhattan_init(void)
{
	int i, j, k;

	int expandidx, mhsize;
	struct manhattan_map_entry expandnode;

	extern int size, singlemode;

	if(singlemode)
		return;

	printf_zero("[cpfw/loadbalance_manhattan_init] Generating a map over Manhattan...\n");

	manhattanmaps = fw_malloc(sizeof(struct manhattan_map_entry *) * size);

	// Initial state: Add node as the first entry to its manhattanmaps array 

	for(i = 0; i < size; i++) {
		manhattanmaps[i] = fw_malloc(sizeof(struct manhattan_map_entry) * size);
		manhattanmaps[i][0].node = i;
		manhattanmaps[i][0].distance = 0;

	}

	// Next, expand nodes so that the entire array is filled out. The manhattanmaps-array
	// is also used as a queue to keep track of which node we are expanding.
	// The 'queue pointer' is expandidx, and the node to expand is 'expandnode'.
	// In the inner loop, indices in manhattanmaps[i] which are < expandnode have been expanded.
	// Expanded nodes are successively added to manhattanmaps[i] array. The current length
	// of the array is in the mhsize parameter.

	for(i = 0; i < size; i++) {

		expandnode = manhattanmaps[i][0];
		expandidx = 1;

		mhsize = 1;

		while(mhsize < size) {

			if((j = coordinate_cart_getleftneighbor(expandnode.node)) >= 0) {
				if(!loadbalance_manhattan_node_exists(i, j, mhsize)) {
					manhattanmaps[i][mhsize].node = j;
					manhattanmaps[i][mhsize++].distance = expandnode.distance + 1;
				}
			}

			if((j = coordinate_cart_getrightneighbor(expandnode.node)) >= 0) {
				if(!loadbalance_manhattan_node_exists(i, j, mhsize)) {
					manhattanmaps[i][mhsize].node = j;
					manhattanmaps[i][mhsize++].distance = expandnode.distance + 1;
				}
			}

			if((j = coordinate_cart_getupperneighbor(expandnode.node)) >= 0) {
				if(!loadbalance_manhattan_node_exists(i, j, mhsize)) {
					manhattanmaps[i][mhsize].node = j;
					manhattanmaps[i][mhsize++].distance = expandnode.distance + 1;
				}
			}

			if((j = coordinate_cart_getlowerneighbor(expandnode.node)) >= 0) {
				if(!loadbalance_manhattan_node_exists(i, j, mhsize)) {
					manhattanmaps[i][mhsize].node = j;
					manhattanmaps[i][mhsize++].distance = expandnode.distance + 1;
				}
			}

			if(expandidx >= mhsize) // done
				break;

			expandnode = manhattanmaps[i][expandidx++];

		}

	}

	printf_zero("[cpfw/loadbalance_manhattan_init] Manhattan map generation done!\n");

}


/* loadbalance_manhattanextract: The Manhattan load-balancing strategy.
 * This is described in the thesis, Chapter 5.
 */

void		loadbalance_manhattanextract(void)
{
	int i, j, l;			// temporary variables
	int partnernode;

	mesh *msh;
	face *f;

	extern int n, m, k;		// grid size
	extern int size, singlemode;	// # of nodes, whether or not running without MPI
	extern mesh *meshlist;		// global mesh list
	extern int mhthres;		// threshold from fw_main
	

	face **faces[size];		// temporary array of face pointers, one for each node (explained further below)

	int nodeidx[size];		// temporary index of the faces array


	// any polygons at all or single mode?

	if(meshlist == NULL || singlemode)
		return;

	// allocate faces[i] array, which is an array of pointers to faces.

	for(i = 0; i < size; i++) {
		if(nodefaces[i] > 0)
			faces[i] = fw_malloc( sizeof(face *) * nodefaces[i] );
		else
			faces[i] = NULL;
	}

	memset(nodeidx, 0, sizeof(int) * size);

	// faces[i][j] points to the jth face of node i.
	// this is done so that we can efficiently access all faces belonging to a node.
	// faces[i] thus corresponds to F^i in the thesis.

	for(msh = meshlist; msh != NULL; msh = msh->next) {
		for(f = msh->faces; f != NULL; f = f->next) {
			faces[f->rnode][nodeidx[f->rnode]++] = f;
		}
	}	




	// This is the main loop of the Manhattan load-balancing algorithm
	// presented in the thesis.

	for(i = 0; i < size; i++) {

		if(faceareas[i] <= totalarea/size + delta)
			continue; // we are underloaded/balanced, nothing to do


		// iterate over nodes in order of increasing Manhattan distance

		for(j = 1; j < size; j++) {
			int partnernode = manhattanmaps[i][j].node;

			// threshold exceeded?
			if(manhattanmaps[i][j].distance > mhthres)
				break;

			// partner node is underloaded

			if(faceareas[partnernode] < totalarea/size + delta) {
				// as described in the thesis, target is the minimum of the available capacity on the partner node
				// and the surplus area stored on node i (which will transfer faces to the partner node)

#define min(a,b) ((a) < (b) ? (a) : (b))
				double target = min( totalarea/size + delta - faceareas[partnernode], faceareas[i] - (totalarea/size + delta) ) ;
#undef min



				
				// the following loop assignes faces until the target area has been reached

				do {


					face *targetface = NULL;
					double targetdiff = n * m * k; // very large value 

					// assign the face which matches the target closely. we thus move one large face
					// rather than several small faces to prevent too much communication.
		
					for(l = 0; l < nodefaces[i]; l++) {
						if(target - faces[i][l]->area < targetdiff && target - faces[i][l]->area >= 0 && faces[i][l]->rnode == i) {
							targetdiff = target - faces[i][l]->area;
							targetface = faces[i][l];
						}
					}

					if(!targetface)
						break;


					// Update areas

					targetface->rnode = partnernode;
					faceareas[i] -= targetface->area;
					faceareas[partnernode] += targetface->area;
					target -= targetface->area;
					nodefaces[i]--;
					nodefaces[partnernode]++;

				} while(target > 0); 

			}

			// Still overloaded? if not, break. if so, continue to the next node.

			if(faceareas[i] <= totalarea/size)
				break;
		}
	}


	// Free temporary array

	for(i = 0; i < size; i++) {
		if(faces[i])
			fw_free(faces[i]);
	}
}


/* loadbalance_reset: Reset the loadbalancer state.
 * This is done at the end of each benchmark run and resets the area counters.
 */

void		loadbalance_reset(void)
{
	int i;

	extern int singlemode;
	extern int size;

	if(singlemode)
		return;

	if(!faceareas)
		faceareas = fw_malloc(sizeof(double) * size);
	if(!nodefaces)
		nodefaces = fw_malloc(sizeof(size_t) * size);

	memset(faceareas, 0, sizeof(double) * size);
	memset(nodefaces, 0, sizeof(size_t) * size);
}



extern double *loadstdev;
extern int cycles, cyclenum;
extern int repeats;

/* loadbalance_statistics_update: Update load-balancing statistics. This involves
 * computing the standard deviation of the load of each node.
 * This routine is not included in any timings.
 *
 */

void		loadbalance_statistics_update(void)
{
	extern mesh *meshlist;
	extern int size;

	double faceareas_tmp[size];
	double totalarea_tmp = 0.0;
	
	double sdev = 0.0;
	
	mesh *msh;
	face *f;

	int i;

	if(size == 1) {
		loadstdev[cyclenum] = 0.0;
		return;
	}

	memset(faceareas_tmp, 0, sizeof(double) * size);

	for(msh = meshlist; msh != NULL; msh = msh->next) {
		for(f = msh->faces; f != NULL; f = f->next) {
			if(!isnan(f->area)) {
				faceareas_tmp[f->rnode] += f->area;
				totalarea_tmp += f->area;
			}		
		}
	}

	for(i = 0; i < size; i++) {
		sdev += (faceareas_tmp[i] - totalarea_tmp/size) * (faceareas_tmp[i] - totalarea_tmp/size);
	}

	sdev = sqrt(sdev/(size - 1));

	loadstdev[cyclenum] += sdev/repeats;
}

