/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 * Implementation based on implementation given in Pre-Master's project.
 * See L. C. Larsen: Utilizing GPUs on Computer Clusters, IDI-NTNU, 2006.
 *
 */

/* 
 * Implementation of Fast LOT algorithm, using even boundary extensions
 *
 * See Malvar, Henrique: Signal Processing with Lapped Transforms. Artech House, 1992.
 *
 * Some code duplication in the cpu_{i}flot_* functions, 
 * done to avoid function call overhead
 *
 * Uses the quickest DCT algorithm (FL MAD) for the DCT/IDCT step.
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

#define __USE_GNU
#include <math.h>

#include "fw_main.h"
#include "fw_lot.h"

/* Maple-calculated coefficients
 */

typedef float f32;

#define FLMAD_C1	0.9807852804032304491261822361342390L	/* cos(1 * Pi / 16)		*/
#define FLMAD_C2	0.9238795325112867561281831893967883L	/* cos(2 * Pi / 16)		*/
#define FLMAD_C4	0.7071067811865475244008443621048490L	/* cos(4*Pi/16) = 1/sqrt(2) 	*/
#define FLMAD_C5	0.5555702330196022247428308139485329L	/* cos(5 * Pi / 16)		*/

#define FLMAD_T1	0.1989123673796580069115976226446762L	/* tan(1*Pi/16)		    	*/
#define FLMAD_T2	0.4142135623730950488016887242096981L	/* tan(2*Pi/16)		    	*/
#define FLMAD_T5	1.4966057626654890176011351349424769L	/* tan(5*Pi/16)		    	*/


#define LOT_COS_THETA1	0.9177546256839811411456038575494851L	/* cos(0.13 * Pi)		*/
#define LOT_COS_THETA2	0.8763066800438635873081159039220626L	/* cos(0.16 * Pi)		*/
#define LOT_COS_THETA3	LOT_COS_THETA1				/* cos(0.13 * Pi)		*/
#define LOT_SIN_THETA1	0.3971478906347806137543773600194771L	/* sin(0.13 * Pi)		*/
#define LOT_SIN_THETA2	0.4817536741017152749871915028721297L	/* sin(0.16 * Pi)		*/
#define LOT_SIN_THETA3	LOT_SIN_THETA1				/* sin(0.13 * Pi)		*/


void cpu_flot_32(float *inarray, float *outarray, int length)
{
	int x, x_start;
	const int xblocks = length/8;
	f32 *in_row, *out_row;
	f32 u0, u1, u2, u3, u4, u5, u6, u7;
	f32 v0, v1, v2, v3, v4, v5, v6, v7;

	/* 1st Step. FL MAD DCT on rows
	 */


	for(x_start = 1; x_start < xblocks - 1; x_start++) {

		/* Transform rows */
		in_row = &inarray[ x_start * 8 ];
		out_row = &outarray[ x_start * 8 ];
				
		/* First step. Multiply with A[0] matrix = 8 adds
		 */

		v0 = in_row[0] + in_row[7];
		v1 = in_row[1] + in_row[6];
		v2 = in_row[3] + in_row[4];
		v3 = in_row[2] + in_row[5];
		v4 = in_row[0] - in_row[7];
		v5 = in_row[1] - in_row[6];
		v6 = in_row[3] - in_row[4];
		v7 = in_row[5] - in_row[2];

		/* Second step. Multiply with A[1] matrix = 6 adds
		 */

		u0 = v0 + v2;
		u1 = v1 + v3;
		u2 = v0 - v2;
		u3 = v1 - v3;
		/* u4 = v4; */
		u5 = v5 + v7;
		/* u6 = v6; */
		u7 = v5 - v7;

		/* Third step. Multiply with A[2] matrix =  4 multiply-add operations. Note: reordered in order to minimize # of variables used 
		 * Two of the multiplications are redundant, since they calculate the same product. But this does not necessarily save time,
		 * particularly on the GPU which can calculate the four multiplications in parallel. If we pre-compute the products,
		 * total instruction count needed is greater.
		 *
		 * Thus instead of explicitly caching the multiplication result, we leave it up to the compiler to decide what to do.
		 */

		/* Optimized to eliminate superfluous assignments
				v0 = u0;
				v1 = u1;
				v2 = u2;
				v3 = u3;
		*/

		v7 = - (FLMAD_C4 * u7) + v4;
		v4 += FLMAD_C4 * u7;
		v5 = FLMAD_C4 * u5 + v6;
		v6 -= FLMAD_C4 * u5;

		/* Fourth step. Multiply with A[3] matrix = 8 MAD operations. 
		 */

		u6 = u2 * (f32) FLMAD_T2 - u3;
		u2 = u3 * (f32) FLMAD_T2 + u2;
		u4 = u0 - u1;
		u0 = u0 + u1;
		u1 = v5 * (f32) FLMAD_T1 + v4;
		u3 = v7 * (f32) FLMAD_T5 - v6;
		u5 = v6 * (f32) FLMAD_T5 + v7;
		u7 = v4 * (f32) FLMAD_T1 - v5;



		/* Fifth step. Multiply with D matrix and scale 
		 * (D-multiply is combined with scaling)
		 *
		 * even coeffs -> out_row[0-3], odd coeffs -> out_row[4-7]
		 */
		out_row[0] = ((f32) FLMAD_C4/2.0) * u0;
		out_row[1] = ((f32) FLMAD_C2/2.0) * u2;
		out_row[2] = ((f32) FLMAD_C4/2.0) * u4;
		out_row[3] = ((f32) FLMAD_C2/2.0) * u6;
		out_row[4] = ((f32) FLMAD_C1/2.0) * u1;
		out_row[5] = ((f32) FLMAD_C5/2.0) * u3;
		out_row[6] = ((f32) FLMAD_C5/2.0) * u5;
		out_row[7] = ((f32) FLMAD_C1/2.0) * u7;
	}

	/* 2nd step - first butterflies */


	for(x_start = 1; x_start < xblocks; x_start++) {
		out_row = &outarray[ x_start * 8 ];
		in_row = &outarray[ x_start * 8 ];

		u0 = in_row[0] + in_row[4];
		u1 = in_row[1] + in_row[5];
		u2 = in_row[2] + in_row[6];
		u3 = in_row[3] + in_row[7];
		u4 = in_row[0] - in_row[4];
		u5 = in_row[1] - in_row[5];
		u6 = in_row[2] - in_row[6];
		u7 = in_row[3] - in_row[7];

		out_row[0] = u0;
		out_row[1] = u1;
		out_row[2] = u2;
		out_row[3] = u3;
		out_row[4] = u4;
		out_row[5] = u5;
		out_row[6] = u6;
		out_row[7] = u7;
	}


	for(x_start = 0; x_start < xblocks - 1; x_start++) {
		out_row = &outarray[ x_start * 8 ];

		u4 = 0.5 * (out_row[4] + out_row[8 + 0]);
		u5 = 0.5 * (out_row[5] + out_row[8 + 1]);
		u6 = 0.5 * (out_row[6] + out_row[8 + 2]);
		u7 = 0.5 * (out_row[7] + out_row[8 + 3]);

		u0 = 0.5 * (out_row[4] - out_row[8 + 0]);
		u1 = 0.5 * (out_row[5] - out_row[8 + 1]);
		u2 = 0.5 * (out_row[6] - out_row[8 + 2]);
		u3 = 0.5 * (out_row[7] - out_row[8 + 3]);

		out_row[4] = u4;
		out_row[5] = u5;		
		out_row[6] = u6;
		out_row[7] = u7;

		out_row[8 + 0] = u0;
		out_row[8 + 1] = u1;
		out_row[8 + 2] = u2;
		out_row[8 + 3] = u3;
	}

	/* 3rd Step. Z-block */


	for(x_start = 0; x_start < xblocks; x_start++) {
		out_row = &outarray[ x_start * 8 ];

		/* The previous iteration has uptaded out_row[0..3], or this is zero for the first iteration
		 */

		u0 = out_row[0] * LOT_COS_THETA1 - out_row[1] * LOT_SIN_THETA1;
		u1 = out_row[1] * LOT_COS_THETA1 + out_row[0] * LOT_SIN_THETA1;

		u2 = out_row[2] * LOT_COS_THETA2 + u1 * LOT_SIN_THETA2;
		u1 = u1 * LOT_COS_THETA2 - out_row[2] * LOT_SIN_THETA2;

		u3 = out_row[3] * LOT_COS_THETA3 + u2 * LOT_SIN_THETA3;
		u2 = u2 * LOT_COS_THETA3 - out_row[3] * LOT_SIN_THETA3;
		
		out_row[0] = u0;
		out_row[1] = u1;
		out_row[2] = u2;
		out_row[3] = u3;

	}


}
