/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_main.c: Main program/utility routines
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <mpi.h>

#include "fw_main.h"
#include "fw_script.h"
#include "fw_extract_cpu.h"
#include "fw_extract_gpu.h"
#include "fw_splitpolygons.h"
#include "fw_loadbalancer.h"

/* The following variables are global and used throughout the program.
 */

int		blocksize = 1; 			// Blocksize for plain extraction method
int 		threadcount = 1, gpumode = 0;	// # of threads and whether to use GPU or not
int 		rank, size;			// rank of process/# of processes
int 		ionode = 0;			// Can print to stdout
int 		singlemode = 0;			// No MPI is used
int		griddims[2];			// # of nodes in x,y directions
int		nodecoords[2];			// Coordinate of this node
int		lbmode;				// Load-balancing strategy
double		delta = 0.0;			// Load-balancing delta value
int		mhthres = 0;			// Manhattan load-balancing threshold

MPI_Comm cpfwcomm;				// MPI 2D cartesian communicator


/* printf_zero: Print format string @fmt on node 0.
 */

void	printf_zero(const char *fmt, ...)
{
	if(!ionode)
		return;
#pragma omp critical
 {
 	char vabuf[1024];
	
	va_list val;
	va_start(val, fmt);
	vsnprintf(vabuf, 1024, fmt, val);
	va_end(val);
	printf("%s", vabuf);
 }
}

/* usage: Usage information. @av0 is argv[0] from main
 */

static void	usage(const char *av0)
{
	printf_zero("\nUsage: mpirun -np <procs> %s <transfer mode> <load-balancing strategy> <gpu-extract|cpu-extract> <threads> <script file>\n", av0);
	printf_zero("<transfer mode> is one of:\n");
	printf_zero("  mc... ....................... : Minimum Communication strategy\n");
	printf_zero("  bt=N ........................ : Block Transfer strategy, blocksize = N\n");
	printf_zero("  bv .......................... : Bounding Volume strategy\n");
	printf_zero("<load-balancing strategy> is one of the following (see thesis on information on the delta parameter):\n");
	printf_zero("  none ........................ : no load-balancing, polygons are mapped to nodes based on polygon centroid\n");
	printf_zero("  global=delta................. : global load-balancing strategy with specified delta value\n");
	printf_zero("  local=delta.................. : local load-balancing strategy with specified delta value\n");
	printf_zero("  manhattan=delta[,th=eps] .... : Manhattan load-balancing strategy with specified delta value, and, optionally, a specificed threshold (tau) value.\n");

	printf_zero("Use programs in the script/ folder to generate script files. For further details see the thesis.\n");
}

/* init_communicator: Initialize MPI communicator. We set up a 2D communicator.
 * If there are N processes, this routine will factor N such that the number of
 * processes in x, y directions are as similar as possible.
 * For example:
 * 	- N = 8 gives 4x4 layout
 *	- N = 7 gives 1x7 layout
 *	- N = 6 gives 2x3 layout
 */

static int	init_communicator(void)
{
	int sz, periodic[2], sqrsz, found = 0;

	MPI_Comm_size(MPI_COMM_WORLD, &sz);

	MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_ARE_FATAL);

	if(sz == 1) {
		printf_zero("[cpfw] SINGLE-NODE MODE. Disabling MPI usage\n");
		singlemode = 1;
		griddims[0] = griddims[1] = 1;
		nodecoords[0] = nodecoords[1] = 0;
		cpfwcomm = MPI_COMM_WORLD;
		size = 1;
		MPI_Comm_rank(cpfwcomm, &rank);
		return 0;
	}

	// find factors a,b such that a * b = sz. sz might be a prime
	// start searching near sqrt(sz)

	sqrsz = round(sqrt(sz));

	for(griddims[0] = sqrsz; griddims[0] > 0 && !found; griddims[0]--) {
		for(griddims[1] = sqrsz; griddims[1] <= sz; griddims[1]++) {
			if(griddims[0] * griddims[1] == sz) {
				found = 1;
				goto loop_exit;
			}
		}
	}

	loop_exit:

	assert(found == 1);
	assert(griddims[0] != 0);
	assert(griddims[1] != 0);

	periodic[0] = periodic[1] = 0;

	printf_zero("[cpfw] Configuration: %d processes, 2D topology %dx%d, %d thread(s) per node\n", sz, griddims[0], griddims[1], threadcount);

	MPI_Cart_create(MPI_COMM_WORLD, 2, griddims, periodic, 1, &cpfwcomm);

	MPI_Errhandler_set(cpfwcomm, MPI_ERRORS_ARE_FATAL);

	MPI_Comm_rank( cpfwcomm, &rank );
	MPI_Comm_size( cpfwcomm, &size );

	MPI_Cart_coords(cpfwcomm, rank, 2, nodecoords);

	return 0;
}

/* main: Main entry point to cpfw application
 *       @argc argument count
 *	 @argv argument vector
 */

#if !defined(VOXELTEST) && !defined(GENSCRIPT)
int 	main(int argc, char **argv)
{
	int tmp;
	FILE *scriptfile;

	void (*extractpoints_f)(void);


	MPI_Init(&argc, &argv);

	MPI_Comm_rank(MPI_COMM_WORLD, &tmp);
	if(tmp == 0)
		ionode = 1;



	printf_zero("\nFramework for Polygonal Structures on Clusters\n");
	printf_zero("Leif Christian Larsen <leifchl@idi.ntnu.no>\n\n");

	// Create a 2D cartesian communicator

	if(init_communicator() < 0)
		goto finalize;
	

	if(argc != 6) {
		usage(argv[0]);
		goto finalize;
	}

	if(!strcasecmp(argv[1], "bv")) {
		printf_zero("[cpfw] Transfer Mode: Bounding volumes\n");
		extractpoints_f = splitpolygons_extractpoints_boundingvolumes;
	} else if(!strcasecmp(argv[1], "mc")) {
		printf_zero("[cpfw] Transfer Mode: Minimum Communication\n");
		extractpoints_f = splitpolygons_extractpoints_minimumtransfer;
	} else if(sscanf(argv[1], "bt=%d", &blocksize) == 1) {
		if(blocksize > 1024 || blocksize <= 1) {
			printf_zero("[cpfw] Error: block size b must 1 < b <= 1024.\n");
			goto finalize;
		}
		printf_zero("[cpfw] Transfer Mode: Blocked, blocksize %d\n", blocksize);
		extractpoints_f = splitpolygons_extractpoints_blockedtransfer;
	} else {
		printf_zero("[cpfw] Invalid transfer mode `%s'.\n", argv[1]);
		usage(argv[0]);
		goto finalize;
	}


	if(!strcasecmp(argv[2], "none")) {
		printf_zero("[cpfw] Load-balancer disabled.\n");
		lbmode = LOADBALANCE_NONE;
	} else if(sscanf(argv[2], "local=%lf", &delta) == 1) {
		printf_zero("[cpfw] Load-balancing on each extract operation, local mode (delta = %.2f).\n", delta);
		lbmode = LOADBALANCE_LOCAL_EXTRACT;
	} else if(sscanf(argv[2], "global=%lf", &delta) == 1) {
		printf_zero("[cpfw] Load-balancing on each extract operation, global mode (delta = %.2f).\n", delta);
		lbmode = LOADBALANCE_GLOBAL_EXTRACT;
	} else if(sscanf(argv[2], "manhattan=%lf,th=%d", &delta, &mhthres) == 2) {
		if(mhthres <= 0) {
			printf("[cpfw] error: Manhattan threshold must be > 0.\n");
			goto finalize;
		}
		printf_zero("[cpfw] Load-balancing on each extract operation, manhattan mode (delta = %.2f, threshold = %d)\n", delta, mhthres);
		lbmode = LOADBALANCE_MANHATTAN_EXTRACT;
	} else if(sscanf(argv[2], "manhattan=%lf", &delta) == 1) {
		mhthres = 0xffffff;
		printf_zero("[cpfw] Load-balancing on each extract operation, manhattan mode (delta = %.2f)\n", delta, mhthres);
		lbmode = LOADBALANCE_MANHATTAN_EXTRACT;
	} else {
		printf_zero("[cpfw] Invalid load-balancing strategy `%s'. (Did you remember to specify the delta value?)\n", argv[2]);
		usage(argv[0]);
		goto finalize;
	}


	if(!strcasecmp(argv[3], "gpu-extract")) {
#ifndef GPU_DISABLE
		printf_zero("[cpfw] using GPU voxelization algorithm for suitably sized triangles, CPU algorithm will be used for other triangles\n");
		gpumode = 1;
#else
		printf_zero("[cpfw] Program was compiled with GPU_DISABLE flag. GPU support not enabled. Recompile without this flag. Aborting.\n");
		goto finalize;
#endif
	} else if(!strcasecmp(argv[3], "cpu-extract")) {
		printf_zero("[cpfw] using CPU voxelization algorithm\n");
		gpumode = 0;
	} else {
		printf_zero("error: argument 2 must be either 'gpu-extract' or 'cpu-extract'\n");
		usage(argv[0]);
		goto finalize;
;	}



	threadcount = atoi(argv[4]);
	if(threadcount <= 0) {
		printf_zero("[cpfw] Invalid threadcount specified (%s), using 1 thread.\n", argv[3]);
		threadcount = 1;
	}


	if(!(scriptfile = fopen(argv[5], "r"))) {
		printf_zero("[cpfw] unable to open script file %s: %s\n", argv[5], strerror(errno));
		goto finalize;
	}

	extract_cpu_init();
#ifndef GPU_DISABLE
	if(gpumode)
		extract_gpu_init(&argc, argv);
#endif
	// Execute the script file

	script_execute(scriptfile, extractpoints_f, splitpolygons_addpolygon, splitpolygons_removepolygon, splitpolygons_movevertex);

finalize:
	printf_zero("[cpfw] Finalizing...\n");
	MPI_Finalize();
	printf_zero("[cpfw] Terminate\n");
	return 0;
}

#endif
