/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_quaternion.c: Implementation of quaternions
 * This file implements support for quaternions. The main functions are:
 *  - creating quaternions
 *  - combining quaternions, and
 *  - obtaining the 4x4 rotation matrix from a quaternion.
 */

#include <assert.h>
#include "fw_vector.h"
#include "fw_quaternion.h"


/* quaternion_set_angle: Update/set the angle of the quaternion q.
 * Prior to calling this function, q->s or q->v would be set by the calling function.
 */

static void		quaternion_set_angle(quaternion *q)
{
	q->q[0] = cos(q->s/2);
	q->q[1] = sin(q->s/2) * q->v.x;
	q->q[2] = sin(q->s/2) * q->v.y;
	q->q[3] = sin(q->s/2) * q->v.z;

	quaternion_normalize(q);
}

/* quaternion_create: Create a new quaternion with vector @v and angle @s.
 * Retuns the new quaternion.
 */

quaternion		quaternion_create(vector v, double s)
{
	struct quaternion res;

	res.qtype = Q_NORMAL;
	res.v = vector_normalize(v);
	res.s = s;

	quaternion_set_angle(&res);

	return res;
}

/* quaternion_normalize: Normalize the quaternion @q. Quaternions must be normalized in order
 * to compute the rotation matrix correctly. This is also done after combining quaternions to one.
 * Unnormalized quaternions result in scaling.
 */

void			quaternion_normalize(quaternion *q)
{
	double iqlength = 1.0/sqrt(q->q[0] * q->q[0] + q->q[1] * q->q[1] + q->q[2] * q->q[2] + q->q[3] * q->q[3]);

	q->q[0] *= iqlength;
	q->q[1] *= iqlength;
	q->q[2] *= iqlength;
	q->q[3] *= iqlength;

	q->qcachestatus = QC_INVALID;
}

/* quaternion_rotate: Set the angle of the quaternion @q to @angle.
 */

void			quaternion_rotate(quaternion *q, double angle)
{
	q->s = angle;
	quaternion_set_angle(q);
}

/* quaternion_invert: Invert the rotation direction of @q.
 */

void			quaternion_invert(quaternion *q)
{
	q->q[1] = -q->q[1];
	q->q[2] = -q->q[2];
	q->q[3] = -q->q[3];
	q->qcachestatus = QC_INVALID;
}

/* quaternion_combine: Combine to quaternions @*qa and @*qb and return the combined quaternion.
 * Note that the angles must be set in advance. 
 */

quaternion		quaternion_combine(quaternion *qa, quaternion *qb)
{
        quaternion res;

	res.qtype = Q_COMBINED;

        res.q[0] = qa->q[0] * qb->q[0] - qa->q[1] * qb->q[1] - qa->q[2] * qb->q[2] - qa->q[3] * qb->q[3];
        res.q[1] = qa->q[0] * qb->q[1] + qa->q[1] * qb->q[0] + qa->q[2] * qb->q[3] - qa->q[3] * qb->q[2];
        res.q[2] = qa->q[0] * qb->q[2] + qa->q[2] * qb->q[0] + qa->q[3] * qb->q[1] - qa->q[1] * qb->q[3];
        res.q[3] = qa->q[0] * qb->q[3] + qa->q[3] * qb->q[0] + qa->q[1] * qb->q[2] - qa->q[2] * qb->q[1];

	quaternion_normalize(&res);

        return res;
}


/* quaternion_update_matrix: Update the rotation matrix of @q. Rotation matrices are cached along
 * with the quaternion to avoid re-computing the matrix each time.
 */

void			quaternion_update_matrix(quaternion *q)
{
	if(q->qcachestatus == QC_VALID)
		return;

	q->mtx[0] = (1 - 2*q->q[2]*q->q[2] - 2*q->q[3]*q->q[3]);
	q->mtx[1] = (2*q->q[1]*q->q[2] - 2*q->q[0]*q->q[3]);
	q->mtx[2] = (2*q->q[0]*q->q[2] + 2*q->q[1]*q->q[3]);
	q->mtx[3] = 0;

	q->mtx[4] = (2*q->q[1]*q->q[2] + 2*q->q[0]*q->q[3]);
	q->mtx[5] = (1 - 2*q->q[1]*q->q[1] - 2*q->q[3]*q->q[3]);
	q->mtx[6] = (2*q->q[2]*q->q[3] - 2*q->q[0]*q->q[1]);
	q->mtx[7] = 0;

	q->mtx[8] = (2*q->q[1]*q->q[3] - 2*q->q[0]*q->q[2]);
	q->mtx[9] = (2*q->q[0]*q->q[1] + 2*q->q[2]*q->q[3]);
	q->mtx[10] = (1 - 2*q->q[1]*q->q[1] - 2*q->q[2]*q->q[2]);
	q->mtx[11] = 0;

	q->mtx[12] = 0;
	q->mtx[13] = 0;
	q->mtx[14] = 0;
	q->mtx[15] = 1;

	q->qcachestatus = QC_VALID;


}
