/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_script.c: Routines for parsing/executing script files
 * This file contains one function which executes a CPFW script, made by
 * the genscript program. The CPFW scripts consists of a sequence of instructions. For example,
 * m(0[0,0,0][20,20,20]) says that vertex (0,0,0) of mesh 0 should be moved to location (20,20,20).
 * See the comments of genscript.c for further information on the file format.
 * 
 * In this file, we also do timings and print benchmark results.
 */


#include <sys/time.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>
#include <mpi.h>

#include "fw_loadbalancer.h"
#include "fw_script.h"
#include "fw_voxel.h"
#include "fw_main.h"

// See fw_main.c for comments on these variables.

extern int griddims[2];
extern MPI_Comm cpfwcomm;
extern int size;
extern int ionode;


int n, m, k;				// Extents in x,y,z direction of 3D space, read from script file
native_voxel_space localvxspace;	// The local voxel space, storing all the voxels belonging to a specific node. Eg.  
					// we have a 3D space of (100,100,100), each node will have a local voxel space
					// of size 50^3. See also fw_voxel.c

voxel_cache *voxelcache = NULL;		// The voxel cache. This voxel cache is used to store voxels received from other nodes.
					// (As opposed to the native_voxel_space structure, which only stores voxels "naturally"
					// belonging to a node. The voxel cache is reset at each repeat but not at each
					// move-extract cycle.


/* Statistics variables accessed in fw_loadbalancer.c */
double *loadstdev;			// StdDev of load in each cycle
int cycles, cyclenum = 0;		// # of cycles / current cycle
int repeats, repeatnum = 0;		// # of repeats / current repeat

/* script_execute: This is the general script execution function, for executing the script @scriptfile. No algorithm is implemented here.
 * Instead, the algorithms are passed as function pointers to functions in fw_splitpolygons.c:
 *   - extractpoints is the Extract-All algorithm to use (minimum transfer, blocked transfer or bounding volumes)
 *   - addpolygon/removepolygon/movevertex point to algorithms in fw_splitpolygons.c (but are implemented as pointers so that future
 *     extensions to use other routines for add/remove/move may be implemented)
 */

void		script_execute(FILE *scriptfile, void (*extractpoints)(void), void (*addpolygon)(mesh *m, vertex *v0, vertex *v1, vertex *v2), void (*removepolygon)(mesh *m, vertex *v), void (*movevertex)(mesh *m, vertex *old, vertex *new))
{

// The following two macros start and stop the timer. We only time actual operations.

#define FW_TIMER_START do { tmptime1 = MPI_Wtime(); } while(0);
#define FW_TIMER_STOP  do { tmptime2 = MPI_Wtime(); wtimetotal += tmptime2 - tmptime1; } while(0);


	int avgsize, npolys, meshes;					// avg size/# of polys/#of meshes, from script file
	int addops = 0, remops = 0, moveops = 0, extractops = 0;	// # of add/moves/rem/cycles
	int meshidx = -1, meshidxtmp;					// current mesh

	vertex v0, v1, v2;
	mesh **meshlist = NULL;
	double tmptime1, tmptime2;
	double wtimetotal = 0.0, maxtime = 0.0, mintime = 0.0;


	extern int lbmode;		// load-balancing mode, from the command line
	extern int rank;		// rank of this process


	printf_zero("[cpfw/script] Reading script\n");

	if(fscanf(scriptfile, "%dx%dx%d:%d:%d:%d:%d:%d\n", &n, &m, &k, &avgsize, &npolys, &repeats, &meshes, &cycles) != 8) {
		printf_zero("[cpfw/script] parse error in script file: file header incorrect\n");
		return;
	}

	printf_zero("[cpfw/script] Script info: meshes(%d) # of polygons(%d) repeats(%d) cycles(%d)\n", meshes, npolys, repeats, cycles);

	if(n <= 0 || m <= 0 || k <= 0 || meshes <= 0 || cycles <= 0 || avgsize <= 0 || npolys <= 0 || repeats <= 0) {
		printf_zero("[cpfw/script] syntax error in script file. please use programs in scripts/ folder to generate.\n");
		return;
	}

	loadstdev = fw_malloc(sizeof(double) * cycles);
	memset(loadstdev, 0, sizeof(double) * cycles);

	// Pad, if necessary. 

	n += griddims[0] - (n % griddims[0]);
	m += griddims[1] - (m % griddims[1]);

	// Create initial voxel space of this node. */

	localvxspace = voxel_space_create(n, m, k, 1);
	voxelcache = voxelcache_create();

	// Reset loadbalancer state 
	loadbalance_reset();

	if(lbmode == LOADBALANCE_MANHATTAN_EXTRACT)
		loadbalance_manhattan_init();

	meshlist = fw_malloc(sizeof(mesh *) * meshes);

	/* All nodes: parse file. The parsing time is not a part of the times
	 * displayed
	 */

	printf_zero("[cpfw/script] Executing script...\n");

	while(!feof(scriptfile)) {
		// Read the first character which gives us the command
		int c = fgetc(scriptfile);
		switch(c) {
			case 'c':	// create mesh
					// NOTE! This is a bit different from the Create-Mesh operation
					// specified in the thesis, since that operation takes three vertices as arguments.
					// Create-Mesh must thus be seen as a 'c' followed by an 'a'. genscript generates
					// scripts which make the behaviour consistent with that described in the thesis.
					// See also comment above we_mesh_create() in fw_wingededge.c
				meshlist[++meshidx] = we_mesh_create();
				break;

			case 'a':	// add face
				if(fscanf(scriptfile, "([%d,%d,%d][%d,%d,%d][%d,%d,%d])", &v0.coords.x, &v0.coords.y, &v0.coords.z, &v1.coords.x, &v1.coords.y, &v1.coords.z, &v2.coords.x, &v2.coords.y, &v2.coords.z) < 9){
					printf_zero("[cpfw/script] parse error in create-mesh/add-face instruction\n");
					return;
				}

				FW_TIMER_START;
				addpolygon(meshlist[meshidx], &v0, &v1, &v2);
				FW_TIMER_STOP;

				++addops;
				break;
			case 'r':	// remove face(s)
				if(fscanf(scriptfile, "(%d[%d,%d,%d])", &meshidxtmp, &v0.coords.x, &v0.coords.y, &v0.coords.z) != 4) {
					printf_zero("[cpfw/script] parse error in remove-faces instruction\n");
					return;
				}

				FW_TIMER_START;
				removepolygon(meshlist[meshidxtmp], &v0);
				FW_TIMER_STOP;
				++remops;
				break;

			case 'm':	// move vertex
				if(fscanf(scriptfile, "(%d[%d,%d,%d][%d,%d,%d])", &meshidxtmp, &v0.coords.x, &v0.coords.y, &v0.coords.z, &v1.coords.x, &v1.coords.y, &v1.coords.z) != 7) {
					printf_zero("[cpfw/script] parse error in move-vertex instruction\n");
					return;
				}

				
				FW_TIMER_START;
				movevertex(meshlist[meshidxtmp], &v0, &v1);
				FW_TIMER_STOP;
				++moveops;

				break;

			case 'e':	// extract-all
				FW_TIMER_START;
				extractpoints();
				FW_TIMER_STOP;

				loadbalance_statistics_update();

				++extractops;
				++cyclenum;
				break;
			case 'd':	// destroy mesh, AFTER all vertices have been removed. we do not time this function but include the instruction to reduce memory usage
				if(fscanf(scriptfile, "(%d)", &meshidxtmp) != 1) {
					printf_zero("[cpfw/script] parse error in destroy-mesh instruction\n");
					return;
				}

				we_mesh_destroy(meshlist[meshidxtmp]);
				break;
			case '!':	// end current repeat				
				++repeatnum;
				cyclenum = 0;
				voxelcache_flushall(voxelcache);
				voxelcache = NULL;
				loadbalance_reset();
				printf_zero("[cpfw/script] ----------------------------------------------- repeat %d done.\n", repeatnum);
				voxelcache = voxelcache_create();
				voxelcache_resetcachemap(&localvxspace);
				meshidx = -1;

				break;
			case -1: /* EOF */
				break;
			default:
				printf_zero("[cpfw/script] parse error: unknown instruction '%c' in script file\n", c);
				return;
		}
	}

	fw_free(meshlist);

	printf_zero("[cpfw/script] Script execution done. Gathering benchmarks.\n\n");


	MPI_Allreduce(&wtimetotal, &maxtime, 1, MPI_DOUBLE, MPI_MAX, cpfwcomm);
	MPI_Allreduce(&wtimetotal, &mintime, 1, MPI_DOUBLE, MPI_MIN, cpfwcomm);

	// Compute avg/min/max add/remove/move/extract times
	if(ionode == 1) {
		double avgstdev = 0.0;

		int i;
			
		// Note: loadstdev-array is accessed in fw_loadbalancer.c

		for(i = 0; i < cycles; i++) {

			avgstdev += loadstdev[i];

		}

		avgstdev /= cycles;


		printf_zero("Benchmark results\n");
		printf_zero("-----------------\n");
		printf_zero("--- Add opertions...............................: %d\n", addops); 
		printf_zero("--- Remove opertions............................: %d\n", remops); 
		printf_zero("--- Move opertions..............................: %d\n", moveops); 
		printf_zero("--- Extract-All/compute opertions...............: %d\n", extractops); 
		printf_zero("--- Slowest node took...........................: %f  Per Repeat:  %f\n", maxtime, maxtime/repeats);
		printf_zero("--- Quickest node took..........................: %f  Per Repeat:  %f\n", mintime, mintime/repeats);
		printf_zero("--- Average load StDev over all cycles..........: %f\n", avgstdev);

	}


	fw_free(loadstdev);
}
