/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_splitpolygons.c: This file implements the three strategies for the Extract-All operation:
 *  - Minimum Communication (splitpolygons_extractpoints_minimumtransfer)
 *  - Blocked Transfer (splitpolygons_extractpoints_blockedtransfer)
 *  - Bounding Volumes (splitpolygons_extractpoints_boundingvolumes)
 *
 * This file contains no MPI communication routines. MPI communication is done in the file
 *    fw_voxeltransfer.c  for splitpolygons_minimumtransfer/blockedtransfer (which share the same communication routines)
 * and
 *    fw_voxelcachetransfer.c  for splitpolygons_boundingvolumes (named 'cachetransfer' for historical reasons).
 *
 * Note that all the strategies use voxelcaches, defined in fw_voxel.c.
 * 
 * This function also contains routines for adding/removing/moving polygons, i.e. Add-Face, Remove-Faces and Move-Vertex operations. 
 * These routines will again call lower-level functions in fw_wingededge.c to 
 * do operations on the polygon mesh, but also computes centroid nodes and calls load-balancing statistics functions.
 */

#include <mpi.h>

#include "fw_main.h"
#include "fw_loadbalancer.h"
#include "fw_wingededge.h"
#include "fw_voxel.h"
#include "fw_voxeltransfer.h"
#include "fw_voxelcachetransfer.h"
#include "fw_extract_cpu.h"
#include "fw_lot.h"

extern int lbmode;			// Load-balancing mode, from fw_main

extern int cycles, cyclenum;		// Current cycle/# of cycles, from fw_script
extern int repeats, repeatnum;		// Repats/# of repeats, from fw_script


/* splitpolygons_addpolygon: Add polygon with vertices @v0,@v1,@v2 to mesh @m0.
 * This is a wrapper function for we_add_polygon in fw_wingededge.c,
 * but this routine computes the centroid node of the polygon and calls load-balancing
 * statistics functions.
 *
 * This function can be viewed as the Add-Face operation (see thesis Chapter 3).
 */

void	splitpolygons_addpolygon(mesh *m0, vertex *v0, vertex *v1, vertex *v2)
{
	coordinate centroid;
	face *f = we_add_polygon(m0, v0, v1, v2);

	// error, ignore (more info printed by we_add_polygon)
	if(!f)
		return;

	// Compute centroid and figure out which node this polygon belongs to
	centroid = coordinate_scale_float(coordinate_add(coordinate_add(v0->coords, v1->coords), v2->coords), 1.0/3.0);
	f->rnode = coordinate_to_rank(centroid);
	f->area = we_triangle_facearea(f);

	if(lbmode != LOADBALANCE_NONE) {
		loadbalance_hook_addpolygon(f);
	}

}


/* splitpolygons_removepolygon: Remove polygons incident to vertex @v
 * in mesh @m0. This can be viewed as the Remove-Faces operation (see thesis Chap. 3)
 */


void	splitpolygons_removepolygon(mesh *m0, vertex *v)
{
	if(lbmode != LOADBALANCE_NONE) {
		loadbalance_hook_removevertex(we_vertex_find(m0, v));
	}

	we_remove_polygons(m0, v);

}


/* splitpolygons_movevertex: Move vertex at @old to location @new.
 * Also we update the centroid node of the triangles connected to the vertex.
 *
 * This can be viewed as the Move-Vertex operation described in the thesis Chapter 3.
 */


void	splitpolygons_movevertex(mesh *m0, vertex *old, vertex *new)
{
	int i;
	vertex *newv = we_move_vertex(m0, old, new);


	// Script error
	if(!newv)
		return;


	if(lbmode != LOADBALANCE_NONE)
		loadbalance_hook_movevertex_pre(newv);

	// After moving, we also need to compute new centroid nodes!
	// Also, the area of connecting faces needs to be re-computed.

	for(i = 0; i < newv->nfaces; i++) {
		face *f = newv->faces[i];
		coordinate centroid = coordinate_scale_float(coordinate_add(coordinate_add(f->vertices[0]->coords, f->vertices[1]->coords), f->vertices[2]->coords), 1.0/3.0);
		f->rnode = coordinate_to_rank(centroid);	
		f->area = we_triangle_facearea(f);
	}

	if(lbmode != LOADBALANCE_NONE)
		loadbalance_hook_movevertex_post(newv);

}

/* splitpolygons_extractpoints_boundingvolumes: Implements the Extract-All operation
 * through the Bounding Volumes strategy. See thesis Chapter 5.
 */

void	splitpolygons_extractpoints_boundingvolumes(void)
{
	int i, j;
	float avg;
	coordinate min, max;
	face *f;
	mesh *m;

	/* In the BV strategy, each node knows precisely what should be 
	 * received and sent from/to other nodes. We use a "receivequeue"
	 * and a "sendqueue" for receive/send buffers. Functions for manipulating
	 * these are located in fw_voxelcachetransfer.c
	 */

	voxelcache_receivequeue *vc_r = voxelcache_receivequeue_create();
	voxelcache_sendqueue *vc_s = voxelcache_sendqueue_create();
	

	int nfaces = 0, facecount = 0;		// # of faces this node is responsible for, and a temp. variable

	coordinate_list **clists;		// this is an array of size nfaces containing the coordinate lists resulting from
						// the voxelization of all the faces this node is responsible for.

	extern int nmeshes;
	extern mesh *meshlist;
	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	extern int rank, size;
	extern MPI_Comm cpfwcomm;

	// Execute load-balancing routines (if any)

	if(lbmode == LOADBALANCE_MANHATTAN_EXTRACT)
		loadbalance_manhattanextract();
	else if(lbmode == LOADBALANCE_GLOBAL_EXTRACT)
		loadbalance_globalextract();
	else if(lbmode == LOADBALANCE_LOCAL_EXTRACT)
		loadbalance_localextract();

	// (See Figure 3.7) First Step. Figure out which voxels must be sent to other nodes.

	for(i = 0; i < size; i++) {
		if(i != rank) {
			vc_s->sendbuffers[i].sendexpect = 1;	// Expecting to send/receive data from all other nodes. Simplifies communication; see fw_voxelcachetransfer.c. 
			vc_r->recvfromnodes[i]++;
		}
	}

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank)
				++nfaces;

			min = we_coordinate_min(f);
			max = we_coordinate_max(f);

			// No intersection, or no transfer necessary (entire triangle is on our node)?
			if((f->rnode == rank && max.x < localvxspace.Xend && min.x >= localvxspace.Xstart && max.y < localvxspace.Yend && min.y >= localvxspace.Ystart) 
				|| (f->rnode != rank && (max.x < localvxspace.Xstart || min.x >= localvxspace.Xend || max.y < localvxspace.Ystart || min.y >= localvxspace.Yend)))
				continue;

			// Add the coordinates intersecting this node's coordinate space to the send queue, if this node is not responsible for the face.
			if(f->rnode != rank) {
				voxelcache_sendqueue_add(vc_s, min, max, f->rnode);
			}
		
		}
	}

	// Second Step. Start exchange.

	voxelcache_exchange_start(vc_r, vc_s); // Exchanges sizes and sends actual data



	// Third Step. Voxelization
	clists = fw_malloc(sizeof(coordinate_list *) * nfaces);

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank) {
				clists[facecount++] = extract_points(f->vertices[0], f->vertices[1], f->vertices[2]);
			}
		}
	}



	// Fourth Step. Finalize transfers and do computations.

	voxelcache_exchange_finalize(vc_r, vc_s);





	double faceavg = 0.0, facevar = 0.0;
	FILE *devnull = fopen("/dev/null", "w"); // we need somewhere to write our data so the compiler does not optimize away the following loop!

	for(i = 0; i < nfaces; i++) {

		faceavg = 0.0;
		facevar = 0.0;

		// Compute variance from the averages, and also compute the LOT

		int lotcounter = 0;

		
		// Array of LOT input/output values

		float *lot_in_array = fw_malloc(sizeof(float) * clists[i]->used);
		float *lot_out_array = fw_malloc(sizeof(float) * clists[i]->used);

		
		// Compute average values (used in variance computation) and initialize LOT input array

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			voxel v = voxel_lookup(&localvxspace, voxelcache, c);
	
			if(v.voxeltype == VOXEL_NATIVE) {
				faceavg += v.vox.nvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.nvox->data;
			} else if(v.voxeltype == VOXEL_CACHED) {
				faceavg += v.vox.cvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.cvox->data;
			}

		}

		// Execute LOT (see fw_lot.c)

		cpu_flot_32(lot_in_array, lot_out_array, lotcounter);

		lotcounter = 0;

		// Compute variance and write LOT values to /dev/null (in order to avoid compiler optimizing computation away)

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			
			if(clists[i]->used == 1) { // varaince is zero
				facevar = 0;
				break;
			}

			voxel v = voxel_lookup(&localvxspace, voxelcache, c);

			if(v.voxeltype == VOXEL_NATIVE) {
				facevar += (v.vox.nvox->data - faceavg) * (v.vox.nvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			} else if(v.voxeltype == VOXEL_CACHED) {
				facevar += (v.vox.cvox->data - faceavg) * (v.vox.cvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			}

		}

		// Write variance also to /dev/null, and free data structures.

		fprintf(devnull, "%f", facevar);
		
		fw_free(lot_in_array);
		fw_free(lot_out_array);

		coordinate_list_free(clists[i]);
	}


	// Close/free devnull-file and coordinate lists

	fclose(devnull);

	fw_free(clists);

	voxelcache_receivequeue_free(vc_r);
	voxelcache_sendqueue_free(vc_s);

}

/* splitpolygons_extractpoints_minimumtransfer: Implements the Extract-All operation
 * through the Minimum Communication strategy. See thesis Chapter 5.
 */

void		splitpolygons_extractpoints_minimumtransfer(void)
{
	int i, j;
	int facecount = 0, nfaces = 0;	// nfaces = # of faces this node is responsible for


	/* We use a "receivequeue"
	 * and a "sendqueue" for receive/send buffers. Functions for manipulating
	 * these are located in fw_voxeltransfer.c
	 */

	voxeltransfer_receivequeue *rq = voxeltransfer_receivequeue_create();
	voxeltransfer_sendqueue *sq = voxeltransfer_sendqueue_create();

	coordinate_list **clists;		// this is an array of size nfaces containing the coordinate lists resulting from
						// the voxelization of all the faces this node is responsible for.

	mesh *m;
	face *f;

	extern mesh *meshlist;
	extern int rank, size;

	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	// Do load-balancing (if any)

	if(lbmode == LOADBALANCE_MANHATTAN_EXTRACT)
		loadbalance_manhattanextract();
	else if(lbmode == LOADBALANCE_GLOBAL_EXTRACT)
		loadbalance_globalextract();
	else if(lbmode == LOADBALANCE_LOCAL_EXTRACT)
		loadbalance_localextract();

	// (See Figure 3.5) First Step. Voxelize all faces node is responsible for and figure out which voxels must be sent to other nodes.
	
	for(i = 0; i < size; i++) {
		if(i != rank) {
			sq->sendbuffers[i].sendexpect = 1;		// Simplifies communication - see fw_voxeltransfer.c
			rq->recvfromnodes[i]++;
		}
	}

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank)
				++nfaces;
		}
	}

	clists = fw_malloc(sizeof(coordinate_list *) * nfaces);

	// Do voxelization

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank) {
				clists[facecount] = extract_points(f->vertices[0], f->vertices[1], f->vertices[2]);
				// For each coordinate returned from voxelization, the node which has the coordinate space where the coordinate belongs is computed by coordinate_to_rank.
				// Then, the coordinate is added to the send queue to that node. 
				for(i = 0; i < clists[facecount]->used; i++) {
					int crank = coordinate_to_rank(clists[facecount]->coords[i]);
					if(sq->sendbuffers[crank].sendexpect && voxel_lookup(&localvxspace, voxelcache, clists[facecount]->coords[i]).voxeltype == VOXEL_NOTFOUND) {
						voxeltransfer_sendqueue_add_coordinate(sq, &clists[facecount]->coords[i], crank);
					}
				}
				++facecount;
			}
		}
	}


	// Second/third Step. Exchange voxels

	voxeltransfer_exchange(rq, sq);

	// Receive/send queues can now be freed.

	voxeltransfer_sendqueue_free(sq);
	voxeltransfer_receivequeue_free(rq);


	// Fourth Step. Do computations.


	double faceavg = 0.0, facevar = 0.0;
	FILE *devnull = fopen("/dev/null", "w"); // we need somewhere to write our data so the compiler does not optimize away the following loop!

	for(i = 0; i < nfaces; i++) {

		faceavg = 0.0;
		facevar = 0.0;

		// Compute variance from the averages, and also compute the LOT

		int lotcounter = 0;

		// Array of LOT input/output values
		

		float *lot_in_array = fw_malloc(sizeof(float) * clists[i]->used);
		float *lot_out_array = fw_malloc(sizeof(float) * clists[i]->used);

		
		// Compute average values (used in variance computation) and initialize LOT input array

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			voxel v = voxel_lookup(&localvxspace, voxelcache, c);
	
			if(v.voxeltype == VOXEL_NATIVE) {
				faceavg += v.vox.nvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.nvox->data;
			} else if(v.voxeltype == VOXEL_CACHED) {
				faceavg += v.vox.cvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.cvox->data;
			}

		}

		// Execute LOT (see fw_lot.c)

		cpu_flot_32(lot_in_array, lot_out_array, lotcounter);


		// Compute variance and write LOT values to /dev/null (in order to avoid compiler optimizing computation away)

		lotcounter = 0;

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			
			if(clists[i]->used == 1) { // varaince is zero
				facevar = 0;
				break;
			}

			voxel v = voxel_lookup(&localvxspace, voxelcache, c);

			if(v.voxeltype == VOXEL_NATIVE) {
				facevar += (v.vox.nvox->data - faceavg) * (v.vox.nvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			} else if(v.voxeltype == VOXEL_CACHED) {
				facevar += (v.vox.cvox->data - faceavg) * (v.vox.cvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			}

		}

		// Write variance also to /dev/null, and free data structures.



		fprintf(devnull, "%f", facevar);
		
		fw_free(lot_in_array);
		fw_free(lot_out_array);

		coordinate_list_free(clists[i]);
	}


	// Close/free devnull-file and coordinate lists


	fclose(devnull);


	fw_free(clists);

}


/* splitpolygons_extractpoints_minimumtransfer: Implements the Extract-All operation
 * through the Blocked Transfer strategy. See thesis Chapter 5.
 *
 * Note: This is very similar to splitpolygons_extractpoints_minimumtransfer() above.
 * The only difference is that
 * 	voxeltransfer_sendqueue_add_coordinate_blocked
 * instead of
 *	voxeltransfer_sendqueue_add_coordinate.
 *
 * is called when adding coordinates to the send queue. Also, a temporary voxel cache is created
 * to keep track of the voxels which have already been added. This has to be done since
 * when adding coordinates in blocks, adding one coordinate may add other coordinates which
 * are a part of the triangle's voxelization.
 */


void		splitpolygons_extractpoints_blockedtransfer(void)
{
	int i, j;
	int facecount = 0, nfaces = 0;		// nfaces = # of faces this node is responsible for

	voxel_cache *tmpvoxelcache = voxelcache_create();

	/* We use a "receivequeue"
	 * and a "sendqueue" for receive/send buffers. Functions for manipulating
	 * these are located in fw_voxeltransfer.c
	 */


	voxeltransfer_receivequeue *rq = voxeltransfer_receivequeue_create();
	voxeltransfer_sendqueue *sq = voxeltransfer_sendqueue_create();

	coordinate min, max;

	coordinate_list **clists;		// this is an array of size nfaces containing the coordinate lists resulting from
						// the voxelization of all the faces this node is responsible for.

	mesh *m;
	face *f;

	extern mesh *meshlist;
	extern int rank, size;

	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	// Execute load-balancing (if any)

	if(lbmode == LOADBALANCE_MANHATTAN_EXTRACT)
		loadbalance_manhattanextract();
	else if(lbmode == LOADBALANCE_GLOBAL_EXTRACT)
		loadbalance_globalextract();
	else if(lbmode == LOADBALANCE_LOCAL_EXTRACT)
		loadbalance_localextract();




	// (See Figure 3.5) First Step. Voxelize all faces node is responsible for and figure out which voxels must be sent to other nodes.
	
	for(i = 0; i < size; i++) {
		if(i != rank) {
			sq->sendbuffers[i].sendexpect = 1;		// Simplifies communication - see fw_voxeltransfer.c
			rq->recvfromnodes[i]++;
		}
	}

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank)
				++nfaces;
		}
	}


	clists = fw_malloc(sizeof(coordinate_list *) * nfaces);

	// Do voxelization

	for(m = meshlist; m != NULL; m = m->next) {
		for(f = m->faces; f != NULL; f = f->next) {
			if(f->rnode == rank) {
				clists[facecount] = extract_points(f->vertices[0], f->vertices[1], f->vertices[2]);

				// For each coordinate returned from voxelization, the node which has the coordinate space where the coordinate belongs is computed by coordinate_to_rank.
				// Then, the coordinate AND surrounding coordinates in the block (which are located on this node) are added to the send queue to that node. 

				for(i = 0; i < clists[facecount]->used; i++) {
					int crank = coordinate_to_rank(clists[facecount]->coords[i]);
					if(sq->sendbuffers[crank].sendexpect && voxel_lookup(&localvxspace, voxelcache, clists[facecount]->coords[i]).voxeltype == VOXEL_NOTFOUND) {
						voxeltransfer_sendqueue_add_coordinate_blocked(sq, &clists[facecount]->coords[i], crank, tmpvoxelcache);
					}
				}
				++facecount;
			}
		}
	}

	// Free the temporary voxel cache

	voxelcache_flushall(tmpvoxelcache);

	// Second/third Step. Exchange voxels

	voxeltransfer_exchange(rq, sq);

	// Receive/send queues can now be freed.

	voxeltransfer_sendqueue_free(sq);
	voxeltransfer_receivequeue_free(rq);


	// Fourth Step. Do computations.


	double faceavg = 0.0, facevar = 0.0;
	FILE *devnull = fopen("/dev/null", "w"); // we need somewhere to write our data so the compiler does not optimize away the following loop!

	for(i = 0; i < nfaces; i++) {

		faceavg = 0.0;
		facevar = 0.0;

		// Compute variance from the averages, and also compute the LOT

		int lotcounter = 0;

		// Array of LOT input/output values		

		float *lot_in_array = fw_malloc(sizeof(float) * clists[i]->used);
		float *lot_out_array = fw_malloc(sizeof(float) * clists[i]->used);

		// Compute average values (used in variance computation) and initialize LOT input array

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			voxel v = voxel_lookup(&localvxspace, voxelcache, c);
	
			if(v.voxeltype == VOXEL_NATIVE) {
				faceavg += v.vox.nvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.nvox->data;
			} else if(v.voxeltype == VOXEL_CACHED) {
				faceavg += v.vox.cvox->data/clists[i]->used;
				lot_in_array[lotcounter++] = v.vox.cvox->data;
			}

		}


		// Execute LOT (see fw_lot.c)

		cpu_flot_32(lot_in_array, lot_out_array, lotcounter);

		lotcounter = 0;

		// Compute variance and write LOT values to /dev/null (in order to avoid compiler optimizing computation away)

		for(j = 0; j < clists[i]->used; j++) {
			coordinate c = clists[i]->coords[j];
			
			if(clists[i]->used == 1) { // varaince is zero
				facevar = 0;
				break;
			}

			voxel v = voxel_lookup(&localvxspace, voxelcache, c);

			if(v.voxeltype == VOXEL_NATIVE) {
				facevar += (v.vox.nvox->data - faceavg) * (v.vox.nvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			} else if(v.voxeltype == VOXEL_CACHED) {
				facevar += (v.vox.cvox->data - faceavg) * (v.vox.cvox->data - faceavg) / (clists[i]->used - 1);
				fprintf(devnull, "%f", lot_out_array[lotcounter++]);
			}

		}

		// Write variance also to /dev/null, and free data structures.

		fprintf(devnull, "%f", facevar);
		
		fw_free(lot_in_array);
		fw_free(lot_out_array);

		coordinate_list_free(clists[i]);
	}



	// Close/free devnull-file and coordinate lists


	fclose(devnull);

	fw_free(clists);

}


