/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_vector.h: Vector structures/operations. Note that there is no .c file
 * as we inline all vector operations. These operations are fairly simple,
 * and are therefore suitable for inlining.
 */
#ifndef FW_VECTOR
#define FW_VECTOR

typedef struct vector {
	double x, y, z;
} vector;



#include "fw_coord.h"
#include "fw_quaternion.h"
#include <stdio.h>
#include <math.h>

/* Unit vectors in x/y/z directions
 */

#define VECTOR_UNIT_X	(vector_create_components(1.0, 0.0, 0.0))
#define VECTOR_UNIT_Y	(vector_create_components(0.0, 1.0, 0.0))
#define VECTOR_UNIT_Z	(vector_create_components(0.0, 0.0, 1.0))


/* vector_create: Create a new vector with coordinates @c
 * Returns a new vector object.
 */

static inline vector vector_create(coordinate *c)
{
	vector v;

	v.x = c->x;
	v.y = c->y;
	v.z = c->z;
	
	return v;
}

/* vector_create_components: Create a new vector from the specified @x, @y, @z components. 
 * Returns a new vector object.
 */

static inline vector vector_create_components(double x, double y, double z)
{
	vector v;

	v.x = x;
	v.y = y;
	v.z = z;
	
	return v;
}


/* vector_to_coordinate: Convert the vector @v to coordinates @c. The components of @v are always rounded down.
 * If rounding is needed, use vector_round.
 * Returns a new coordinate object.
 */

static inline coordinate vector_to_coordinate(vector v)
{
	struct coordinate res = { (int) v.x, (int) v.y, (int) v.z };
	return res;
}

/* vector_equal: Check if vectors @v1 and @v2 are equal.
 * Note: Vectors should be rounded prior to calling this routine,
 * due to the instability of the floating-point == operator.
 * Returns 1 if the vectors are equal, 0 if not.
 */

static inline int vector_equal(vector v1, vector v2)
{
	return v1.x == v2.x && v1.y == v2.y && v1.z == v2.z;
}

/* vector_round: Round vector @v = (x,y,z), returning a new vector
 * (round(x), round(y), round(z)).
 */

static inline vector vector_round(vector v)
{
	vector res;

	res.x = round(v.x);
	res.y = round(v.y);
	res.z = round(v.z);

	return res;
}

/* vector_floor: Floor the components of @v, returning a new vector.
 * For negative values we use floor(x) = ceil(x). Thus this is not
 * the mathematical floor function. We want floor to give an "underestimated" value.
 * Consider if we have a vector v = (10.1, -0.9, 0.0). Then vector_floor(v) gives
 * (10, 0, 0), while the mathematical floor would give (10, -1, 0).
 */

static inline vector vector_floor(vector v)
{
	vector res;

	res.x = v.x > 0.0 ? floor(v.x) : ceil(v.x);
	res.y = v.y > 0.0 ? floor(v.y) : ceil(v.y);
	res.z = v.z > 0.0 ? floor(v.z) : ceil(v.z);

	return res;
}

/* vector_ceil: Return a new vector whose components are @v's components ceiled.
 */


static inline vector vector_ceil(vector v)
{
	vector res;

	res.x = ceil(v.x);
	res.y = ceil(v.y);
	res.z = ceil(v.z);

	return res;
}


/* vector_coordinate_equal: Check if vector @v1 and @cv1 are equal, when rounded to integers.
 */

static inline int vector_coordinate_equal(vector v1, vector cv1)
{
	return ((int) v1.x == (int) cv1.x) && ((int) v1.y == (int) cv1.y) && ((int) v1.z == (int) cv1.z);	
}

/* vector_add: Add vectors @v1 and @v2, returning new vector object.
 */


static inline vector vector_add(vector v1, vector v2)
{
	vector res;

	res.x = v1.x + v2.x;
	res.y = v1.y + v2.y;
	res.z = v1.z + v2.z;

	return res;
}

/* vector_addf: Add vectors @v1 and @v2, returning new vector object, but use only
 * float (instead of double) precision. Used
 * when doing voxelization benchmarks to ensure fair comparison between CPU and GPU.
 */


static inline vector vector_addf(vector v1, vector v2)
{
	vector res;

	res.x = ((float) v1.x) + ((float) v2.x);
	res.y = ((float) v1.y) + ((float) v2.y);
	res.z = ((float) v1.z) + ((float) v2.z);

	return res;
}

/* vector_reverse: Reverse vector @v, returning the reversed vector.
 */

static inline vector vector_reverse(vector v)
{
	vector res;

	res.x = -v.x;
	res.y = -v.y;
	res.z = -v.z;

	return res;
}

/* vector_sub: Subtract @v2 from @v1, returning resulting vector.
 */

static inline vector vector_sub(vector v1, vector v2)
{
	vector res;

	res.x = v1.x - v2.x;
	res.y = v1.y - v2.y;
	res.z = v1.z - v2.z;

	return res;
}

/* vector_normalize: Normalize vector @v, returning the normalized vector.
 */

static inline vector vector_normalize(vector v)
{
	double nfac = 1.0/sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
	vector res;

	res.x = v.x * nfac;
	res.y = v.y * nfac;
	res.z = v.z * nfac;

	return res;
}

/* vector_cross: Compute and return the cross product of @v1 and @v2.
 */

static inline vector vector_cross(vector v1, vector v2)
{
	vector res;

	res.x = v1.y * v2.z - v1.z * v2.y;
	res.y = v1.z * v2.x - v1.x * v2.z;
	res.z = v1.x * v2.y - v1.y * v2.x;

	return res;
}

/* vector_parallel: Returns 1 if @v1 and @v2 are (approximately) parallel,
 * 0 if not.
 */

static inline int vector_parallel(vector v1, vector v2)
{
	return vector_coordinate_equal(vector_round(vector_cross(v1, v2)), vector_create_components(0.0, 0.0, 0.0));
}

/* vector_dot: Compute and return the dot product of @v1 and @v2.
 */

static inline double vector_dot(vector v1, vector v2)
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}


/* vector_scale: Scale @v by @c, and return the resulting vector.
 */

static inline vector vector_scale(vector v, double c)
{
	vector res;

	res.x = v.x * c;
	res.y = v.y * c;
	res.z = v.z * c;

	return res;
}

/* vector_length: Compute and return the length of vector @v.
 */

static inline double vector_length(vector v)
{
	return sqrt(vector_dot(v, v));
}

/* vector_rotate: Rotate vector @v about @q.
 * Essentially, the rotation amounts to 4x4 vector-matrix multiplication.
 */

static inline vector vector_rotate(vector v, quaternion *q)
{
	vector res;

	// Update quaternion's rotation matrix

	quaternion_update_matrix(q);

	res.x = q->mtx[0] * v.x + q->mtx[1] * v.y + q->mtx[2] * v.z;
	res.y = q->mtx[4] * v.x + q->mtx[5] * v.y + q->mtx[6] * v.z;
	res.z = q->mtx[8] * v.x + q->mtx[9] * v.y + q->mtx[10] * v.z;

	return res;	
}


/* vector_rotatef: Same as vector_rotate, but with float-precision. Used
 * when doing voxelization benchmarks to ensure fair comparison between CPU and GPU.
 */

static inline vector vector_rotatef(vector v, quaternion *q)
{
	vector res;

	quaternion_update_matrix(q);


	res.x = ((float) q->mtx[0]) * ((float) v.x) + ((float)q->mtx[1]) * ((float)v.y) + ((float)q->mtx[2]) * ((float) v.z);
	res.y = ((float) q->mtx[4]) * ((float) v.x) + ((float)q->mtx[5]) * ((float)v.y) + ((float)q->mtx[6]) * ((float) v.z);
	res.z = ((float) q->mtx[8]) * ((float) v.x) + ((float)q->mtx[9]) * ((float)v.y) + ((float)q->mtx[10]) * ((float) v.z);

	return res;	
}

/* vector_angle: Compute and return angle between @v1 and @v2.
 */ 

static inline double vector_angle(vector v1, vector v2)
{
	double result = acos( vector_dot(v1, v2) / (vector_length(v1) * vector_length(v2)) );

	// In some cases with numbers very near zero, we may get nan as a result. This should be Pi/2. 
	return isnan(result) ? M_PI_2 : result; 
}


/* vector_print: Debugging routine for printing @v to stdout with prefix @prefix.
 */

static inline void vector_print(const char *prefix, vector v)
{
	printf("%s: [%.5f, %.5f, %.5f]\n", prefix, v.x, v.y, v.z);
}



#endif

