/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */
/* fw_voxel.c: Voxel space and voxel caches
 * This file contains structures to keep track of voxels stored at a node.
 * We separate between two types of voxels:
 * - 'native', which are voxels in the node's coordinate space
 * - 'cached', which are voxels that have been received from other nodes as a consequence of
 *             executing the caching strategies.
 *
 * Native voxels are stored in the native_voxel_space structure, while cached voxels
 * are stored in the voxel_cache_structure. Details are described below. See also fw_voxel.h.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fw_main.h"
#include "fw_voxel.h"



#include <mpi.h>

// From fw_main.c

extern int rank;			// This node's rank
extern int griddims[2];		// Extents in x,y dimensions
extern int nodecoords[2];		// This node's coordinates
extern int singlemode;			// Single-CPU mode or not

// MPI data types for cached voxels/coordinates

MPI_Datatype cached_voxel_mpit = NULL;		// used when transmitting voxel data
MPI_Datatype coordinate_mpit = NULL;		// used when transmitting coordinates (in the MT/BT modes)

extern MPI_Comm cpfwcomm;

extern voxel_cache *voxelcache;		// global voxel cache, reset at each repeat but
						// kept between move-extract cycles


/* voxel_space_create: Create native_voxel_space. @n,@m,@k denote the
 * extents of the 3D space in x,y,z directions. We assumes that n,m,k has been appropriately padded (this
 * is done in fw_script.c). Returns a new native_voxel_space object which has allocated the native voxels
 * of this node. (With p=ab processors, this will be of size (n/a) * (m/b) * k).
 *
 * Native voxels also have a cache map, which indicates to which nodes the voxel has been sent. This is 
 * implemented as a bit mask associated with each voxel. For example, if we are on node 5 and node 5 has
 * transmitted a native voxel to node 8, bit number 8 in the voxel's cache map on node 5 has been set.
 *
 * This creates a problem with > 32 nodes. In that case, we use multiple cache maps and SCALABLE_CACHE
 * must be defined when compiling.
 */


native_voxel_space	voxel_space_create(int n, int m, int k, int payload)
{
	int i, x, y, z;
	native_voxel_space res;


#ifdef SCALABLE_CACHE
	int rank_cachemapidx, rank_cacheidx;
	int ulsz = 8 * sizeof(unsigned long);
	rank_cachemapidx = rank / ulsz;
	rank_cacheidx = rank % ulsz;
	
	assert(rank_cachemapidx < 8);
#endif

	res.Xstart = nodecoords[0] * (n/griddims[0]);
	res.Ystart = nodecoords[1] * (m/griddims[1]);
	res.Zstart = 0;

	res.Xend = (nodecoords[0] + 1) * (n/griddims[0]);
	res.Yend = (nodecoords[1] + 1) * (m/griddims[1]);
	res.Zend = k;

	res.payload = payload;
	res.spacesize = (res.Xend - res.Xstart) * (res.Yend - res.Ystart) * (res.Zend - res.Zstart);

	printf_zero("[cpfw/voxel_space_create] Allocating %d voxels per node. Padded voxel space extends from (0, 0, 0) -> (%d, %d, %d) (total size: %d). Wait while allocating...\n", res.spacesize, n, m, k, n * m * k);
	res.voxellist = fw_malloc(sizeof(struct native_voxel) * res.spacesize);



	// Create initial voxels and cache map

	/* res.vxmap[i,j,k] is global coords (i+xstart,j+ystart,k) */

	for(x = res.Xstart; x < res.Xend; x++) {
		for(y = res.Ystart; y < res.Yend; y++) {
			for(z = res.Zstart; z < res.Zend; z++) {
				native_voxel v;
				coordinate c = { x, y, z };

				v.data = random()/(float)RAND_MAX;

#ifdef SCALABLE_CACHE
				v.cachemaps[rank_cachemapidx] |= (1 << rank_cacheidx);
#else
				v.cachemap = (1 << rank);
#endif

				*native_voxel_access(&res, c) = v;
			}
		}
	}

	if(!singlemode)
		MPI_Barrier(cpfwcomm);

	printf_zero("[cpfw/voxel_space_create] Allocation done. Initializing MPI data types\n");

	/* Generate MPI data types, if not already generated.
	 * For the coordinate we use a struct of three ints (see fw_coord.c/coordinate_build_mpitype). For the voxel data
	 * we use a coordinate in addition to a float. 
	 * See fw_voxelcachetransfer.c for further details.
	 */

	if(!cached_voxel_mpit) {
		MPI_Aint	displacements[2];
		MPI_Datatype	types[2];
		int		blocklengths[2];
		MPI_Datatype	floatarray;
		cached_voxel	vc;


		MPI_Aint	startaddr, tmpaddr;

		// First build coordinate data type
		coordinate_build_mpitype( &coordinate_mpit );

		blocklengths[0] = blocklengths[1] = 1;

		types[0] = coordinate_mpit;
		types[1] = MPI_FLOAT;

		displacements[0] = 0;

		MPI_Address(&vc.coord, &startaddr);
		MPI_Address(&vc.data, &tmpaddr);

		displacements[1] = tmpaddr - startaddr;

		MPI_Type_struct( 2, blocklengths, displacements, types, &cached_voxel_mpit );
		MPI_Type_commit( &cached_voxel_mpit );
	}

	printf_zero("[cpfw/voxel_space_create] MPI data types initialized.\n");

	return res;
}


/* voxel_hash_coord: Hash coordinate @c, returning the hash value in @vc
 */

static int voxel_hash_coord(voxel_cache *vc, coordinate c)
{
	return (c.x * c.y * c.z) % vc->hashsize;
}

/* voxel_hash: Wrapper for voxel_hash_coord(vc, v.coord)
 */

static int voxel_hash(voxel_cache *vc, cached_voxel v)
{
	return voxel_hash_coord(vc, v.coord);
}

/* voxel_lookup: Look up coordinate @c in the native voxel space @nvs or in the voxel
 * cache @vc. This can be viewed as computing V(c) (see thesis).
 * A hash table is used for efficient lookups.
 * 
 * A voxel is returned. If the voxel is not found, the voxeltype field is
 * set to VOXEL_NOTFOUND. If the voxel is cached, VOXEL_CACHED is set, and if
 * the voxel is native, VOXEL_NATIVE is set.
 *
 * If the voxel is native, voxel.nvox points to the native_voxel entry in @nvs.
 * The actual data is then voxel.nvox->data. Similarly, if the voxel is cached,
 * voxel.vox.cvox points to the corresponding cached_voxel structure in @vc. In this case,
 * the actual data is located in voxel.vox.cvox->data.
 *
 * Note that 'vox' is defined as (see fw_voxel.h):
 *	union vox { 
 *		native_voxel *nvox;
 *		cached_voxel *cvox;
 *	} vox;
 *
 */

voxel		voxel_lookup(native_voxel_space *nvs, voxel_cache *vc, coordinate c)
{
	int crank = coordinate_to_rank(c);
	voxel result;

	result.voxeltype = VOXEL_NOTFOUND;
	result.vox.nvox = NULL;
	result.vox.cvox = NULL;


	extern int rank;
	extern int n, m, k;

	if(c.x < 0 || c.y < 0 || c.z < 0 || c.x >= n || c.y >= m || c.z >= k)
		return result;


	if(crank == rank) {
		result.voxeltype = VOXEL_NATIVE;
		result.vox.nvox = native_voxel_access(nvs, c);
	} else {
		int hashval = voxel_hash_coord(vc, c);
		voxel_hash_entry *vhe;
		if(vc->voxelhashes[crank] && (vhe = vc->voxelhashes[crank][hashval])) {
			for(; vhe != NULL; vhe = vhe->next) {
				if(coordinate_equal(c, vhe->vx.coord)) {
					result.voxeltype = VOXEL_CACHED; 
					result.vox.cvox = &vhe->vx;
					break;
				}
			}
		}
	}



	return result;
} 

/* voxelcache_insert: Insert voxel @v in @fromrank's coordinate space into the
 * voxel cache @vc. See fw_voxel.h for more information on the hash structure.
 */

void		voxelcache_insert(voxel_cache *vc, cached_voxel v, int fromrank)
{
	int hashval = voxel_hash(vc, v);

	voxel_hash_entry *vhe;
	extern int rank, size;

	extern native_voxel_space localvxspace;

//	// Check if coordinate is already cached/inside local voxel space

	if(coordinate_to_rank(v.coord) == rank)
		return;

	if(vc->voxelhashes[fromrank] && (vhe = vc->voxelhashes[fromrank][hashval])) {
		for(; vhe != NULL; vhe = vhe->next) {
			if(coordinate_equal(v.coord, vhe->vx.coord)) {
				return;
			}
		}
	}

	assert(vc != NULL);
	assert(vc->voxelhashes != NULL);

	vhe = fw_malloc(sizeof(voxel_hash_entry));


	if(!vc->voxelhashes[fromrank]) {
		vc->voxelhashes[fromrank] = fw_malloc(sizeof(voxel_hash_entry *) * vc->hashsize);
		memset(vc->voxelhashes[fromrank], 0, sizeof(voxel_hash_entry *) * vc->hashsize);
	}

	vhe->vx = v;
	vhe->next = vc->voxelhashes[fromrank][hashval];
	vhe->prev = NULL;

	if(vc->voxelhashes[fromrank][hashval])
		vc->voxelhashes[fromrank][hashval]->prev = vhe;

	vc->voxelhashes[fromrank][hashval] = vhe;	

}

/* voxelcache_create: Create a new voxel cache, called from fw_script.c at the start of every repeat.
 * Returns a new, empty voxel cache.
 */

voxel_cache	*voxelcache_create(void)
{
	voxel_cache *res;
	extern int size;
	extern native_voxel_space localvxspace;


	res = fw_malloc(sizeof(voxel_cache));

	res->voxelhashes = fw_malloc(sizeof(voxel_hash_entry **) * size);
	res->hashsize = 1 + (localvxspace.Xend - localvxspace.Xstart) * (localvxspace.Yend - localvxspace.Ystart) * (localvxspace.Zend - localvxspace.Zstart) / 16; 
	memset(res->voxelhashes, 0, sizeof(voxel_hash_entry **) * size);

	return res;
}

/* voxelcache_flushall: Flush and free the voxel cache @vc. After this routine has
 * been called, @vc is unusable and new voxel caches must be allocated by voxelcache_create.
 */

void		voxelcache_flushall(voxel_cache *vc)
{
	extern int size;

	int i;


	for(i = 0; i < size; i++)
		voxelcache_flush(vc, i);

	fw_free(vc->voxelhashes);
	fw_free(vc);

}


/* voxelcache_flush: Flush and free the voxel cache line from @fromrank in @vc. After this routine has
 * been called, all voxels previously transferred from @fromrank are no longer cached in @vc.
 */


void		voxelcache_flush(voxel_cache *vc, int fromrank)
{
	if(vc->voxelhashes[fromrank]) {
		int i;
		voxel_hash_entry *vhe1, *vhe2;

		for(i = 0; i < vc->hashsize; i++) {
			for(vhe1 = vc->voxelhashes[fromrank][i]; vhe1 != NULL; vhe1 = vhe2) {
				vhe2 = vhe1->next;
				fw_free(vhe1);
			}
		}

		fw_free(vc->voxelhashes[fromrank]);
		vc->voxelhashes[fromrank] = NULL;
	}
}

/* voxelcache_resetcachemap: Reset the cache map of all the voxels in @nvs.
 * See also comments regarding scalable caches above voxel_space_create.
 */


void		voxelcache_resetcachemap(native_voxel_space *nvs)
{
	int x, y, z;
	extern int rank;

	for(x = nvs->Xstart; x < nvs->Xend; x++) {
		for(y = nvs->Ystart; y < nvs->Yend; y++) {
			for(z = nvs->Zstart; z < nvs->Zend; z++) {
				coordinate c = { x, y, z };
#ifdef SCALABLE_CACHE
				native_voxel *nv = native_voxel_access(nvs, c);
				nv->cachemaps[0] = 0;
				nv->cachemaps[1] = 0;
				nv->cachemaps[2] = 0;
				nv->cachemaps[3] = 0;
				nv->cachemaps[4] = 0;
				nv->cachemaps[5] = 0;
				nv->cachemaps[6] = 0;
				nv->cachemaps[7] = 0;
#else
				// This might look insecure, but native_voxel_access on c returns zero,
				// we are in serious trouble anyway!

				native_voxel_access(nvs, c)->cachemap = (1 << rank);
#endif
			}
		}
	}
}

