/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_voxel.h: Voxel structures. The following figure explains the relationships between the structures
 *
 *    native_voxel_space ---> array of native_voxel                      [                ]
 *    voxel_cache        ---> array of linked lists of voxel_hash_entrys [     X          ]
 *                                                                             |--> voxel_hash_entry  
 *                                                                             |--> voxel_hash_entry --- each entry has one cached_voxel
 * native_voxel_space has an array of native_voxels.
 * voxel_cache has an array of a linked list of voxel_hash_entries. This array acts as a hash table. Each linked list 
 * corresponds to one hash value. Each element in the linked list has a link to a cached_voxel, which contains the actual data.
 */

#ifndef FW_VOXEL
#define FW_VOXEL

#include "fw_coord.h"

#include <mpi.h>



// MPI tags
#define		CPFW_MPI_LENGTHDATA	0x00	// Length info, single int
#define		CPFW_MPI_VOXELDATA	0x01	// Voxel data, array of cached_voxel_mpit (used in BV strategy)
#define		CPFW_MPI_RAWVOXELDATA	0x02	// Raw voxel data, plain array of floats (used in MT/BT strategies)
#define		CPFW_MPI_COORDDATA	0x03	// Coordinates, used in MT/BT strategies


#define		VOXEL_NATIVE		0x00	// Native voxel (see fw_voxel.c)
#define		VOXEL_CACHED		0x01	// Cached voxel
#define		VOXEL_NOTFOUND		0x02	// Not found

// Native voxels

typedef struct native_voxel {
	float data;
#ifndef SCALABLE_CACHE
	unsigned long cachemap; 	// Bitmask of where this voxel has been cached - max 32/64 nodes, depending on integer size
#else
	unsigned long cachemaps[8]; 	// Larger cache, but requires more memory.
#endif
} native_voxel;

// Cached voxels

typedef struct cached_voxel {
	coordinate coord;
	float data;
} cached_voxel;

// Native voxel space

typedef struct native_voxel_space {
	int Xstart, Ystart, Zstart;
	int Xend, Yend, Zend;
	int spacesize;
	int payload;
	struct native_voxel *voxellist; // 3D array of voxels
} native_voxel_space;

// Voxel (see fw_voxel.c)

typedef struct voxel {

	union vox { 
		native_voxel *nvox;
		cached_voxel *cvox;
	} vox;

	char voxeltype;
} voxel;

// Voxel hash entry (see beginning of this file)

typedef struct voxel_hash_entry {
	struct cached_voxel vx;
	struct voxel_hash_entry *next;
	struct voxel_hash_entry *prev;
} voxel_hash_entry;

// Voxel cache (see beginning of this file and fw_voxel.c)

typedef struct voxel_cache {
	int hashsize;
	struct voxel_hash_entry ***voxelhashes;
} voxel_cache;


native_voxel_space		voxel_space_create(int n, int m, int k, int payload);

voxel_cache 			*voxelcache_create(void);
void				voxelcache_insert(voxel_cache *vc, cached_voxel v, int fromrank);
void				voxelcache_flush(voxel_cache *vc, int fromrank);
void				voxelcache_flushall(voxel_cache *vc);
void				voxelcache_resetcachemap(native_voxel_space *nvs);

voxel				voxel_lookup(native_voxel_space *nvs, voxel_cache *vc, coordinate c);


/* native_voxel_access: Look up a native voxel at coordinate @c in the native voxel space @nvs.
 */

static inline native_voxel *native_voxel_access(native_voxel_space *nvs, coordinate c) { return &nvs->voxellist[(c.x - nvs->Xstart) + (c.y - nvs->Ystart) * (nvs->Xend - nvs->Xstart) + (c.z - nvs->Zstart) * (nvs->Yend - nvs->Ystart) * (nvs->Xend - nvs->Xstart)]; }

#endif
