/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_voxelcachetransfer.c: Implements communication routines for the Bounding Volumes
 * transfer strategy. Functions in this file are called from
 *   splitpolygons_extractpoints_boundingvolumes
 * in fw_splitpolygons.c
 *
 * NOTE! While fw_voxelcachetransfer and fw_voxeltransfer might look the same,
 * they are not. Different structures (voxelcache_receive/sendqueues vs. voxeltransfer_receive/sendqueues)
 * are used and the communication protocol is different.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpi.h>

#include "fw_main.h"
#include "fw_voxel.h"
#include "fw_voxelcachetransfer.h"



// from fw_main.c

extern int rank;
extern int griddims[2];
extern int nodecoords[2];
extern int singlemode;

extern MPI_Comm cpfwcomm;


// from fw_voxel.c

extern MPI_Datatype cached_voxel_mpit;
extern MPI_Datatype coordinate_mpit;

extern voxel_cache *voxelcache;


/* voxelcache_receivequeue_create: Create and return a new, empty receivequeue.
 * Receivequeues act as wrappers from a set of receive buffers.
 * Returns a new, empty receivequeue.
 */

voxelcache_receivequeue *voxelcache_receivequeue_create(void)
{
	extern int size;

	voxelcache_receivequeue *res = fw_malloc(sizeof(voxelcache_receivequeue));

	/* ->recvfromnodes: The nodes we are going to receive from.
	 * ->requests: MPI_Request objects for nonblocking receives.
	 * ->recvlengths: The length of the voxels each node is going to transmit to this node.
	 * ->recvbufs: The actual receive buffer.
	 */

	res->recvfromnodes = fw_malloc(sizeof(char) * size);
	res->requests = fw_malloc(sizeof(MPI_Request) * size);
	res->recvlengths = fw_malloc(sizeof(int) * size);
	res->recvbufs = fw_malloc(sizeof(cached_voxel *) * size);

	memset(res->recvfromnodes, 0, sizeof(char) * size); // sizeof(char) =def 1
	memset(res->requests, 0, sizeof(MPI_Request) * size);
	memset(res->recvlengths, 0, sizeof(int) * size);
	memset(res->recvbufs, 0, sizeof(cached_voxel *) * size);

	return res;
}

/* voxelcache_receivequeue_free: Free the receive queue @rq
 */

void			voxelcache_receivequeue_free(voxelcache_receivequeue *rq)
{
	int i;
	extern int size;

	fw_free(rq->recvfromnodes);	
	fw_free(rq->recvlengths);
	fw_free(rq->requests);

	for(i = 0; i < size; i++) {
		if(rq->recvbufs[i])
			fw_free(rq->recvbufs[i]);
	}
	
	fw_free(rq);
}

/* voxelcache_sendqueue_create: Create a new sendqueue.
 * Sendqueues are as receivequeues, only for buffering sends. (See exchange routines below)
 */


voxelcache_sendqueue *voxelcache_sendqueue_create(void)
{
	extern int size;

	voxelcache_sendqueue *res = fw_malloc(sizeof(voxelcache_sendqueue));
	res->sendbuffers = fw_malloc(sizeof(voxelcache_sendbuf) * size);
	memset(res->sendbuffers, 0, sizeof(voxelcache_sendbuf) * size);

	return res;
}

/* voxelcache_sendqueue_add: Add all voxels to sendqueue @sq which intersect (1) this node's coordinate space and (2) the box 
 * starting at coordinate @min, going to coordinate @max for sending to node @rrank. @min and @max 
 * are the extents of a triangle. Only voxels not sent before are sent. (See fw_voxel.c for caching information)
 */

void			voxelcache_sendqueue_add(voxelcache_sendqueue *sq, coordinate min, coordinate max, int rrank)
{
	// Find the part of this node's local space which intersects the cube starting from min extending to max.

	int x, y, z;
	coordinate min_local, max_local;


	extern native_voxel_space localvxspace;


#ifdef SCALABLE_CACHE
	int rrank_cachemapidx, rrank_cacheidx;
	int ulsz = 8 * sizeof(unsigned long);
	rrank_cachemapidx = rrank / ulsz;
	rrank_cacheidx = rrank % ulsz;
	

	assert(rrank_cachemapidx < 8);
#endif

	if(min.x < localvxspace.Xstart)
		min_local.x = localvxspace.Xstart;
	else
		min_local.x = min.x;


	if(min.y < localvxspace.Ystart)
		min_local.y = localvxspace.Ystart;
	else
		min_local.y = min.y;


	if(max.x >= localvxspace.Xend)
		max_local.x = localvxspace.Xend;
	else
		max_local.x = max.x;



	if(max.y >= localvxspace.Yend)
		max_local.y = localvxspace.Yend;
	else
		max_local.y = max.y;


	assert(min.z >= localvxspace.Zstart);
	assert(max.z < localvxspace.Zend);	


	min_local.z = min.z;
	max_local.z = max.z + 1;
	
	/* Over-allocate the send buffer if the cube we are processing does not fit. This ensures that
	 * even though we over-allocate now (since we do not know how many of the (maxx-minx)*(maxy-miny)*(maxz-minz) voxels
	 * have been previously transmitted), we will save space if other voxels are to be sent to the other node
	 */

	assert(sq != NULL);
	assert(sq->sendbuffers != NULL);

	if(sq->sendbuffers[rrank].size == 0)
		sq->sendbuffers[rrank].buffer = NULL;

	if(sq->sendbuffers[rrank].size < sq->sendbuffers[rrank].used + (max_local.x - min_local.x) * (max_local.y - min_local.y) * (max_local.z - min_local.z)) {
		sq->sendbuffers[rrank].size += (max_local.x - min_local.x) * (max_local.y - min_local.y) * (max_local.z - min_local.z);
		sq->sendbuffers[rrank].buffer = fw_realloc(sq->sendbuffers[rrank].buffer, sq->sendbuffers[rrank].size * sizeof(cached_voxel));
	}

	for(x = min_local.x; x < max_local.x; x++) {
		for(y = min_local.y; y < max_local.y; y++) {
			for(z = min_local.z; z < max_local.z; z++) {
				coordinate c = { x, y, z };
				native_voxel *v = native_voxel_access(&localvxspace, c);

				// Already sent?

#ifdef SCALABLE_CACHE
				if(!(v->cachemaps[rrank_cachemapidx] & (1 << rrank_cacheidx))) {
					v->cachemaps[rrank_cachemapidx] |= (1 << rrank_cacheidx);
#else
				if(!(v->cachemap & (1 << rrank))) {
					v->cachemap |= 1 << rrank;
#endif
					sq->sendbuffers[rrank].buffer[sq->sendbuffers[rrank].used].coord = c;
					sq->sendbuffers[rrank].buffer[sq->sendbuffers[rrank].used++].data = v->data;
				}
			}
		}
	}

	// Mark that data is to be sent to rrank

	sq->sendbuffers[rrank].sendexpect = 1;

}


/* voxelcache_sendqueue_free: Free sendqueue @sq.
 */

void			voxelcache_sendqueue_free(voxelcache_sendqueue *sq)
{
	int i;
	extern int size;

	for(i = 0; i < size; i++) {
		if(sq->sendbuffers[i].buffer)
			fw_free(sq->sendbuffers[i].buffer);
	}

	fw_free(sq->sendbuffers);
	fw_free(sq);
}

/* voxelcache_exchange_start: Start the exchange of the bounding volumes (see thesis Chapter 5). This is Step 2/3 of the BV strategy.
 */


void			voxelcache_exchange_start(voxelcache_receivequeue *rq, voxelcache_sendqueue *sq)
{
	int i;
	
	extern int size, rank;
	extern MPI_Comm cpfwcomm;

	int nrecvs = 0;
	int flag;


	memset(rq->recvbufs, 0, sizeof(cached_voxel *) * size);
	memset(rq->recvlengths, 0, sizeof(int) * size);

	/* Step A.
	 * 
	 * We send two items to each node:
	 *  1) the length of the voxel data (may be zero)
	 *  2) the actual voxel data
	 * We use nonblocking communication. The MPI_Request for 1 is stored in the sendbuffer's lenrequest, while
	 * for 2 it is stored in the datarequest field.
	 *
 	 * We also start a length receive operation from all nodes. Node i's MPI_Request for the receive is stored in the requests[i]
	 * in the receive queue.
	 */

	for(i = 0; i < size; i++) {
		if(sq->sendbuffers[i].sendexpect > 0) {
			MPI_Isend(&sq->sendbuffers[i].used, 1, MPI_INT, i, CPFW_MPI_LENGTHDATA, cpfwcomm, &sq->sendbuffers[i].lenrequest);
			if(sq->sendbuffers[i].used > 0)
				MPI_Isend(sq->sendbuffers[i].buffer, sq->sendbuffers[i].used, cached_voxel_mpit, i, CPFW_MPI_VOXELDATA, cpfwcomm, &sq->sendbuffers[i].datarequest);

		}

		if(rq->recvfromnodes[i] > 0) {
			MPI_Irecv(&rq->recvlengths[i], 1, MPI_INT, i, CPFW_MPI_LENGTHDATA, cpfwcomm, &rq->requests[i]);
			++nrecvs;
		}
	}


	rq->recvrequests = 0;

	/* Step B. We wait for the length receives to complete and initiate a voxel receive from node i if 
	 * the received length from node i is > 0.
	 */

	while(nrecvs > 0) {
		for(i = 0; i < size; i++) {
			if(rq->recvfromnodes[i] > 0 && rq->recvbufs[i] == NULL) {
				MPI_Test(&rq->requests[i], &flag, MPI_STATUS_IGNORE);
				if(flag) { // Operation completed
					
					--nrecvs;

					if(rq->requests[i] != MPI_REQUEST_NULL)
						MPI_Request_free(&rq->requests[i]);

					if(rq->recvlengths[i] == 0) {
						rq->recvfromnodes[i] = 0;
						continue;
					}
					rq->recvbufs[i] = fw_malloc(sizeof(cached_voxel) * rq->recvlengths[i]);
					MPI_Irecv(rq->recvbufs[i], rq->recvlengths[i], cached_voxel_mpit, i, CPFW_MPI_VOXELDATA, cpfwcomm, &rq->requests[i]);
					rq->recvrequests++;
				}
			}
		}
	}
}


/* voxelcache_exchange_finalize: Finalize transfers in @rq and @sq. This is called after
 * voxelcache_exchange_start and after voxelization is done (see fw_splitpolygons.c). This is a part of Step 4
 * of the BV strategy. After this, computations can be done since all voxels are received.
 */


void			voxelcache_exchange_finalize(voxelcache_receivequeue *rq, voxelcache_sendqueue *sq)
{
	int i;
	int flag;


	extern MPI_Comm cpfwcomm;
	extern voxel_cache *voxelcache;

	extern int size, rank;

	/* Step C. Wait for the receives initiated in voxelcache_exchange_start
	 * to finish.
	 */

	while(rq->recvrequests > 0) {

		for(i = 0; i < size; i++) {
			if(rq->recvfromnodes[i] > 0 && rq->recvbufs[i] != NULL) {
				MPI_Test(&rq->requests[i], &flag, MPI_STATUS_IGNORE);
				if(flag) { // OK

					if(rq->requests[i] != MPI_REQUEST_NULL)
						MPI_Request_free(&rq->requests[i]);

					int j;

					for(j = 0; j < rq->recvlengths[i]; j++) {
						voxelcache_insert(voxelcache, rq->recvbufs[i][j], i);
					}
			

					fw_free(rq->recvbufs[i]);
					rq->recvbufs[i] = NULL;
					--rq->recvrequests;
				}
			}
		}
	}

	/* Step D. Wait for send requets to finish, to ensure
	 * that the other nodes have received the data (and that we do not
	 * continue into the next move-extract cycle before all nodes have finished!)
	 */



	for(i = 0; i < size; i++) {
		if(sq->sendbuffers[i].sendexpect > 0) {
			MPI_Wait(&sq->sendbuffers[i].lenrequest, MPI_STATUS_IGNORE);

			if(sq->sendbuffers[i].lenrequest != MPI_REQUEST_NULL)
				MPI_Request_free(&sq->sendbuffers[i].lenrequest);


			if(sq->sendbuffers[i].used > 0) {
				MPI_Wait(&sq->sendbuffers[i].datarequest, MPI_STATUS_IGNORE);
				if(sq->sendbuffers[i].datarequest != MPI_REQUEST_NULL)
					MPI_Request_free(&sq->sendbuffers[i].datarequest);
				fw_free(sq->sendbuffers[i].buffer);
				sq->sendbuffers[i].buffer = NULL;
			}
		}
	}
}


