/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_voxeltransfer.c: Implement the communication routines for the
 * Minimum Communication and Block Transfer routines, called from
 * the functions
 *   splitpolygons_extractpoints_minimumtransfer
 *   splitpolygons_extractpoints_blockedtransfer
 * fw_splitpolygons.c
 *
 * NOTE! While fw_voxelcachetransfer and fw_voxeltransfer might look the same,
 * they are not. Different structures (voxelcache_receive/sendqueues vs. voxeltransfer_receive/sendqueues)
 * are used and the communication protocol is different.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpi.h>

#include "fw_main.h"
#include "fw_voxel.h"
#include "fw_voxeltransfer.h"
#include "fw_voxelcachetransfer.h"

/* voxeltransfer_sendqueue_create: Create and return a new voxeltransfer sendqueue.
 * Sendqueues contain buffers and information related to sending (see the exchange function below).
 */

voxeltransfer_sendqueue *voxeltransfer_sendqueue_create(void)
{
	extern int size;

	voxeltransfer_sendqueue *res = fw_malloc(sizeof(voxeltransfer_sendqueue));
	res->sendbuffers = fw_malloc(sizeof(voxeltransfer_sendbuf) * size);
	memset(res->sendbuffers, 0, sizeof(voxeltransfer_sendbuf) * size);

	return res;
}

/* voxeltransfer_sendqueue_free: Free the sendqueue @sq
 */

void			voxeltransfer_sendqueue_free(voxeltransfer_sendqueue *sq)
{
	int i;
	extern int rank, size;

	for(i = 0; i < size; i++) {
		if(sq->sendbuffers[i].buffer_v) {
			fw_free(sq->sendbuffers[i].buffer_v);
		}

		if(sq->sendbuffers[i].buffer_c) {
			fw_free(sq->sendbuffers[i].buffer_c);
		}
	}

	fw_free(sq->sendbuffers);
	fw_free(sq);
}

/* voxeltransfer_sendqueue_add_coordinate: Add coordinate request for @c, going to @rrank to the send queue @sq. Thus, 
 * this ensures that @c is requested from @rrank.
 */

void				voxeltransfer_sendqueue_add_coordinate(voxeltransfer_sendqueue *sq, coordinate *c, int rrank)
{
	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	extern int n, m, k;	

	assert(sq != NULL);
	assert(sq->sendbuffers != NULL);

	/* Sanity check - rounding errors may in very special cases give coordinates outside the space
	 * If so, we just drop them. This has only been observed once in hundreds of tests.
	 */

	if(c->x < 0 || c->y < 0 || c->z < 0 || c->x >= n || c->y >= m || c->z >= k)
		return;

	if(sq->sendbuffers[rrank].size == 0)
		sq->sendbuffers[rrank].buffer_c = NULL;

	/* We do over-allocation to avoid many calls to
	 * fw_realloc
	 */


	if(sq->sendbuffers[rrank].size < sq->sendbuffers[rrank].used + 1) {
		sq->sendbuffers[rrank].size += 1 + 2 * sq->sendbuffers[rrank].size;
		sq->sendbuffers[rrank].buffer_c = fw_realloc(sq->sendbuffers[rrank].buffer_c, sq->sendbuffers[rrank].size * sizeof(coordinate));
	}

	if(voxel_lookup(&localvxspace, voxelcache, *c).voxeltype == VOXEL_NOTFOUND)
		sq->sendbuffers[rrank].buffer_c[sq->sendbuffers[rrank].used++] = *c;
}



/* voxeltransfer_sendqueue_add_coordinate_blocked: Add coordinate @c and k (specified by the user from the command line in bt=..) 
 * surrounding coordinates (which are on this node) to the , going to @rrank to the send queue @sq.
 * This is just like voxeltransfer_sendqueue_add_coordinate, except we also request the surrounding coordinates. 
 * This is what separates the MT and BT strategies.
 *
 * @tmpvoxelcache is a temporary voxel cache, allocated by the caller, where voxels which are added to the request
 * are stored temporarily. This is done since when adding the k surrounding coordinates of @c might lead to that
 * the next coordinate is already added, and we do not want to request coordinates twice! 
 */

void				voxeltransfer_sendqueue_add_coordinate_blocked(voxeltransfer_sendqueue *sq, coordinate *c, int rrank, voxel_cache *tmpvoxelcache)
{
	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	extern int blocksize;
	extern int size, rank;
	extern int n, m, k;

	int x, y, z;
	cached_voxel cv;
	voxel v;
	
	assert(sq != NULL);
	assert(sq->sendbuffers != NULL);

	if(sq->sendbuffers[rrank].size == 0)
		sq->sendbuffers[rrank].buffer_c = NULL;


	/* We do over-allocation to avoid many calls to
	 * fw_realloc
	 */

	if(sq->sendbuffers[rrank].size < sq->sendbuffers[rrank].used + (blocksize + 1)*(blocksize + 1)*(blocksize + 1)) {
		sq->sendbuffers[rrank].size += (blocksize + 1) * (blocksize + 1) * (blocksize + 1) + sq->sendbuffers[rrank].size;
		sq->sendbuffers[rrank].buffer_c = fw_realloc(sq->sendbuffers[rrank].buffer_c, sq->sendbuffers[rrank].size * sizeof(coordinate));
	}


	for(x = c->x - blocksize/2; x <= c->x + blocksize/2; x++) {
		for(y = c->y - blocksize/2; y <= c->y + blocksize/2; y++) {
			for(z = c->z - blocksize/2; z <= c->z + blocksize/2; z++) {
				coordinate c0 = { x, y, z };
				// Check if coordinate is valid and in our coordinate space
				if(c0.x >= 0 && c0.x < n &&
					c0.y >= 0 && c0.y < m &&
					c0.z >= 0 && c0.z < k && coordinate_to_rank(c0) == rrank) {
						v = voxel_lookup(&localvxspace, tmpvoxelcache, c0);

						// Add the coordinate if it has not already been added in a previous call

						if(v.voxeltype == VOXEL_NOTFOUND) {
							sq->sendbuffers[rrank].buffer_c[sq->sendbuffers[rrank].used++] = c0;
							cv.coord = c0;
							voxelcache_insert(tmpvoxelcache, cv, rrank);
						}
				}
			}
		}
	} 
}


/* voxeltransfer_receivequeue_create: Create and return a new, empty receivequeue.
 * Receivequeues act as wrappers from a set of receive buffers.
 * Returns a new, empty receivequeue.
 */

voxeltransfer_receivequeue *voxeltransfer_receivequeue_create(void)
{
	extern int size;

	voxeltransfer_receivequeue *res = fw_malloc(sizeof(voxeltransfer_receivequeue));


	/* ->recvfrmonodes: Nodes to receive from
	 * ->coordrequests: Receive requests for coordinate requests
	 * ->voxelrequests: Receive requests for voxel requests
	 * ->recvlengths: recvlengths[i] denots # of coordinates to receive from node i
	 * ->recvbufs_voxels/recvbufs_coords: Receive buffers for voxels/coordinates
	 */

	res->recvfromnodes = fw_malloc(sizeof(char) * size);
	res->coordrequests = fw_malloc(sizeof(MPI_Request) * size);
	res->voxelrequests = fw_malloc(sizeof(MPI_Request) * size);
	res->recvlengths = fw_malloc(sizeof(int) * size);
	res->recvbufs_voxels = fw_malloc(sizeof(float *) * size);
	res->recvbufs_coords = fw_malloc(sizeof(coordinate *) * size);
	
	memset(res->recvfromnodes, 0, sizeof(char) * size); // sizeof(char) =def 1
	memset(res->coordrequests, 0, sizeof(MPI_Request) * size);
	memset(res->voxelrequests, 0, sizeof(MPI_Request) * size);
	memset(res->recvlengths, 0, sizeof(int) * size);
	memset(res->recvbufs_voxels, 0, sizeof(float *) * size);
	memset(res->recvbufs_coords, 0, sizeof(coordinate *) * size);


	return res;
}


/* voxeltransfer_receivequeue_free: Free the receive queue @rq
 */

void			voxeltransfer_receivequeue_free(voxeltransfer_receivequeue *rq)
{
	int i;
	extern int rank, size;


	fw_free(rq->recvfromnodes);	
	fw_free(rq->recvlengths);

	if(rq->coordrequests)
		fw_free(rq->coordrequests);
	if(rq->voxelrequests)
		fw_free(rq->voxelrequests);

	for(i = 0; i < size; i++) {
		if(rq->recvbufs_voxels[i]) {
			fw_free(rq->recvbufs_voxels[i]);
		}

		if(rq->recvbufs_coords[i]) {
			fw_free(rq->recvbufs_coords[i]);
		}

	}
	
	fw_free(rq);
}

/* voxeltransfer_exchange: This does the actual exchange of voxels. Voxels to be sent are in @sq, and receive
 * buffers are in @rq.
 * 
 * The communication consists of the following steps:
 *   A. Request coordinates we need from other nodes, and receive requests from other nodes.
 *   B. Send the voxels to the requesting nodes.
 *   C. Receive voxels requested in A from other nodes.
 */


void				voxeltransfer_exchange(voxeltransfer_receivequeue *rq, voxeltransfer_sendqueue *sq)
{
	int i;
	int nreqs = 0, nrecvs = 0, ncoordrecvs = 0;

	extern int rank, size;
	extern native_voxel_space localvxspace;
	extern voxel_cache *voxelcache;

	extern MPI_Datatype coordinate_mpit;

	extern MPI_Comm cpfwcomm;


	/* Step A. Send requests to other nodes, and receive requests from other nodes.
	 * We send both the length and voxel data at once and in two separate calls.
	 */

	for(i = 0; i < size; i++) {

		/* Requests Added: sq->sendbuffers[i].lengthrequest
		 *                 sq->sendbuffers[i].coordrequest
		 *                 rq->voxelrequests[i]
		 */

		if(sq->sendbuffers[i].sendexpect > 0) {
			MPI_Isend(&sq->sendbuffers[i].used, 1, MPI_INT, i, CPFW_MPI_LENGTHDATA, cpfwcomm, &sq->sendbuffers[i].lengthrequest);
			++nreqs;

			if(sq->sendbuffers[i].used > 0) {
				MPI_Isend(sq->sendbuffers[i].buffer_c, sq->sendbuffers[i].used, coordinate_mpit, i, CPFW_MPI_COORDDATA, cpfwcomm, &sq->sendbuffers[i].coordrequest);
				// Start receive
				rq->recvbufs_voxels[i] = fw_malloc(sizeof(float) * sq->sendbuffers[i].used);
				MPI_Irecv(rq->recvbufs_voxels[i], sq->sendbuffers[i].used, MPI_FLOAT, i, CPFW_MPI_RAWVOXELDATA, cpfwcomm, &rq->voxelrequests[i]);
			}
		}

		/* Requests Added: rq->coordrequests[i]
		 */

		if(rq->recvfromnodes[i] > 0) {

			MPI_Irecv(&rq->recvlengths[i], 1, MPI_INT, i, CPFW_MPI_LENGTHDATA, cpfwcomm, &rq->coordrequests[i]);

			++nrecvs;
		}
	}


	/* Start receiving the coordinate requests from other nodes. This must be done here since 
	 * in the previous loop we do not know the lengths and thus can not allocate the buffers.
	 */


	while(nrecvs > 0) {
		for(i = 0; i < size; i++) {

			/* Requests Completed: rq->coordrequests[i]
			 * Requests Added:     rq->coordrequests[i] IF length of request from other node > 0 voxels
			 */


			if(rq->recvfromnodes[i] > 0) {
				int flag;

				MPI_Test(&rq->coordrequests[i], &flag, MPI_STATUS_IGNORE);

				if(!flag)
					continue;

				--nrecvs;
				if(rq->coordrequests[i] != MPI_REQUEST_NULL)
					MPI_Request_free(&rq->coordrequests[i]);

				rq->recvfromnodes[i] = 0;
		

				if(rq->recvlengths[i] > 0) {
					rq->recvbufs_coords[i] = fw_malloc(sizeof(coordinate) * rq->recvlengths[i]);
					MPI_Irecv(rq->recvbufs_coords[i], rq->recvlengths[i], coordinate_mpit, i, CPFW_MPI_COORDDATA, cpfwcomm, &rq->coordrequests[i]);
					++ncoordrecvs;
					rq->recvfromnodes[i] = -1;

				}
			}
		}
	}


	/* Step B. Send requested voxels to requesting nodes.
	 */


	while(ncoordrecvs > 0) {
		for(i = 0; i < size; i++) {

			/* Requests Completed:  rq->coordrequests[i]
			 * Requests Added       sq->sendbuffers[i].voxelrequest
			 */

			if(rq->recvfromnodes[i] == -1) {
				int flag, j;

				MPI_Test(&rq->coordrequests[i], &flag, MPI_STATUS_IGNORE);

				if(!flag)
					continue;




				--ncoordrecvs;
				if(rq->coordrequests[i] != MPI_REQUEST_NULL)
					MPI_Request_free(&rq->coordrequests[i]);

				sq->sendbuffers[i].buffer_v = fw_malloc(sizeof(float) * rq->recvlengths[i]);
				for(j = 0; j < rq->recvlengths[i]; j++) {
					coordinate c = rq->recvbufs_coords[i][j];
					native_voxel *v = native_voxel_access(&localvxspace, c);
					sq->sendbuffers[i].buffer_v[j] = v->data;
				}



				MPI_Isend(sq->sendbuffers[i].buffer_v, rq->recvlengths[i], MPI_FLOAT, i, CPFW_MPI_RAWVOXELDATA, cpfwcomm, &sq->sendbuffers[i].voxelrequest);

				// rq->recvfrmnodes[i] is set to -2 if we have sent voxels to node i.
				// This is used when waiting for the requests to complete, below.
				rq->recvfromnodes[i] = -2;
			}
		}
	}


	/* Step C. Receive requested voxels. We also wait for the previous send requsts to complete.
	 */

	while(nreqs > 0) {
		for(i = 0; i < size; i++) {
			if(sq->sendbuffers[i].used > 0) {
				int flag, j;

				MPI_Test(&rq->voxelrequests[i], &flag, MPI_STATUS_IGNORE);

				if(!flag)
					continue;


				if(rq->voxelrequests[i] != MPI_REQUEST_NULL)
					MPI_Request_free(&rq->voxelrequests[i]);

				for(j = 0; j < sq->sendbuffers[i].used; j++) {
					cached_voxel cv;
					cv.coord = sq->sendbuffers[i].buffer_c[j];
					cv.data = rq->recvbufs_voxels[i][j];
					voxelcache_insert(voxelcache, cv, i);
				}

				// We can now safely wait for sq->sendbuffers[i].coordrequest/lengthrequest without being afraid of blocking too long, since the fact that the voxel
				// request is completed means that the other requests should be completed. 

				MPI_Wait(&sq->sendbuffers[i].coordrequest, MPI_STATUS_IGNORE);
				if(sq->sendbuffers[i].coordrequest != MPI_REQUEST_NULL)
					MPI_Request_free(&sq->sendbuffers[i].coordrequest);

				MPI_Wait(&sq->sendbuffers[i].lengthrequest, MPI_STATUS_IGNORE);
				if(sq->sendbuffers[i].lengthrequest != MPI_REQUEST_NULL)
					MPI_Request_free(&sq->sendbuffers[i].lengthrequest);

				--nreqs;
				sq->sendbuffers[i].used = -1;

			} else if(sq->sendbuffers[i].used == 0 && sq->sendbuffers[i].sendexpect > 0) { // We only did a length request to this node (i.e. length = 0)
				MPI_Wait(&sq->sendbuffers[i].lengthrequest, MPI_STATUS_IGNORE);
				if(sq->sendbuffers[i].lengthrequest != MPI_REQUEST_NULL)
					MPI_Request_free(&sq->sendbuffers[i].lengthrequest);



				--nreqs;
				sq->sendbuffers[i].used = -1;

			}
		}
	}

	/* Step C (continued) - Finally, we wait for the voxel send requests to complete.
	 * After this, everything has been sent and everything has been received.
	 * rq->recvfromnodes is set to -2 if we have sent voxels to node i.
	 */



	for(i = 0; i < size; i++) {
		if(rq->recvfromnodes[i] == -2) {
			MPI_Wait(&sq->sendbuffers[i].voxelrequest, MPI_STATUS_IGNORE);
			if(sq->sendbuffers[i].voxelrequest != MPI_REQUEST_NULL)
				MPI_Request_free(&sq->sendbuffers[i].voxelrequest);
		}
	}
}
