/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_wingededge.c: The winged-edge structure. This file contains functions for adding/removing polygons and
 * creating/destroying meshes, and moving vertices of polygons.
 *
 * See thesis Chapter 2 for more informtion on Winged-Edge. See fw_wingededge.h for more information
 * on the structures used in this file.
 *
 * Most functions in this file are called from
 * - fw_script.c
 * - fw_splitpolygons.c, which provides some wrappers around the functions in this file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "fw_main.h"
#include "fw_vector.h"
#include "fw_wingededge.h"

mesh	*meshlist = NULL;	// List of all meshes
int	nmeshes = 0;		// Number of meshes

/* we_mesh_create: Create a new, empty mesh and return a new mesh object.
 * This is slightly different from what is specified in the thesis, where the Create-Mesh
 * operation takes three vertices as arguments.
 * 
 * The Create-Mesh operation can be viewed as a 
 * call to we_mesh_create() followed by a call to we_add_polygon(). The Add-Face operation
 * is a call to we_add_polygon() where two of the vertices have already been added.
 * (The genscript program generates scripts which behave this way).
 */

mesh	*we_mesh_create(void)
{
	int i;

	mesh *res = fw_malloc(sizeof(mesh));
	res->faces = NULL;
	memset(res->vertices, 0, sizeof(struct vertex *) * VERTEX_HASH_SIZE);
	res->edges = NULL;

	res->next = meshlist;
	res->prev = NULL;

	res->nfaces = res->nedges = res->nvertices = 0;

	if(meshlist) {
		meshlist->prev = res;
	}

	meshlist = res;
	++nmeshes;

	return res;
}

/* we_mesh_destroy: Destroy the mesh m.
 * This function removes all faces, edges and vertices of the mesh.
 */

void	we_mesh_destroy(mesh *m)
{
	int i;
	face *tmpf, *tmpf2;
	edge *tmpe, *tmpe2;
	vertex *tmpv, *tmpv2;

	if(m->next)
		m->next->prev = m->prev;
	if(m->prev)
		m->prev->next = m->next;

	if(m == meshlist)
		meshlist = m->next;

	// Remove faces

	for(tmpf = m->faces; tmpf != NULL;) {
		tmpf2 = tmpf->next;
		fw_free(tmpf);
		tmpf = tmpf2;
	}


	// Remove edges

	for(tmpe = m->edges; tmpe != NULL;) {
		tmpe2 = tmpe->next;
		fw_free(tmpe);
		tmpe = tmpe2;
	}

	// Remove vertices
	for(i = 0; i < VERTEX_HASH_SIZE; i++) {
		for(tmpv = m->vertices[i]; tmpv != NULL;) {	
			tmpv2 = tmpv->next;
			if(tmpv->edges)
				fw_free(tmpv->edges);
			if(tmpv->faces)
				fw_free(tmpv->faces);
			fw_free(tmpv);
			tmpv = tmpv2;
		}
	}


	fw_free(m);
	--nmeshes;
}

/* we_add_polygon: Add triangle with vertices @v1, @v2 and @v3 to mesh @m, and 
 * return the resulting face object. As described in the thesis, only manifold meshes are supported.
 * This uses we_edge_add, we_vertex_add and we_face_add as auxiliary functions.
 */

face 	*we_add_polygon(mesh *m, vertex *v1, vertex *v2, vertex *v3)
{
	vertex *v1_m = we_vertex_find(m, v1),
	       *v2_m = we_vertex_find(m, v2),
		*v3_m = v3;

	edge *e1, *e2, *e3;

	// Do the vertices exist? Either NO vertices should exist OR both vertices should exist.

	if(! ((v1_m && v2_m) || (!v1_m && !v2_m))) {
		printf_zero("[cpfw/we_polygon_add] only one of two vertices already added for polygon face; ignoring face\n");
		return NULL;
	}

	if(!v1_m && !v2_m) {
		// First triangle - part of the Create-Mesh operation
		v1_m = v1;
		v2_m = v2;
		we_vertex_add(m, &v1_m);
		we_vertex_add(m, &v2_m);
		we_vertex_add(m, &v3_m);
		e1 = we_edge_add(m, v1_m, v2_m);
		e2 = we_edge_add(m, v2_m, v3_m);
		e3 = we_edge_add(m, v3_m, v1_m);
		return we_face_add(m, e1, e2, e3); // note: can never return NULL here
	} else {
		// Add vertex to already existing edge 
		edge *commonedge = we_edge_find(v1_m, v2_m);

		assert(commonedge != NULL);

		if(commonedge->rightface != NULL && commonedge->leftface != NULL) {
			// error! one edge shared by more than two polygons. ignore
			printf_zero("[cpfw/we_polygon_add] warning: attempted to add face [(%d,%d,%d),(%d,%d,%d),(%d,%d,%d)] including edge [(%d,%d,%d),(%d,%d,%d)] which already has two incident faces. ignoring.\n", v1->coords.x, v1->coords.y, v1->coords.z, v2->coords.x, v2->coords.y, v2->coords.z, v3->coords.x, v3->coords.y, v3->coords.z, commonedge->start->coords.x, commonedge->start->coords.y, commonedge->start->coords.z, commonedge->end->coords.x, commonedge->end->coords.y, commonedge->end->coords.z);
			return NULL;
		}

		we_vertex_add(m, &v3_m);
		e2 = we_edge_add(m, v2_m, v3_m);
		e3 = we_edge_add(m, v3_m, v1_m);
		return we_face_add(m, commonedge, e2, e3); // note: can also never return NULL here (after the above assert)
	}
}



/* we_vertex_hash: We employ a small hash table to index vertices. This makes lookup
 * quicker (but is of limited importance to our speedup timings since
 * we are interested in speedups when using many nodes - and for this,
 * using hash tables is irrelevant)
 */

static inline int we_vertex_hash(vertex *v) { return (v->coords.x * v->coords.y * v->coords.z) % VERTEX_HASH_SIZE; }

/* we_vertex_find: Find vertex object with coordinates specified by @v.coords in mesh @m.
 * This is used in we_add_polygon, see above.
 */

vertex	*we_vertex_find(mesh *m, vertex *v)
{
	int hashidx = we_vertex_hash(v);
	int found = 0;
	vertex *tmp;

	for(tmp = m->vertices[hashidx]; tmp != NULL && tmp->used != VERTEX_SLOT_UNUSED; tmp = tmp->next) {
		if(coordinate_equal(tmp->coords, v->coords)) {
			++found;
			break;
		}
	}

	return found ? tmp : NULL;
}

/* we_vertex_add: Add vertex with coordinates (*v)->coords to mesh @m.
 * The corresponding vertex object in the winged-edge structure of @m is returned in *v.
 */

void	we_vertex_add(mesh *m, vertex **v)
{

	int hashidx = we_vertex_hash(*v);

	vertex *newv = fw_malloc(sizeof(vertex));

	newv->coords = (*v)->coords;

	newv->nedges = newv->nfaces = 0;
	newv->edges = NULL;
	newv->faces = NULL;
	newv->used = VERTEX_SLOT_USED;

	newv->next = m->vertices[hashidx];
	newv->prev = NULL;
	if(newv->next)
		newv->next->prev = newv;

	m->vertices[hashidx] = newv;


	*v = newv;

	++m->nvertices;
}

/* we_edge_add: Add a new edge to mesh @m from vertex @v1 to @v2, returning pointer to
 * the edge object in the winged-edge structure of @m.
 */

edge	*we_edge_add(mesh *m, vertex *v1, vertex *v2)
{
	edge *newedge = fw_malloc(sizeof(edge));

	newedge->start = v1;
	newedge->end = v2;

	newedge->next_cw = newedge->next_ccw = newedge->prev_cw = newedge->prev_ccw = NULL;
	newedge->leftface = newedge->rightface = NULL;


	v1->edges = realloc(v1->edges, ++v1->nedges * sizeof(edge *));
	v2->edges = realloc(v2->edges, ++v2->nedges * sizeof(edge *));

	v1->edges[v1->nedges - 1] = newedge;
	v2->edges[v2->nedges - 1] = newedge;

	newedge->prev = NULL;
	newedge->next = m->edges;
	if(m->edges)
		m->edges->prev = newedge;

	++m->nedges;

	return newedge;	
}

/* we_edge_find: Find edge object between @v1 and @v2. Returns NULL
 * if no such edge exists.
 */

edge	*we_edge_find(vertex *v1, vertex *v2)
{
	int i;

	for(i = 0; i < v1->nedges; i++) {
		if(v1->edges[i]->start == v2 || v1->edges[i]->end == v2)
			return v1->edges[i];
	}

	return NULL;
}

/* we_face_add: Add face specified by edges @e1, @e2 and @e3 to mesh @m, and return new face object in the winged-edge structure
 * of the mesh @m.
 *
 * This is called from we_add_polygon above. (The difference between we_add_polygon and we_face_add is that we_face_add only
 * adds a face object to the WE structure, while we_add_polygon adds vertices, edges and faces (through we_face_add)
 * 
 */

face	*we_face_add(mesh *m, edge *e1, edge *e2, edge *e3)
{
	face *newface = fw_malloc(sizeof(face));

	newface->startedge = e1;
	if(!e1->leftface && !e2->leftface && !e3->leftface) {

		e1->leftface = e2->leftface = e3->leftface = newface;
		e1->next_cw = e2;
		e2->next_cw = e3;
		e3->next_cw = e1;
		e1->prev_cw = e3;
		e2->prev_cw = e1;
		e3->prev_cw = e2;

	} else if(!e1->rightface && !e2->rightface && !e3->rightface) {

		e1->rightface = e2->rightface = e3->rightface = newface;
		e1->next_ccw = e2;
		e2->next_ccw = e3;
		e3->next_ccw = e1;
		e1->prev_ccw = e3;
		e2->prev_ccw = e1;
		e3->prev_ccw = e2;

	} 

	/* else case checked in calling function (we_add_polygon) */

	newface->vertices[0] = e1->start;
	newface->vertices[1] = e1->end;
	newface->vertices[2] = e2->end;

	newface->vertices[0]->faces = realloc(newface->vertices[0]->faces, ++newface->vertices[0]->nfaces * sizeof(face *));
	newface->vertices[1]->faces = realloc(newface->vertices[1]->faces, ++newface->vertices[1]->nfaces * sizeof(face *));
	newface->vertices[2]->faces = realloc(newface->vertices[2]->faces, ++newface->vertices[2]->nfaces * sizeof(face *));

	newface->vertices[0]->faces[newface->vertices[0]->nfaces - 1] = newface;
	newface->vertices[1]->faces[newface->vertices[1]->nfaces - 1] = newface;
	newface->vertices[2]->faces[newface->vertices[2]->nfaces - 1] = newface;

	newface->prev = NULL;
	newface->next = m->faces;
	if(m->faces)
		m->faces->prev = newface;
	m->faces = newface;	

	++m->nfaces;

	return newface;
}


/* we_remove_vertex: Remove vertex @v from mesh @m. Should only be called internally (from we_remove_edge)
 */

static void	we_remove_vertex(mesh *m, vertex *v)
{
	int hashidx = we_vertex_hash(v);

	if(v->faces)
		fw_free(v->faces);
	if(v->edges)
		fw_free(v->edges);

	// Unlink the vertex from the mesh.

	if(v->prev)
		v->prev->next = v->next;
	if(v->next)
		v->next->prev = v->prev;

	if(v == m->vertices[hashidx]) { 
		m->vertices[hashidx] = v->next;
	}
	
	fw_free(v);
	
}


/* we_remove_edge: Remove edge @e from mesh @m. 
 */

void	we_remove_edge(mesh *m, edge *e)
{
	vertex *v1 = e->start, *v2 = e->end;

	// compact v1 and v2's edge/face arrays
	
	int i;

	for(i = 0; i < v1->nedges; i++) {
		if(v1->edges[i] == e)
			break;
	}

	for(; i + 1 < v1->nedges; i++) {
		v1->edges[i] = v1->edges[i+1];
	}


	for(i = 0; i < v2->nedges; i++) {
		if(v2->edges[i] == e)
			break;
	}

	for(; i + 1 < v2->nedges; i++) {
		v2->edges[i] = v2->edges[i+1];
	}



	v1->edges = fw_realloc(v1->edges, --v1->nedges * sizeof(edge *));
	v2->edges = fw_realloc(v2->edges, --v2->nedges * sizeof(edge *));

	if(v1->nedges == 0)
		we_remove_vertex(m, v1);
	if(v2->nedges == 0)
		we_remove_vertex(m, v2);

	// Unlink the edge from the mesh.

	if(e->prev)
		e->prev->next = e->next;
	if(e->next)
		e->next->prev = e->prev;

	if(e == m->edges)
		m->edges = e->next;
	
	fw_free(e);

}


/* we_remove_face: Remove face @f from mesh @m. 
 */

void	we_remove_face(mesh *m, face *f)
{
	// First remove this face from the vertices.

	int i, j;


	if(f == f->startedge->rightface) {
		// follow CCW order and remove/unlink edges. this may result in vertices being removed
		edge *e, *e2;



		for(e = f->startedge->next_ccw; e != f->startedge;) {
			e2 = e->next_ccw;
			e->next_ccw = NULL;
			if(e->leftface == NULL) { // edge with no face, remove it
				we_remove_edge(m, e);
			}

			e = e2;
		}
		
		f->startedge->next_ccw = NULL;
		f->startedge->rightface = NULL;
		if(f->startedge->leftface == NULL) {
			we_remove_edge(m, f->startedge);
		}
		
	} else {
		// follow CW order and remove/unlink edges
		edge *e, *e2;



		for(e = f->startedge->next_cw; e != f->startedge;) {
			e2 = e->next_cw;
			e->next_cw = NULL;
			if(e->rightface == NULL) { // edge with no face, remove it
				we_remove_edge(m, e);
			}

			e = e2;
		}

		f->startedge->next_cw = NULL;
		f->startedge->leftface = NULL;
		if(f->startedge->rightface == NULL) {
			we_remove_edge(m, f->startedge);
		}
	}

	// Unlink the face from the mesh.

	if(f->prev)
		f->prev->next = f->next;
	if(f->next)
		f->next->prev = f->prev;

	if(f == m->faces)
		m->faces = f->next;

	fw_free(f);

}



/* we_remove_polygons: Remove polygons incident to @v0 from mesh @m.
 *
 * This is the Remove-Faces operation. 
 * Notice that each vertex points to a set of faces to which it is incident.
 * Therefore, these arrays may need compaction after the face has been removed. This is done
 * in the function below.
 */

void	we_remove_polygons(mesh *m, vertex *v0)
{
	int i, j, k;
	vertex *v;

	v = we_vertex_find(m, v0);
	
	if(!v) {
		return;
	}


	for(i = 0; i < v->nfaces; i++) {
		face *f = v->faces[i];

		for(j = 0; j < 3; j++) {
			vertex *vtx = f->vertices[j];

			if(vtx == v)
				continue;

			// Compact the vertex's face arrays
			for(k = 0; k < vtx->nfaces; k++) {
				if(vtx->faces[k] == f)
					break;
			}

			for(; k + 1 < vtx->nfaces; k++) {
				vtx->faces[k] = vtx->faces[k+1];
			}


			vtx->faces = fw_realloc(vtx->faces, --vtx->nfaces * sizeof(face *));
		}

		we_remove_face(m, f);
	}


	if((v = we_vertex_find(m, v0)) != NULL)
		we_remove_vertex(m, v);
}

/* we_move_vertex: Move vertex @old of mesh @m to coordinates specified by @new->coords.
 * This operation is actually simple, but complicated by the fact that the vertex hash location
 * may change if the coordinates change. Therefore, the hash table must be updated.
 */

vertex	*we_move_vertex(mesh *m, vertex *old, vertex *new)
{
	int oldhashidx, newhashidx;

	vertex *v_old = we_vertex_find(m, old),
		*v_new = we_vertex_find(m, new);

	/* Check that the old vertex exists in @m, and that we are not trying to move @new to an already-existing vertex. 
	 * (Multiple vertices at a single point is not allowed.)
	 */

	if(!v_old) {
		return NULL;
	}

	if(v_new) {
		return v_old;
	}

	/* Change the coordinates, but topological relationships remain intact
	 */

	oldhashidx = we_vertex_hash(v_old);

	v_old->coords = new->coords;

	newhashidx = we_vertex_hash(v_old);

	/* Small optimization: if oldhashidx == newhashidx, no moving required! 
	 * (Probability of this occuring decreases with increasing hash table size :-() 
	 */

	if(newhashidx == oldhashidx)
		return v_old;

	/* Move from old hash location */

	if(v_old->prev)
		v_old->prev->next = v_old->next;
	if(v_old->next)
		v_old->next->prev = v_old->prev;

	if(v_old == m->vertices[oldhashidx]) { 
		m->vertices[oldhashidx] = v_old->next;
	}

	/* ..to new hash location */

	v_old->next = m->vertices[newhashidx];
	if(m->vertices[newhashidx])
		m->vertices[newhashidx]->prev = v_old;
	v_old->prev = NULL;
	m->vertices[newhashidx] = v_old;
	
	/* DONE! */

	return v_old;
}

/* we_coordinate_min: Find the minimum x,y,z extents of face @f.
 * Used in fw_splitpolygons/loadbalancer.c
 */

coordinate	we_coordinate_min(face *f)
{
	coordinate res;

	assert(f != NULL);
	assert(f->vertices[0] != NULL);
	assert(f->vertices[1] != NULL);
	assert(f->vertices[2] != NULL);

	res.x = Vmin( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, x );
	res.y = Vmin( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, y );
	res.z = Vmin( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, z );

	return res;
}


/* we_coordinate_min: Find the maximum x,y,z extents of face @f.
 * Used in fw_splitpolygons/loadbalancer.c
 */

coordinate	we_coordinate_max(face *f)
{
	coordinate res;

	assert(f != NULL);
	assert(f->vertices[0] != NULL);
	assert(f->vertices[1] != NULL);
	assert(f->vertices[2] != NULL);

	res.x = Vmax( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, x );
	res.y = Vmax( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, y );
	res.z = Vmax( f->vertices[0]->coords, f->vertices[1]->coords, f->vertices[2]->coords, z );

	return res;
}



/* we_triangle_area: Compute the area of the triangle specified by vertices @v1, @v2 and @v3.
 * Used in fw_splitpolygons/loadbalancer.c
 *
 * Retuns 0 if the triangle is degenerate. 
 */

double		we_triangle_area(vertex *v1, vertex *v2, vertex *v3)
{
	vector v1_v2 = vector_sub(vector_create(&v2->coords), vector_create(&v1->coords));
	vector v1_v3 = vector_sub(vector_create(&v3->coords), vector_create(&v1->coords));

	double area = 0.5 * vector_length(v1_v2) * vector_length(vector_cross(v1_v2, v1_v3)) / vector_length(v1_v2);

	return isnan(area) ? 0 : area;
}

