/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* fw_wingededge.h: The winged-edge structure. See also fw_wingededge.c
 */


#ifndef FW_WINGEDEDGE
#define FW_WINGEDEDGE

#include <stdio.h>
#include <assert.h>

#include "fw_coord.h"

/* Forward declarations 
 */

struct face;
struct edge;
struct vertex;

#define VERTEX_HASH_SIZE	100		// Size of the vertex hash. 100 is reasonable; see fw_wingededge.c

#define VERTEX_SLOT_UNUSED	0		// Constants defining used/unused hash slots
#define VERTEX_SLOT_USED	1

/* The winged-edge structure has four components: 
 * - vertex
 * - face
 * - edge
 * - mesh
 */

typedef struct vertex {
	char used;		// Denote if this object is available or used in the vertex hash table
	coordinate coords;	// Coordinates of this vertex
	struct face **faces; 	// Faces incident to this vertex
	struct edge **edges; 	// Edges incident to this vertex
	int nfaces, nedges;	// Size of two above lists
	struct vertex *next;	// Next/previous vertex in hash table
	struct vertex *prev;
} vertex;

typedef struct face {
	double area;				// Area of face
	struct edge *startedge;		// Starting edge of face
	struct vertex *vertices[3];		// Vertices
	int rnode; 				// Responsible node
	struct face *next, *prev;		// Next/previous in linked list of all mesh faces
} face;

typedef struct edge {
	struct edge *prev_ccw, *prev_cw, *next_ccw, *next_cw;	// See thesis Chapter 2 for a description of prev/next cw/ccw edges
	struct face *leftface, *rightface;			// Left/right face
	struct vertex *start, *end;				// Start/end vertex
	struct edge *next, *prev;				// Next/previous in linked list of all mesh edges
} edge;


typedef struct mesh {
	struct face *faces;				// List of faces
	struct edge *edges;				// List of edges
	struct vertex *vertices[VERTEX_HASH_SIZE];	// List of vertices, but implemented as a hash table for quick lookups
	int nfaces, nedges, nvertices;			// # of faces/edges/vertices
	struct mesh *next, *prev;			// Next/previous mesh in the global mesh list (declared in fw_wingededge.c)
} mesh;

// See fw_wingededge.c for a description of the below functions.

mesh				*we_mesh_create(void);
void				we_mesh_destroy(mesh *m);

face 				*we_add_polygon(mesh *m, vertex *v1, vertex *v2, vertex *v3);
void				we_remove_polygons(mesh *m, vertex *v0);

vertex				*we_vertex_find(mesh *m, vertex *v);
void				we_vertex_add(mesh *m, vertex **v);

edge				*we_edge_add(mesh *m, vertex *v1, vertex *v2);
edge				*we_edge_find(vertex *v1, vertex *v2);

void				we_remove_face(mesh *m, face *f);
face				*we_face_add(mesh *m, edge *e1, edge *e2, edge *e3);

vertex				*we_move_vertex(mesh *m, vertex *old, vertex *new);

coordinate			we_coordinate_min(face *f);
coordinate			we_coordinate_max(face *f);

double				we_triangle_area(vertex *v1, vertex *v2, vertex *v3);

/* Wrapper around we_triangle_area for face objects.
 */

static inline double		we_triangle_facearea(face *f) {
	return we_triangle_area(f->vertices[0], f->vertices[1], f->vertices[2]);
}


#endif
