/* Framework for Polygon Structures on Clusters
 *
 * Leif Christian Larsen <leifchl@idi.ntnu.no>
 *
 */

/* genscript.c: Program for generating script files
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <assert.h>

#include "../fw_wingededge.h"
#include "../fw_coord.h"
#include "../fw_main.h"

// Generate a random number between 0 and M (using higher-order bits)

#define RANDVTX(M) ((int) (random()/(RAND_MAX + 1.0) * M ))

#define	MESHES		16

// User-defined parameters
static int n = 0, m = 0, k = 0;
static int avgsize = 0;
static int npolys = 0;
static int repeats = 0;
static int cycles = 0;
static int movedistance = 0;

static double avgarea = 0.0;		 // average triangle area


/* usage: Print usage information
 * @av0 original program string, as typed by user
 */

static void usage(const char *av0)
{
	printf("usage: %s <NxMxK> <approx. average polygon size> <#polygons> <move-extract cycles> <move distance> <repeats> <output file>\n\n", av0);
	printf("<NxMxK> specifies the voxel space size\n");
	printf("<average polygon size> specifies the APPROXIMATE average polygon size used in the script.\n");
	printf("<#polygons> specifies the number of polygons.\n");
	printf("<move-extract cycles> specifies the number of move-extract cycles (see thesis).\n");
	printf("<move distance> specifies the APPROXIMATE average distance a vertex is moved.\n");
	printf("<repeats> specifies how many repetitions to run\n");
	printf("Note: If the average area deviates significantly from specified area, try:\n");
	printf(" - increasing voxel space size\n");
	printf(" - changing the approx. area parameter\n");
	printf(" - rerun program\n");

}

/* randvertex_distance: Generate a random vertex connected to v0 and v1 such that the area of the convex hull (triangle)
 * is approximately (and that is, VERY approximately) equal to the average size specified when running genscript
 */

static coordinate randvertex_distance(coordinate v0, coordinate v1)
{
	float dist;
	coordinate mid_v0v1;


	// assuming equilateral triangle;  A = 1/2 |v1-v0| * dist((v1,v0), V)
	// => dist(v1,v0) = 2A/|v1-v0|

	dist = 2*avgsize/(float)coordinate_length(coordinate_sub(v1, v0));

	// assuming equal extents in x,y,z direction: d = sqrt(3*e^2) = sqrt(3)*e
	// divide by three to approximately compensate for the RANDVTX(20) adds/subtracts
	dist /= 3;

	// midpoint of v0, v1
	mid_v0v1 = coordinate_add(v0, coordinate_scale_float(coordinate_sub(v1,v0), 0.5));

	if(RANDVTX(2) == 0)
		mid_v0v1.x += round(dist) - RANDVTX(20);
	else
		mid_v0v1.x -= round(dist) + RANDVTX(20);

	if(RANDVTX(2) == 0)
		mid_v0v1.y += round(dist) - RANDVTX(20);
	else
		mid_v0v1.y -= round(dist) + RANDVTX(20);

	if(RANDVTX(2) == 0)
		mid_v0v1.z += round(dist) - RANDVTX(20);
	else
		mid_v0v1.z -= round(dist) + RANDVTX(20);

	if(mid_v0v1.x >= n)
		mid_v0v1.x -= 2*round(dist) + 20;
	if(mid_v0v1.x < 0)
		mid_v0v1.x += 2*round(dist) + 20;

	if(mid_v0v1.y >= m)
		mid_v0v1.y -= 2*round(dist) + 20;
	if(mid_v0v1.y < 0)
		mid_v0v1.y += 2*round(dist) + 20;

	if(mid_v0v1.z >= k)
		mid_v0v1.z -= 2*round(dist) + 20;
	if(mid_v0v1.z < 0)
		mid_v0v1.z += 2*round(dist) + 20;
	
	// In special cases, we still may not be >= 0 or < nmk. ensure that we are

	mid_v0v1.x %= n;
	mid_v0v1.y %= m;
	mid_v0v1.z %= k;


	return coordinate_abs(mid_v0v1);	
}


/* randvertex_distance_single: Generate a random vertex which is approximately (VERY approximately) movedistance away from v0
 */

static coordinate randvertex_distance_single(coordinate v0)
{

	// rationale: movedistance = sqrt(3*d^2) => movedistance = sqrt(3) d => d = movedistance/sqrt(3)
	// but since we 50% of the time move in negative direction, we need to multiply by 2
	
	int d = round(movedistance*2/sqrt(3.0));
	coordinate result = v0;


	if(RANDVTX(2) == 0)
		result.x += RANDVTX(d);
	else
		result.x -= RANDVTX(d);

	if(RANDVTX(2) == 0)
		result.y += RANDVTX(d);
	else
		result.y -= RANDVTX(d);

	if(RANDVTX(2) == 0)
		result.z += RANDVTX(d);
	else
		result.z -= RANDVTX(d);

	if(result.x >= n)
		result.x -= d;
	if(result.x < 0)
		result.x += d;

	if(result.y >= m)
		result.y -= d;
	if(result.y < 0)
		result.y += d;

	if(result.z >= k)
		result.z -= d;
	if(result.z < 0)
		result.z += d;

	// In special cases, we still may not be >= 0 or < nmk. ensure that we are

	result.x %= n;
	result.y %= m;
	result.z %= k;

	return coordinate_abs(result);
}


/* generate: Generate script file. The general format of the script file is, in EBNF: 
 *     <NxMxK>:<approx. average size>:<polygons>:<repeats>:<#polygon meshes>:<move-extract cycles per repeat>\n
 *     ((<c>|<a(m,[v1][v2][v3])>|<r(m[v])>|<m(m[v][NewV])|<e>|d(m))+!)+ 
 * where:
 * <c = create mesh>
 * <a = add vertex to mesh m, vertices v1,v2,v3>
 * <r = remove vertex v of mesh m>
 * <m = move vertex v of mesh m, to NewV>
 * <e = extract all voxels and do computation>
 * <d = destroy mesh id m>
 * <! = end of current repeat>
 *
 * This program generates files of the form
 * <sequence of create operations> <sequence of add operations> e <sequence of move operations> e <sequence of move operations> ... (repeated 'cycles' times) 
 * d <sequence of remove operations> <sequence of mesh destruction operations> ......... <sequence of create operations> repeated 'repeats' time
 */


static void generate(FILE *f)
{
	int i, j, u, v, w;
	coordinate_list *allcoords;	 // list holding the coordinates "currently" in the mesh


	srandom(time(NULL));
	fprintf(f, "%dx%dx%d:%d:%d:%d:%d:%d\n", n, m, k, avgsize, MESHES * (npolys/MESHES), repeats, MESHES, cycles);

	for(i = 0; i < repeats; i++) {
		allcoords = coordinate_list_create((npolys + 16) * 3);

		for(j = 0, v = 0; j < MESHES; j++, v += npolys/MESHES + 3) {
			coordinate v0 = coordinate_create(RANDVTX(n-1), RANDVTX(m-1), RANDVTX(k-1)), 
				v1 = randvertex_distance_single(v0), 
				v2 = randvertex_distance(v0, v1);

			coordinate_list_add(allcoords, &v0);
			coordinate_list_add(allcoords, &v1);
			coordinate_list_add(allcoords, &v2);

			fprintf(f, "ca([%d,%d,%d][%d,%d,%d][%d,%d,%d])", v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);

			for(u = 0; u < npolys/MESHES; u++) {
				coordinate vc;
				int found, k;

				// Find a vertex such that triangle approximately has specified area.
				// Since the same vertex may be returned twice, which we want to avoid,
				// this is encapsulated in the do-while loop.

				do { 

					found = 0;

					vc = randvertex_distance(v0, v1);
						
					for(k = v; k < v + npolys/MESHES + 3 && k < allcoords->used; k++) {
						if(coordinate_equal(allcoords->coords[k], vc)) {
							++found;
							break;
						}
					} 

				} while(found > 0);

				fprintf(f, "a([%d,%d,%d][%d,%d,%d][%d,%d,%d])", v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, vc.x, vc.y, vc.z);
			
				coordinate_list_add(allcoords, &vc);

				
				// compute triangle area

				vertex v1_V, v2_V, v3_V;
				v1_V.coords = v0;
				v2_V.coords = v1;
				v3_V.coords = vc;
				avgarea += we_triangle_area(&v1_V, &v2_V, &v3_V)/npolys;
				


				if(RANDVTX(2) == 0)
					v0 = vc;
				else
					v1 = vc;
			}
		}

		fprintf(f, "e");

		/* Create move-extract cycles. Note that the first add/extract cycle counts as one move-extract cycle.
		 */

		for(w = 0; w < cycles - 1; w++) {

			/* Move vertices. We check if the destination coordinate is already used. This algorithm is slow and could be sped up using a hash table. This does not matter since:
			 * -- This is the genscript program, only used for generating random, fictional test data and not used in any timings
			 * -- There will not be nearly enough coordinates to make these loops impact running time in practice. This holds even for very large data sets.
			 */
			

			for(j = 0, v = 0; j < MESHES; j++, v += npolys/MESHES + 3) {
				for(u = 0; u < npolys/MESHES + 3; u++) {
					int found, k;
					coordinate vc;
					do { 
						found = 0;

						vc = randvertex_distance_single(allcoords->coords[u + v]);
						
						for(k = v; k < v + npolys/MESHES + 3 && k < allcoords->used; k++) {
							if(coordinate_equal(allcoords->coords[k], vc)) {
								++found;
								break;
							}
						} 
					} while(found > 0);


					fprintf(f, "m(%d[%d,%d,%d][%d,%d,%d])", j, allcoords->coords[v + u].x, allcoords->coords[v + u].y, allcoords->coords[v + u].z, vc.x, vc.y, vc.z);
					allcoords->coords[v + u] = vc;
				}
			}



			fprintf(f, "e");
		}
			
		for(j = 0, u = 0; j < MESHES; j++) {
			for(v = 0; v < npolys/MESHES + 3; v++, u++)
				fprintf(f, "r(%d[%d,%d,%d])", j, allcoords->coords[u].x, allcoords->coords[u].y, allcoords->coords[u].z);
		}

		for(u = 0; u < MESHES; u++) {
			fprintf(f, "d(%d)", u);
		}

		fprintf(f, "!");

		coordinate_list_free(allcoords);
	
	}
}

/* main: Program entry point
 * @argc argument count
 * @argv argument vector
 */

int main(int argc, char **argv)
{
	FILE *outfile;
	printf("genscript - generate script files for cpfw\n");

	if(argc != 8) {
		usage(argv[0]);
		return -1;
	}

	if(sscanf(argv[1], "%dx%dx%d", &n, &m, &k) != 3) {
		printf("--- Error: Invalid argument #1. Must be on the form NxMxK\n");
		usage(argv[0]);
		return -1;
	}

	avgsize = atoi(argv[2]);

	if(avgsize < 10) {
		printf("--- Warning: Average size < 10, using average size = 10\n");
		avgsize = 10;	
	}
	
	npolys = atoi(argv[3]);

	if(npolys < MESHES) {
		printf("--- Error: Use at least %d polygons.\n", MESHES);
		usage(argv[0]);
		return -1;
	}


	cycles = atoi(argv[4]);
	if(cycles <= 0) {
		printf("--- Error: # of move-extract cycles invalid\n");
		usage(argv[0]);
		return -1;
	}	

	movedistance = atoi(argv[5]);
	if(movedistance <= 0) {
		printf("--- Error: move-distance must be > 0\n");
		usage(argv[0]);
		return -1;
	}	


	repeats = atoi(argv[6]);
	
	if(repeats <= 0) {
		printf("--- Warning: No repeats specified. Using repeat = 1\n");
		repeats = 1;
	}

	outfile = fopen(argv[7], "w");

	if(!outfile) {
		printf("--- Error: Unable to open output file %s: %s\n", argv[7], strerror(errno));
	}

	generate(outfile);

	printf("Wrote cpfw script to %s - actual average area is %f\n", argv[7], avgarea);



	fclose(outfile);

	return 0;
}

